#!/bin/sh
# each test script sources this file to import main() and its dependencies.

MDVIEW_BIN="${MDVIEW_BIN:-./mdview}"
CMARK_BIN="${CMARK_BIN:-cmark}"
GFM_BIN="${GFM_BIN:-cmark-gfm}"
PANDOC_BIN="${PANDOC_BIN:-pandoc}"

TESTSCRIPTNAME="${0##*/}"

__usage () { # {{{1
	cat << EOF
Usage: env [MDVIEW_BIN=./mdview] [CMARK_BIN=cmark] [PANDOC_BIN=pandoc] \\
       ${0##*/} [OPTIONS] [FORMAT...] [ | highlight_diff.awk ]

Optionally, pipe output to 'highlight_diff.awk' to colorize differences.

$(fmt -80 << MAN
For each FORMAT: $TESTSCRIPTNAME (this script) runs mdview on a SUBJECT file
to produce a RESULT file for the output FORMAT; then it compares an existing
REFERENCE file against the RESULT file and reports differences. The test run is
marked PASS if no differences can be found, otherwise it is marked FAIL. This
script exists non-zero if at least one run was marked FAIL or if other errors
occurred.
Typically:
- Write SUBJECT file (markdown).
- Pass --create to create a REFERENCE file for each specified FORMAT.
- Run $TESTSCRIPTNAME to test regressions or non-conformance.
Optionally:
- Edit $TESTSCRIPTNAME to change test FORMAT(s).
- Edit $TESTSCRIPTNAME to change the default comparison procedure.[1]
- Edit $TESTSCRIPTNAME to change the default creation procedure.[2]

[1] Change the body of function 'test_diff'. Choose a diff_* function from file
'common.sh' or write your own function instead, and call it in 'test_diff'.

[2] Change the body of function 'test_create'. Choose a predefined create_*
function from file 'common.sh' or write your own function instead, and call it
in 'create_diff'.
MAN
)

FORMATS:
  ansi, html (default), po, text, tty.

OPTIONS:
  Unrecognized options are passed to function 'test_diff'.
  --cmark   Compare cmark https://github.com/commonmark/cmark against mdview.
  --create  Create REFERENCE file - overwriting without prompt.
  --eol     Insert '␤' before each compared line ending to reveal end-of-line.
  --gfm     Compare cmark-gfm https://github.com/github/cmark-gfm against mdview.
  --pandoc  Compare pandoc https://github.com/jgm/pandoc against mdview.
  --viewer  Helper to start mdview GUI passing the test options; nothing else.
  --vim     Inspect differences with vim instead of test_diff(), and PASS test.
EOF
}

# Set traps and create temporary folder. {{{1
TMPD=$(mktemp -d -p "${TMPDIR:-/tmp}" "${0##*/}_XXXXXX")
# created with permissions u+rwx minus umask restrictions
chmod 700 "$TMPD" || exit 1
trap 'command rm -rf "'"${TMPD:?}"'"; exit $STATUS' \
	HUP INT ALRM TERM 0 # USR1 USR2
mkdir -p "$TMPD"

# {{{1}}}

# DO NOT CHANGE COMMAND VALUES HERE!
# if you need to change do it in the test script, not here
# change AFTER sourcing this file
MDVIEW="${MDVIEW:-$MDVIEW_BIN}"
CMARK="${CMARK:-$CMARK_BIN --smart --nobreaks}"
GFM="${GFM:-$GFM_BIN --smart --nobreaks}"
PANDOC="${PANDOC:-$PANDOC_BIN -f markdown-smart-auto_identifiers -t html}"

# The test script initializes TEST_LABEL.
REFERENCE_DIR='test/reference'
REFERENCE_STEM="${REFERENCE_DIR:?}/${TEST_LABEL:?}"
RESULT_STEM="$TMPD/${TEST_LABEL:?}"

show_line_ending () {
local script
# U+2424 symbol for newline
[ -n "$_opt_eol" ] && script='s/$/␤/' || script=''
sed -e "$script" "$@"
}

# $@-main_args
test_create () {
	create_default "$@"
}

### create reference file from subject file
# $@-main_args
create_default () {
	# stop here unless explicit overwrite is requested
	[ -n "$OVERWRITE_TEST_FILES" ] || return 0

	write_reference_files "$@"
}

### create subject and reference file from commonmark spec
# $1-short_description $2@-commonmark_example_id... (list of)
# commonmark_example_id from CM specification, single number <n> or range <n>-<m>
# optionally prepend 'T', e.g. T<n> or T<n>-<m> to mark acceptable deviations
create_cm_conformance () {
	# stop here unless explicit overwrite is requested
	[ -n "$OVERWRITE_TEST_FILES" ] || return 0

	# create subject file
	test/script/lib/prepare_commonmark_test_range.sh "$@" &&
		# create reference file
		write_reference_files html
}

# The test script initializes TEST_ARGS, TEST_LABEL
# $@-main-args
write_reference_files () {
	: "ensure: ${TEST_LABEL:?} ${MDVIEW:?} ${REFERENCE_STEM:?}"
	local p ret=0
	for p in "$@"; do
		# sift valid test formats
		case $p in
			ansi|html|po|text|tty) :;; *) continue ;;
		esac
		: echo "RUN $MDVIEW $TEST_ARGS --$p" >&2
		$MDVIEW $TEST_ARGS --$p > "$REFERENCE_STEM.$p"
		if [ $? -eq 0 ]; then
			printf "\033[7m  %s  \033[0m %s %s\n" "CREATED" "$TEST_LABEL" "$p"
		else
			printf "\033[7;31m  %s  \033[0m %s %s\n" "ERROR" "$TEST_LABEL" "$p"
			ret=1
		fi
	done
return $ret
}

# $* ansi|html|text|tty ...
# The test script initializes TEST_ARGS, TEST_FORMATS
run_test () { # => REFERENCE_STEM
	: "ensure: ${MDVIEW:?} ${RESULT_STEM:?}"
		# convenience helper to simply start the viewer
		if [ -n "$_opt_viewer" ]; then
			echo "RUN> $MDVIEW $TEST_ARGS $mdview_opts" >&2
			$MDVIEW $TEST_ARGS $mdview_opts
			exit $?
		fi

		[ $? -eq 0 ] || ret=1
	local p ret=0 mdview_opts
	for p in ${@:-$TEST_FORMATS}; do
		mdview_opts=
		[ -n "$_opt_cmark" ] && mdview_opts="$mdview_opts --cm-block-end"

		########################
		#  output result file  #
		########################
		echo "RUN> $MDVIEW $TEST_ARGS $mdview_opts --$p" >&2
		$MDVIEW $TEST_ARGS $mdview_opts --$p > "$RESULT_STEM.$p"
		[ $? -eq 0 ] || ret=1


		##################################################
		#  output alternative reference file (optional)  #
		##################################################
		if [ -n "$_opt_cmark" ]; then
			: "ensure: ${CMARK:?}"
			if [ $p = html ]; then
				echo "RUN< $CMARK ${TEST_ARGS##* }" >&2
				$CMARK ${TEST_ARGS##* } |
					massage_cmark_output "$RESULT_STEM.cmark.$p"
				[ $? -eq 0 ] || ret=1
				REFERENCE_STEM="$RESULT_STEM.cmark"
			fi


		elif [ -n "$_opt_gfm" ]; then
			: "ensure ${GFM:?}"
			if [ $p = html ]; then
				echo "RUN< $GFM ${TEST_ARGS##* }" >&2
				$GFM ${TEST_ARGS##* } |
					massage_cmark_output "$RESULT_STEM.gfm.$p"
				[ $? -eq 0 ] || ret=1
				REFERENCE_STEM="$RESULT_STEM.gfm"
			fi


		elif [ -n "$_opt_pandoc" ]; then
			: "ensure: ${PANDOC:?}"
			if [ $p = html ]; then
				:; echo "RUN< $PANDOC ${TEST_ARGS##* }" >&2
				$PANDOC ${TEST_ARGS##* } > "$RESULT_STEM.pandoc.$p"
				[ $? -eq 0 ] || ret=1
				REFERENCE_STEM="$RESULT_STEM.pandoc"
			elif [ $p = text ]; then
				$PANDOC_BIN -f markdown -t plain ${TEST_ARGS##* } > "$RESULT_STEM.pandoc.$p"
				[ $? -eq 0 ] || ret=1
				REFERENCE_STEM="$RESULT_STEM.pandoc"
			fi
		fi
	done
	return $ret
}

massage_cmark_output () {
	# slurp the stream
	gawk -v 'RS=a\\yb' -v OUT="$1" '###gawk{{{
	{
		# cmark --smart replaces ellipsis but mdview does not
		gsub(/…/, "...")
		# they go n-dash where we go m-dash
		gsub(/–/, "—")
		# they may end the stream with an extra \n, we do not
		sub(/\n\n$/, "")

		# This is the most important difference: cmark joins
		# lines inside <p> or <li> with line feed while mdview
		# with space. We try to make make this unnoticeable by
		# replacing line feed with space as needed.
		buf = $0
		sub(/\n$/, "", buf)
		gsub (/<br \/>\n/, "HARDBREAK_LINEFEED", buf)
		alen = split(buf, a, /<p>|<\/p>/, sep)
		buf  = ""
		for (i = 1; i <= alen; i++) {
			if (sep[i - 1] == "<p>")
				gsub ("\n", " ", a[i])
			buf = (buf sep[i - 1] a[i])
		}
		gsub ("HARDBREAK_LINEFEED", "<br />\n", buf)
		print buf > OUT
		close(OUT)
	}
	END { exit !(NR > 0) }
	###gawk}}}'
}
# $* html|tty|text ...
run_test_diffs () {
	: "ensure ${REFERENCE_STEM:?} ${RESULT_STEM:?}"
	local ret differ reff tmpf
	[ -n "$_opt_vim" ] && differ=vimdiff || differ=test_diff
	for p in "${@:-html}"; do
			reff="$REFERENCE_STEM.$p" tmpf="$RESULT_STEM.$p" &&
			if ! [ -e "$reff" ]; then
				>&2 printf "error: no reference file: pass --create to initialize: %s\n" "$reff"
				ret=1
			else
				show_line_ending "$reff" > "$tmpf.e.nl" &&
				show_line_ending "$tmpf" > "$tmpf.a.nl" &&
				: echo "CALL $differ $_opts_test_diff \"$tmpf.e.nl\" \"$tmpf.a.nl\"" >&2 &&
				$differ $_opts_test_diff "$tmpf.e.nl" "$tmpf.a.nl"
			[ $? -eq 0 ] || ret=1
			fi
	done
	return $ret
}

# $1-reference_file $2-result_file
test_diff () {
	diff_default "${1:?}" "${2:?}"
}

# $1-reference_file $2-result_file
diff_default () {
	diff_git_word_diff "${1:?}" "${2:?}"
}

# $1-reference_file $2-result_file
diff_plain () {
	diff "${1:?}" "${2:?}"
}

# $1-reference_file $2-result_file
diff_y () {
	diff -y --suppress-common-lines "${1:?}" "${2:?}"
}

# $1-reference_file $2-result_file
diff_git_word_diff () {
	git -P diff --word-diff "$1" "$2"
}

# $1-reference_file $2-result_file $3-subject_file
# for test files prepared by test/prepare_commonmark_test_range.sh
# use test script option --eof for accurate results
diff_cm_conformance () {
	if [ -n "$_opt_vim" ]; then
		vimdiff "${1:?}" "${2:?}"
	else
		sub_diff_cm_conformance "${1:?}" "${2:?}" "${3:?}"
	fi

	return $?
}

# $1-reference_file $2-result_file $3-subject_file
# !!! For accurate results run the test script with option --eol !!!
sub_diff_cm_conformance () {
	# $2 is located in $TMPDIR
	local referencef="${1:?}" resultf="${2:?}" subjectf="${3:?}"

	local capture cases test_case_header_regex
	test_case_header_regex='(<p>)?([0-9]+) (DEVIATING|CONFORMING)'
	cases=$(grep -Ec "$test_case_header_regex" "$referencef")

	# capture stderr leaving stdout alone
	{ capture=$(\
	git -P diff --no-index --word-diff=porcelain -- "$referencef" "$resultf" |
		gawk \
		-v TEST_CASE_HEADER_REGEX="$test_case_header_regex" \
		-v CASES=$cases '
###gawk {{{
##### purpose: minimize git diff output for humans

BEGIN {
	cases=CASES+0
	non_conforming_cases = deviating_cases = start = 0
}

# git diff headers
/^--- a\// || /^\+\+\+ b\// { next }

# differences before first test case header
! start && /^[-+]/ {
	print
	++non_conforming_cases
}

# end previous block
$0 ~ TEST_CASE_HEADER_REGEX && start > 0 {
	if (show) {
		show = 0
		do_show()
	}
}
function do_show(   a) {
	# do not count a small deviation from CM spec as FAIL
	match (R[start], TEST_CASE_HEADER_REGEX, a)
	if (a[3] == "DEVIATING") {
		--non_conforming_cases
		++deviating_cases
	}

	for (i = start; i < NR; i++)
		print R[i]
}

# start new block
$0 ~ TEST_CASE_HEADER_REGEX {
	delete R; start = NR; show = 0
}

start > 0 {
	R[NR] = $0
	# remember to show the current block iff it differs
	if (!show && $0 ~ /^[-+]/)
		show = ++non_conforming_cases
}

END {
	if (show)
		do_show()

	printf "%d %d %.2f %d %.2f\n", cases, \
		non_conforming_cases, non_conforming_cases * 100 / cases, \
		deviating_cases, deviating_cases * 100 / cases \
	> "/dev/stderr"
	exit non_conforming_cases > 0
}
###gawk }}}' 2>&1 >&3 3>&-); } 3>&1

	### "validate" output with html-tidy: ###
	# -mute less-serious warning types; unmuted types could be serious errors,
	# such as improperly nested tags, i.e. <em><strong></em></strong>
	tidy -eqi -file "$resultf.err" \
		--doctype omit --vertical-space no --show-body-only yes --wrap 0 --mute-id yes \
		--mute MISSING_ATTRIBUTE,MISSING_TITLE_ELEMENT,\
NESTED_EMPHASIS,COERCE_TO_ENDTAG,DISCARDING_UNEXPECTED \
	"$resultf"
	local tidy_status=$?

	# report tidy warnings/errors in context
	if [ -s "$resultf.err" ]; then
		>&2 gawk \
			-v TEST_CASE_HEADER_REGEX="$test_case_header_regex" '
###gawk {{{
BEGINFILE { ++fnum }
fnum == 1 { S[FNR] = $0; next } # subject file   (markdown)
fnum == 2 { R[FNR] = $0; next } # reference file (cmark)
fnum == 3 { I[FNR] = $0; next } # result file    (mdview)
fnum == 4 {                     # error file     (tidy)
# error line example: line 1 column 1 - Warning: inserting implicit <body> (INSERTING_TAG)
	if ($0 ~ /<body>/) next
	lineno = $2+0
	E[lineno] = (E[lineno] "\n" $0)
}
END {
	# read test case number a[2] from block comment
	max = length(S)
	for (i = 1; i <= max; i++) {
		if (match(S[i], TEST_CASE_HEADER_REGEX, a)) {
			testnum = a[2]
			if (1 == index(S[i+2], "<!--")) {
				for (j = i + 3; j <= max; j++) {
					if (1 == index(S[j], "-->")) {
						i = j
						break
					}
					T[testnum] = T[testnum] "\n\t" S[j]
				}
			}
		}
	}
}
END {
	if (length(E) > 0)
		print "SELECTION OF POSSIBLY SERIOUS ERRORS:"
	for (lineno in E) {
		#print (substr(E[lineno], 2) "\n\t" I[lineno])
		#printf "line % 4s: %s\n", lineno, I[lineno]
		testnum = lineno_to_testnum(lineno)
		printf "case % 4s: %s%s\n", testnum, I[lineno], T[testnum]
	}
}
function lineno_to_testnum (lineno,   a) {
	while (--lineno > 0)
		if (match(R[lineno], TEST_CASE_HEADER_REGEX, a))
			return a[2]
	return -1
}
###gawk }}}' \
			"$subjectf" "$referencef" "$resultf" "$resultf.err"

	fi

	### summarize results ###

	# using diff -U to count diffs is accurate as long as the two files are quite similar
	local dff=$(diff -U 0 "$1" "$2" | grep ^@ -c)

	set -- $capture
	local tc=$1 ncc=$2 nccpc=$3 dvc=$4 dvcpc=$5

	set --
	set -- "$@" 'total test cases' 0 "$tc" ""
	set -- "$@" 'line diff count - reference vs. result'
		[ $dff -eq 0 ] && set -- "$@" 0 || set -- "$@" 31; set -- "$@" "$dff" ""
	set -- "$@" 'non-conforming cases'
		[ $ncc -eq 0 ] && set -- "$@" 0 || set -- "$@" 31; set -- "$@" "$ncc" "$nccpc %"
	set -- "$@" 'known deviations' 0 $dvc "$dvcpc %"

	printf >&2 '%-40s : \033[%dm% 4d\033[0m %8s\n' "$@"

	[ $ncc -eq 0 -a $tidy_status -ne 2 ]
}

# optional: $* ansi|html|tty|text (valid test formats)
# $* overrides TEST_FORMATS.
# The test script initializes TEST_ARGS, TEST_LABEL, TEST_FORMATS.
main () { # => $STATUS
	: "ensure: ${TEST_LABEL:?}"
	local p failed
	unset _opts_test_diff \
	_opt_cmark \
	_opt_create \
	_opt_eol \
	_opt_gfm \
	_opt_pandoc \
	_opt_viewer \
	_opt_vim \
	;
	[ -n "$TEST_FORMATS" ] || TEST_FORMATS='html'

	while ! [ "${1#-}" = "$1" ]; do
		case "$1" in
			-h|--help) __usage; STATUS=0 exit ;;
			--cmark  ) _opt_cmark=1
				#recommended MDVIEW="$MDVIEW_BIN --cm-block-end --no-smart --no-ext"
				;;
			--create ) OVERWRITE_TEST_FILES=1 ;;
			--eol    ) _opt_eol=1 ;;
			--gfm    ) _opt_gfm=1 ;;
				#recommended MDVIEW="$MDVIEW_BIN --cm-block-end --no-smart --no-ext"
			--pandoc ) _opt_pandoc=1
				#recommended MDVIEW="$MDVIEW_BIN --cm-block-end --no-smart --no-ext --unsafe-html"
			;;
			--viewer ) _opt_viewer=1 ;;
			--vim    ) _opt_vim=1 ;;
			-*       )
				echo >&2 "OPTION> $1 will be passed to function test_diff"
				_opts_test_diff="$_opts_test_diff $1"
				;;
		esac
		shift
	done

	STATUS=0
	export G_DEBUG=fatal-warnings
	if ! test_create ${@:-$TEST_FORMATS}; then
		STATUS=1
		exit $STATUS
	fi

	for p in ${@:-$TEST_FORMATS}; do
		case $p in
			ansi|po) [ -z "$_opt_cmark$_opt_gfm_$_opt_pandoc" ] || continue ;;
			html) : ;; # common ground among mdview, cmark, cmark-gfm and pandoc
			text) [ -z "$_opt_cmark$_opt_gfm_" ] || continue ;;
			tty ) [ -z "$_opt_cmark$_opt_gfm_$_opt_pandoc" ] || continue ;;
			*   ) continue ;;
		esac

		run_test "$p"
		if [ $? -ne 0 ]; then
			printf "\033[7m  %s  \033[0m %s %s\n" \
			"INVALID" "$TEST_LABEL" "$p"
			false
		else
			run_test_diffs "$p"
			failed=$?
			if [ $failed -eq 0 ]; then
				printf "\033[7m  %s  \033[0m %s %s\n" \
				"PASS" "$TEST_LABEL" "$p"
			else
				STATUS=1
				printf "\033[7;31m  %s  \033[0m %s %s\n" \
				"FAIL $failed" "$TEST_LABEL" "$p"
			fi
		fi
	done
	return $STATUS
}

