/* -*- Mode: C; tab-width: 4; indent-tabs-mode: t; c-basic-offset: 4 -*-
 *
 * Copyright (C) 2008 Richard Hughes <richard@hughsie.com>
 * Copyright (C) 2009 Leandro Pereira <leandro@hardinfo.org>
 * Copyright (C) 2015 James B
 * Copyright (C) 2016, 2023 step
 *
 * Licensed under the GNU General Public License Version 2
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#define _GNU_SOURCE
#include <stdio.h>
#include <string.h>
		/* disable G_ADD_PRIVATE macro deprecation warning introduced in 2.58 */
#define GLIB_VERSION_MIN_REQUIRED	GLIB_VERSION_2_56
#include <glib.h>
#include <libintl.h>
#include <ctype.h>

//#define DEBUG g_print
#define DEBUG

#include "egg-markdown.h"
#include "egg-markdown-style.h"

/*******************************************************************************
 *
 * This is a simple Markdown parser.
 *
 * Features:
 * - Output to Pango, HTML, vt100 codes, ANSI CSI escape codes or plain text.
 * - Partially conform to CommonMark specification 0.30 (with multiple limitations).
 *
 * Extensions:
 * - Auto-code: auto-replace certain word patterns with code spans.
 * - Smart text: auto-replace single/double quotes and m-dash.
 * - Pipe tables: display markdown pipe tables as code blocks.
 * - Multi-line heading: heading requires blank line to end.
 * - Extract or replace gettext translation data.
 *
 * The following limitations are already known, and properly deliberate:
 * - Single level unordered lists only
 * - No ordered list support
 * - No blockquote section support
 * - No HTML entity escaping support
 *   Not thread-safe (GLib data structures aren't)
 *
 * It does support the rest of the standard pretty well, although it's not
 * been run against any conformance tests. The parsing is single pass, with
 * a simple enumerated interpreter mode and a single line back-memory.
 *
 ******************************************************************************/

/*******************************************************************************
 * We implement some text formatting operations using Unicode Private Use Area
 * (PUA) codepoints. Input markdown shall not include the PUA characters we use.
 ******************************************************************************/

/**
 * https://www.unicode.org/faq/private_use.html
 * https://github.com/silnrsi/unicode-resources/tree/main/sil-pua
 **/

/*
* Markdown hard line-break (2+ spaces at the end of a line) is stored in
* pending memory as sUNIPUA_BR, which is presented to gettext as MSGSTR_BR.
* See egg_markdown_to_text_line_formatter_gettext.
*/
#define iUNIPUA_BR         0xF48A
#define sUNIPUA_BR         "\357\222\212"     /* EF928A */
#define MSGSTR_BR               "<%BR/%>"
/*
* Inline code spans are stored internally as sUNIPUA_CODE.
*/
#define iUNIPUA_CODE       0xF489
#define sUNIPUA_CODE       "\357\222\211"
#define rUNIPUA_CODE       "\\x{F489}"
/*
* Inline link and image spans are stored internally as sUNIPUA_LINK.
*/
#define iUNIPUA_LINK       0xF488
#define sUNIPUA_LINK       "\357\222\210"
#define rUNIPUA_LINK       "\\x{F488}"
/*
* Inline raw HTML segments are stored internally as sUNIPUA_HTML.
*/
#define iUNIPUA_HTML       0xF487
#define sUNIPUA_HTML       "\357\222\207"
#define rUNIPUA_HTML       "\\x{F487}"

/*
* Input lines are joined together in back-memory self->priv->pending separated
* by PENDING_SEPARATOR (egg_markdown_add_pending) - chosen to be a space) " ".
* Space separated text is easy to parse. If the separator included non-space
* section- or word-boundary characters, parsing would be more complex.
*/
#define PENDING_SEPARATOR " "
/* make it stand out while debugging */
//#define PENDING_SEPARATOR "┼"

/*
CommonMark's cmark replaces raw HTML with the following comment
*/
#define SAFE_HTML "<!-- raw HTML omitted -->"

/*******************************************************************************
 ******************************************************************************/

static void     egg_markdown_finalize	(GObject		*object);

#define EGG_MARKDOWN_GET_PRIVATE(o) (G_TYPE_INSTANCE_GET_PRIVATE ((o), EGG_TYPE_MARKDOWN, EggMarkdownPrivate))

typedef gchar *(EggMarkdownLinkBuilder)(const gchar *text, const gchar *dest, const gchar *title, const gint link_dest_id);
typedef gchar *(EggMarkdownImageBuilder)(const gchar *text, const gchar *dest, const gchar *title, const gint link_dest_id);
typedef gchar *(EggMarkdownFenceBuilder)(const EggMarkdown *self);

typedef enum {
	EGG_MARKDOWN_MODE_BLANK        = 1 << 0,
	EGG_MARKDOWN_MODE_RULE         = 1 << 1,
	EGG_MARKDOWN_MODE_BULLET       = 1 << 2,
	EGG_MARKDOWN_MODE_PARA         = 1 << 3,
	EGG_MARKDOWN_MODE_CODEBLOCK    = 1 << 4,
	EGG_MARKDOWN_MODE_H1           = 1 << 5,
	EGG_MARKDOWN_MODE_H2           = 1 << 6,
	EGG_MARKDOWN_MODE_H3           = 1 << 7,
	EGG_MARKDOWN_MODE_H4           = 1 << 8,
	EGG_MARKDOWN_MODE_H5           = 1 << 9,
	EGG_MARKDOWN_MODE_H6           = 1 << 10,
	EGG_MARKDOWN_MODE_COMMENTBLOCK = 1 << 11,
	EGG_MARKDOWN_MODE_UNKNOWN      = 1 << 12,
	EGG_MARKDOWN_MODE_BULLET_FIRST = 1 << 13,
} EggMarkdownMode;

#define EGG_MARKDOWN_MODE_HEADING ( \
	EGG_MARKDOWN_MODE_H1 | \
	EGG_MARKDOWN_MODE_H2 | \
	EGG_MARKDOWN_MODE_H3 | \
	EGG_MARKDOWN_MODE_H4 | \
	EGG_MARKDOWN_MODE_H5 | \
	EGG_MARKDOWN_MODE_H6   \
	)

typedef enum {
	EGG_MARKDOWN_CODEBLOCK_NONE,
	EGG_MARKDOWN_CODEBLOCK_STD,
	EGG_MARKDOWN_CODEBLOCK_SHEBANG,
	EGG_MARKDOWN_CODEBLOCK_FENCED,
	EGG_MARKDOWN_CODEBLOCK_TABLE,
} EggMarkdownCodeblockMode;

typedef enum {
	EGG_MARKDOWN_CODEBLOCK_SYNTAX_NONE,
	EGG_MARKDOWN_CODEBLOCK_SYNTAX_GETTEXT
} EggMarkdownCodeblockSyntax;

typedef struct {
	const gchar *em_start;
	const gchar *em_end;
	const gchar *strong_start;
	const gchar *strong_end;
	const gchar *code_span_start;
	const gchar *code_span_end;
	const gchar *codeblock_start;
	const gchar *codeblock_end;
	const gchar *codeblock_table_start;
	const gchar *codeblock_table_end;
	const gchar *codeblock_shebang_start;
	const gchar *codeblock_shebang_end;
	const gchar *strikethrough_start;
	const gchar *strikethrough_end;
	const gchar *h1_start;
	const gchar *h1_end;
	const gchar *h2_start;
	const gchar *h2_end;
	const gchar *h3_start;
	const gchar *h3_end;
	const gchar *h4_start;
	const gchar *h4_end;
	const gchar *h5_start;
	const gchar *h5_end;
	const gchar *h6_start;
	const gchar *h6_end;
	const gchar *ulist_start;
	const gchar *ulist_end;
	const gchar *bullet_start;
	const gchar *bullet_end;
	const gchar *rule;
	const gchar *para_start;
	const gchar *para_end;
	const gchar *br;

	EggMarkdownLinkBuilder *link_builder;
	EggMarkdownImageBuilder *image_builder;
	EggMarkdownFenceBuilder *fence_builder;
} EggMarkdownTags;

struct EggMarkdownPrivate
{
	EggMarkdownExtensions      extensions;
	EggMarkdownMode            mode;
	EggMarkdownMode            blockcomment_prev_mode;
	EggMarkdownCodeblockMode   codeblock_mode;
	EggMarkdownCodeblockSyntax codeblock_syntax;
	gchar                     *codeblock_fence_info; //  CM info string
	gchar                      codeblock_fence_char[1];
	guint                      codeblock_fence_char_count;
	guint                      codeblock_fence_indent;
	gint                       codeblock_start_line_idx;
	gint                       codeblock_first_line_idx;
	gint                       line_curr_idx;     // parser's, egg_markdown_parse
	guint                      line_last_idx;     // parser's
	gint                       line_delta;        // parser's
	gchar                    **lines;             // parser's input lines
	gboolean                   eof;               // no more lines, TRUE after parser's loop
	gboolean                   ulist_loose;
	guint                      ulist_length;
	EggMarkdownTags            tags;
	EggMarkdownOutput          output;
	gint                       max_lines;
	guint                      line_count;        // count of output lines
	gboolean                   smart_text;
	gboolean                   escape;
	gboolean                   autocode;
	gboolean                   autocode_shebang;  // autocode entire file that has shebang
	gboolean                   use_gettext;       // translate content via gettext
	gboolean                   cm_block_end;       // see COMMONMARK_FENCED_CODEBLOCK_LINE_ENDING
	gboolean                   unsafe_html;       // imitate cmark's --unsafe
	GString                    *pending;
	GString                    *processed;
	GArray                     *regex_table;
	GArray                     *code_table;
	GArray                     *link_table;
	GArray                     *html_table;

	// for pot_output
	gchar*                     input_filename;
	FILE*                      pot_output;        // where to send pot output
	gboolean                   temporary_disable_pot;

	// code blocks mode save
	gboolean                   codeblock_prev_pot;

	// code block table mode
	guint                      table_columns;
};

/**
 * EggMarkdownWordType
 **/
typedef enum {
	EGG_MARKDOWN_WORD_UNKNOWN = 0,     /* any word not of the following types */
	EGG_MARKDOWN_WORD_URI,             /* http:// ... */
	EGG_MARKDOWN_WORD_URI_FLANKED,     /* [<(]http://[)>] ... */
	EGG_MARKDOWN_WORD_ABS_PATH,        /* reasonably long /pathname */
	EGG_MARKDOWN_WORD_FILE_DIFF,       /* ".diff" ".patch" file name */
	EGG_MARKDOWN_WORD_BUGZILLA,        /* #asciiWord */
	EGG_MARKDOWN_WORD_FUNCNAME,        /* ident"()" */
	EGG_MARKDOWN_WORD_EMAIL,           /* @email address */
	EGG_MARKDOWN_WORD_UIDENT,          /* uppercase identifier with "_" */
} EggMarkdownWordType;

/**
 * EggMarkdownRegexType
 **/
typedef enum {
	EGG_MARKDOWN_REGEX_CODE = 0,        /* md code spans */
	EGG_MARKDOWN_REGEX_CODE_REF,        /* internal code_refs */
	EGG_MARKDOWN_REGEX_LINK,            /* md link and image element */
	EGG_MARKDOWN_REGEX_LINK_DEST_REF,   /* internal link_dest_ref for internet URI and image path */
	EGG_MARKDOWN_REGEX_LINK_TEXT_REF,   /* internal link_text_ref for link text and image alt text */
	EGG_MARKDOWN_REGEX_LINK_TITLE_REF,  /* internal link_title_ref for link and image title */
	EGG_MARKDOWN_REGEX_HTML,            /* HTML inline */
	EGG_MARKDOWN_REGEX_HTML_REF,        /* internal html_ref */

	/* keep last */
	EGG_MARKDOWN_REGEX_LEN,             /* length of regex_table */
} EggMarkdownRegexType;

/**
 * EggMarkdownSliceType
 **/
typedef enum {
	/* any match */
	EGG_MARKDOWN_SLICE_MATCH = 0,     /* text of a whole match */
	EGG_MARKDOWN_SLICE_BEFORE,        /* text before a MATCH */
	EGG_MARKDOWN_SLICE_AFTER,         /* text after the last MATCH */
	EGG_MARKDOWN_SLICE_IGNORED,
	EGG_MARKDOWN_SLICE_UNKNOWN,

	/* code spans */
	EGG_MARKDOWN_SLICE_CODE_CODE,      /* code between backticks */
	EGG_MARKDOWN_SLICE_CODE_MID,       /* text before optional TICK and CODE */
	EGG_MARKDOWN_SLICE_CODE_TICK,      /* if exactly 1 space, it signals that CODE includes backtick(s) */
	EGG_MARKDOWN_SLICE_CODE_END,       /* text after CODE and optional TICK */

	/* links */
	EGG_MARKDOWN_SLICE_LINK_TYPE_URL,  /* "[" */
	EGG_MARKDOWN_SLICE_LINK_TYPE_IMG,  /* "![" */
	EGG_MARKDOWN_SLICE_LINK_TEXT,      /* text between square brackets */
	EGG_MARKDOWN_SLICE_LINK_DEST,      /* URL or path between brackets */
	EGG_MARKDOWN_SLICE_LINK_TITLE,     /* optional accessibility title */
	EGG_MARKDOWN_SLICE_LINK_MID,       /* text before the next TYPE or between TEXT and DEST or between DEST and TITLE */
	EGG_MARKDOWN_SLICE_LINK_END,       /* text after the CODE and optional TITLE */

	/* inline html */
	EGG_MARKDOWN_SLICE_HTML_TAG_NAME,  /* open/closing tag name */
} EggMarkdownSliceType;

/* debugging helpers C99 */
// Define DEBPRTELM={0,1,2} to trace protected elements.
#if DEBPRTELM > 0
static const gchar * const _slice_type_name[] = {
	[EGG_MARKDOWN_SLICE_MATCH]         = "MATCH",
	[EGG_MARKDOWN_SLICE_BEFORE]        = "BEFORE",
	[EGG_MARKDOWN_SLICE_AFTER]         = "AFTER",
	[EGG_MARKDOWN_SLICE_IGNORED]       = "IGNORED",
	[EGG_MARKDOWN_SLICE_UNKNOWN]       = "UNKNOWN",

	[EGG_MARKDOWN_SLICE_CODE_CODE]     = "CODE_CODE",
	[EGG_MARKDOWN_SLICE_CODE_MID]      = "CODE_MID",
	[EGG_MARKDOWN_SLICE_CODE_TICK]     = "CODE_TICK",
	[EGG_MARKDOWN_SLICE_CODE_END]      = "CODE_END",

	[EGG_MARKDOWN_SLICE_LINK_TYPE_URL] = "LINK_TYPE_URL",
	[EGG_MARKDOWN_SLICE_LINK_TYPE_IMG] = "LINK_TYPE_IMG",
	[EGG_MARKDOWN_SLICE_LINK_TEXT]     = "LINK_TEXT",
	[EGG_MARKDOWN_SLICE_LINK_DEST]     = "LINK_DEST",
	[EGG_MARKDOWN_SLICE_LINK_TITLE]    = "LINK_TITLE",
	[EGG_MARKDOWN_SLICE_LINK_MID]      = "LINK_MID",
	[EGG_MARKDOWN_SLICE_LINK_END]      = "LINK_END",

	[EGG_MARKDOWN_SLICE_HTML_TAG_NAME] = "HTML_TAG_NAME",
};
#endif /* DEBPRTELM */

static const gchar * const _tag_info[] = {
	[EGG_MARKDOWN_TAG_DEST_LINK_URI_ID]   = "dest=Lu",
	[EGG_MARKDOWN_TAG_DEST_IMAGE_PATH_ID] = "dest=Ip",
};

/**
 * _strchug:
 * Modified from glib's g_strchug, which removes a larger white space set.
 * @string: a string to remove the leading ' ' and '\t' from.
 * _strchug modifies @string in place.
 */
static inline gchar *
_strchug (gchar *string)
{
	guchar *start;
	g_return_val_if_fail (string != NULL, NULL);

	for (start = (guchar*) string;*start && (*start == ' ' || *start == '\t'); start++)
		;
	memmove (string, start, strlen ((gchar *) start) + 1);
	return string;
}

/**
 * _strchomp:
 * Modified from glib's g_strchomp, which removes a larger white space set.
 * @string: a string to remove the trailing ' ' and '\t' from.
 * _strchomp modifies @string in place.
 */
static inline gchar *
_strchomp (gchar *string)
{
	gsize len;
	g_return_val_if_fail (string != NULL, NULL);

	len = strlen (string);
	while (len--) {
		if (' ' == (guchar)string[len] || '\t' == (guchar)string[len])
			string[len] = '\0';
		else
			break;
	}
	return string;
}

/* remove leading and trailing ' ' and '\t' characters */
#define _strstrip(string)   g_strchomp (g_strchug (string))

/**
 * egg_markdown_tag_get_info:
 * @tag: pango "font" tag
 * @subject: EggMarkdownTagInfo
 * Return: `gint` @subject's value parsed from @tag
 **/
gint
egg_markdown_tag_get_info (EggMarkdown *self, const gchar *tag, const EggMarkdownTagInfo subject) {
	gchar *p = NULL;
	gint ret  = -1;

	g_return_val_if_fail (self != NULL, -1);
	g_return_val_if_fail (tag  != NULL, -1);
	g_return_val_if_fail (subject < EGG_MARKDOWN_TAG_INFO_LEN, -1);
	p = strstr (tag, _tag_info[subject]);
	if (p)
		ret = atoi (p + strlen (_tag_info[subject]));
	return ret;
}

/**
 * EggMarkdownSlice
 * String slice specialized for markdown text
 **/
typedef struct {
	EggMarkdownSliceType type;
	gint    start;
	gint    len;   /* bytes */
	gint    data;
} EggMarkdownSlice;

G_DEFINE_TYPE (EggMarkdown, egg_markdown, G_TYPE_OBJECT)

static inline void
skip_blanks (gchar **p) {
	while (**p == ' ' || **p == '\t') (*p)++;
}

/**
 * egg_markdown_to_text_line_is_rule:
 *
 * Horizontal rules are created by placing three or more hyphens, asterisks,
 * or underscores on a line by themselves.
 * You may use spaces between the hyphens or asterisks.
 **/
static gboolean
egg_markdown_to_text_line_is_rule (const gchar *line)
{
	guint i;
	guint len;
	guint count = 0;
	gchar *copy = NULL;
	gboolean ret = FALSE;

	len = strnlen (line, EGG_MARKDOWN_MAX_LINE_LENGTH);
	if (len == 0)
		goto out;

	/* replace non-rule chars with ~ */
	copy = g_strdup (line);
	g_strcanon (copy, "-*_ ", '~');
	for (i=0; i<len; i++) {
		if (copy[i] == '~')
			goto out;
		if (copy[i] != ' ')
			count++;
	}

	/* if we matched, return true */
	if (count >= 3)
		ret = TRUE;
out:
	g_free (copy);
	return ret;
}

/**
 * egg_markdown_to_text_line_is_bullet:
 * Deviation from CM spec: no support for bullet level > 1
 **/
static inline gboolean
egg_markdown_to_text_line_is_bullet (const gchar *line)
{
	/* regex /^ ?[-*+] / */
	gchar *p = (gchar *)line;
	return *p &&
	(*p == '-' || *p == '*' || *p == '+') && *(++p) == ' ' ||
	(*p == ' ' && (*(++p) == '-' || *p == '*' || *p == '+') && *(++p) == ' ');
}

/**
 * egg_markdown_to_text_line_is_heading_atx:
 * Test if @line is an ATX-style heading level 1 through 6.
 * Return: the heading's level mode enum or 0 if @line isn't a heading.
 * Deviations from CM spec: see comments below.
 **/
static inline EggMarkdownMode
egg_markdown_to_text_line_is_heading_atx (const gchar *line, EggMarkdownSlice *slice)
{
	/*
	https://spec.commonmark.org/0.30/#atx-headings
	CM: "ATX headings need not be separated from surrounding content by
	blank lines, and they can interrupt paragraphs."

	Deviation: we don't support optional trailing #'s.
	Deviation: we don't ignore trailing spaces - two or more spaces map to <br>.

	Mdview extension: Contrary to CM spec, non-empty lines following '#'
	are joined together to form the heading.
	*/

	gchar *p = (gchar *)line, *q;
	guint level;

	while (*p == ' ')
		++p;
	if ((ptrdiff_t)(p - line) > 3)
		return 0;
	q = p;
	while (*q == '#')
		++q;
	if (p == q || (level = (ptrdiff_t)(q - p)) > 6 || *q != ' ')
		return 0;
	slice->data = level;
	slice->start = (ptrdiff_t)(q - line) + 1;
	/* TODO future: use slice->len */
	return EGG_MARKDOWN_MODE_H1 << (level - 1);
}

/**
 * egg_markdown_to_text_line_is_heading1_setext:
 * Deviations from CM spec: see comments below.
 * Same mdview extension as egg_markdown_to_text_line_is_heading_atx.
 **/
static inline gboolean
egg_markdown_to_text_line_is_heading1_setext (const gchar *line)
{
	/* non-conforming! to https://spec.commonmark.org/0.30/#setext-headings */
	return line[0] == '=' && line[1] == '=' && line[2] == '=' && !strcspn (line + 3, "= \t");
}

/**
 * egg_markdown_to_text_line_is_heading2_setext:
 * Deviations from CM spec: see comments below.
 * Same mdview extension as egg_markdown_to_text_line_is_heading_atx.
 **/
static inline gboolean
egg_markdown_to_text_line_is_heading2_setext (const gchar *line)
{
	/* non-conforming! to https://spec.commonmark.org/0.30/#setext-headings */
	return line[0] == '-' && line[1] == '-' && line[2] == '-' && !strcspn (line + 3, "- \t");
}

/**
 * egg_markdown_to_text_line_is_textdomain:
 **/
static inline gboolean
egg_markdown_to_text_line_is_textdomain (const gchar *line)
{
	return g_str_has_prefix (line, TEXTDOMAIN_DIRECTIVE);
}

/**
 * egg_markdown_to_text_line_is_nopot:
 **/
static inline gboolean
egg_markdown_to_text_line_is_nopot (const gchar *line)
{
	return g_str_has_prefix (line, NOPOT_DIRECTIVE);
}

/**
 * egg_markdown_to_text_line_is_shebang:
 **/
static inline gboolean
egg_markdown_to_text_line_is_shebang (const gchar *line)
{
	return line[0] == '#' && line[1] == '!';
}

/**
 * egg_markdown_to_text_line_is_indented_code
 **/
static inline gboolean
egg_markdown_to_text_line_is_indented_code (const EggMarkdown *self, const gchar *line)
{
	/*
	https://spec.commonmark.org/0.30/#indented-code-blocks
	CM: An indented code block cannot interrupt a paragraph, so there must
	be a blank line between a paragraph and a following indented code block.
	CM: Any non-blank line with fewer than four spaces of indentation ends
	the code block. A paragraph may occur immediately after indented code.
	*/
	return self->priv->mode != EGG_MARKDOWN_MODE_PARA &&
		(*line == '\t' || strspn (line, " ") >= 4);
}

/**
 * egg_markdown_to_text_line_is_codeblock_fence_start:
 * @line markdown
 * Return: TRUE iff a start fence can be parsed, in which case set @self's
 * `codeblock_syntax` and `codeblock_lang` - the latter to the CommonMark
 * "info_string" value.
 **/
static gboolean
egg_markdown_to_text_line_is_codeblock_fence_start (const EggMarkdown *self, const gchar *line)
{
	/*
	https://spec.commonmark.org/0.30/#fenced-code-blocks
	https://spec.commonmark.org/0.30/#example-119 - 147
	Exceptions:
	- mdview is compiled with -UCOMMONMARK_FENCED_CODEBLOCK_LINE_ENDING
	- see detailed cases in test/script/commonmark_codeblock_fence.sh
	*/
	gchar *p = (gchar *)line;
	guint len;
	gboolean match = FALSE;

	len = strspn (p, " ");
	if (len <= 3) {
		self->priv->codeblock_fence_indent = len;
		p += len;
		len = strspn (p, "`");
		if (len == 0)
			len = strspn (p, "~");
		if (len >= 3) {
			self->priv->codeblock_fence_char[0] = *p;
			p += len;
			self->priv->codeblock_fence_char_count = len;
			self->priv->codeblock_syntax = EGG_MARKDOWN_CODEBLOCK_SYNTAX_NONE;
			g_free (self->priv->codeblock_fence_info);
			self->priv->codeblock_fence_info = NULL;
			match = *p == '\0';
			if (!match) { /* not yet but still possible */
				self->priv->codeblock_fence_info = _strstrip (g_strdup (p));

				if (/*example 145*/ !(self->priv->codeblock_fence_char[0] == '`' &&
					strchr (self->priv->codeblock_fence_info, '`')) &&
						/*example 138*/ strcspn (self->priv->codeblock_fence_info, self->priv->codeblock_fence_char) > 0) {
					match = TRUE;
					if (strcmp (self->priv->codeblock_fence_info, "gettext") == 0)
						self->priv->codeblock_syntax = EGG_MARKDOWN_CODEBLOCK_SYNTAX_GETTEXT;
					/* g_free (self->priv->codeblock_fence_info); */
				}
			}
		}
	}
	return match;
}

/**
 * egg_markdown_to_text_line_is_codeblock_fence_end:
 * @line markdown
 * Return: FALSE if no end fence can be parsed, otherwise return TRUE.
 **/
static gboolean
egg_markdown_to_text_line_is_codeblock_fence_end (const EggMarkdown *self, const gchar *line)
{
	gchar *p = (gchar*)line;
	guint len;
	gboolean match = FALSE;
	gchar fence[1];

	len = strspn (p, " \t");
	if (len <= 3) {
		*fence = self->priv->codeblock_fence_char[0];
		p += len;
		len = strspn (p, fence);
		if (len >= self->priv->codeblock_fence_char_count) {
			p += len;
			p += strspn (p, " \t");
			match = *p == '\0';
		}
	}
	return match;
}

/**
 * egg_markdown_fencebuilder_pango:
 * @info markdown fenced codeblock info string
 **/
static inline gchar *
egg_markdown_fencebuilder_pango (const EggMarkdown *self)
{
	return g_strdup (self->priv->tags.codeblock_start);
}

/**
 * egg_markdown_fencebuilder_html:
 * @info markdown fenced codeblock info string
 **/
static inline gchar *
egg_markdown_fencebuilder_html (const EggMarkdown *self)
{
	/* the CM specification doesn't mandate this treatment of the info string
	 * but we imitate cmark's output for now, e.g. example  143 */
	if (self->priv->codeblock_fence_info) {
		guint len = strcspn (self->priv->codeblock_fence_info, " \t");
		gchar c = self->priv->codeblock_fence_info[len];
		self->priv->codeblock_fence_info[len] = '\0';
		gchar *p = g_strconcat ("<pre><code class=\"language-", self->priv->codeblock_fence_info, "\">", NULL);
		self->priv->codeblock_fence_info[len] = c;
		return p;
	}
	return g_strdup (self->priv->tags.codeblock_start);
}

/**
 * egg_markdown_fencebuilder_text:
 * @info markdown fenced codeblock info string
 **/
static inline gchar *
egg_markdown_fencebuilder_text (const EggMarkdown *self)
{
	return g_strdup (self->priv->tags.codeblock_start);
}

static inline gboolean
egg_markdown_extension_enabled(EggMarkdown *self, const EggMarkdownExtensions ext) {
	return (self->priv->extensions & ext);
}

/**
 * egg_markdown_to_text_line_is_blank:
 **/
static gboolean
egg_markdown_to_text_line_is_blank (const gchar *line)
{
	guint i;
	guint len;
	gboolean ret = TRUE;

	len = strnlen (line, EGG_MARKDOWN_MAX_LINE_LENGTH);

	/* https://spec.commonmark.org/0.30/#characters-and-lines */

	for (i=0; i<len; i++)
		if (line[i] != ' ' && line[i] != '\t') {
			ret = FALSE;
			break;
		}

	return ret;
}

/**
 * egg_markdown_to_text_line_is_blockcomment_start:
 **/
static inline gboolean
egg_markdown_to_text_line_is_blockcomment_start (const gchar *line) {
	return line[0] == '<' && line[1] == '!' && line[2] == '-' && line[3] == '-';
}

/**
 * egg_markdown_to_text_line_is_blockcomment_end:
 *
 * Deviation from CM: EGG MD deliberately does not end the block
 * comment on a blank line.
 **/
static inline gboolean
egg_markdown_to_text_line_is_blockcomment_end (const gchar *line) {
	/*
	CM: "Note that anything on the last line [of the HTML block] after
	the end tag will be included in the HTML block".{1}
	step: Therefore no output for TEXT in "<!-- \n --> TEXT" and
	"<!-- --> TEXT".
	{1} https://spec.commonmark.org/0.30/#html-blocks (type 2)
	*/
	return g_strstr_len(line, -1, "-->") != 0;
}

/**
 * egg_markdown_to_text_line_has_line_break:
 **/
static inline gboolean
egg_markdown_to_text_line_has_line_break (const gchar *line)
{
	return g_str_has_suffix (line, "  ");
}

/**
 * _count_pipe_table_columns: helper
 * @line: *line != 0
 **/
static inline guint
_count_pipe_table_columns (const gchar *line)
{
	guint n;
	gboolean b;

	/* strip up to 3 spaces (cmark-gfm) */
	gchar *p = (gchar *)line;
	for (n = 0; n < 3 && *p == ' '; p++)
		;

	/* count inner non-escaped pipes */
	gchar *q = p;
	if (*p == '|')
		p++; /* don't count optional opening pipe */

	for (n = b = 0; *p; p++)
		switch (*p) {
			case '\\': b = !b; break;
			case ' ' : case '\t': break;
			case '|' : if (!b) n++;
			default: q = p;
		}

	if (*q == '|')
		n--; /* don't count optional closing pipe */

	return n > 0 ? n + 1 : 0;
}

/**
 * egg_markdown_to_text_line_is_pipe_table_start:
 * one-line look-ahead
 **/
static gboolean
egg_markdown_to_text_line_is_pipe_table_start (const gchar *line, const gchar *next_line, guint *columns)
{
	gint dashes, col;
	gchar *p, *q;

	if (!*line || next_line == NULL)
		return FALSE;

	/* column header - count inner pipes */
	*columns = _count_pipe_table_columns (line);

	/* header underlines - count columns */

	/* strip up to 3 spaces (cmark-gfm) */
	p = (gchar *)next_line;
	for (guint i = 0; i < 3 && *p == ' '; p++)
		;
	if (*p == '|')
		p++; /* ignore optional opening pipe */

	for (col = 0; col < *columns && *p; ++col, p++) {
		dashes = 0;
		while (*p && *p == ' ')
			p++;
		if (*p && *p == ':') /* left-align option */
			p++, dashes++;
		while (*p && *p == '-')
			p++, dashes++;
		if (*p && *p == ':') /* right-align option */
			p++, dashes++;
		while (*p && *p == ' ')
			p++;
		if (*p && *p != '|' && *p != '+')
			break;
		if (dashes < 3)
			break;
	}

	return col > 0 && col == *columns;
}

/**
 * egg_markdown_to_text_line_is_pipe_table_end:
 **/
static inline gboolean
egg_markdown_to_text_line_is_pipe_table_end (const gchar *line, const guint columns)
{
	if (line == NULL || !*line)
		return TRUE;
	return _count_pipe_table_columns(line) != columns;
}

/**
 * _replace_f1f2:
 * Replace all matches of regex pattern @f1(@f2)? where @f1 and @f2 are fixed strings.
 * @f1 string
 * @f2 string, nullable
 * @replace string replacement
 * @pcount   pointer to signed integer, nullable
 * Return:   the replaced string and the number of replacements in *@pcount
 **/
static gchar *
_replace_f1f2
(const gchar *haystack, const gchar *f1, const gchar *f2, const gchar *replace, gint *pcount)
{
	gint count = 0;
	gchar *ret;

	if (*haystack == '\0' || *f1 == '\0') {
		ret = g_strdup (haystack);
	}
	else {
		GString *hay = g_string_new (haystack);
		GString *f12 = g_string_new (f1);
		if (f2) {
			g_string_append (f12, f2);
			count += g_string_replace (hay, f12->str, replace, 0);
			g_string_truncate (f12, strlen (f1));
		}
		count += g_string_replace (hay, f12->str, replace, 0);
		ret = hay->str;
		g_string_free (hay, FALSE);
		g_string_free (f12, TRUE);
	}
	if (pcount)
		*pcount = count;

	return ret;
}

#ifdef DEBPRTELM
/* standout, standout end */
#define _SO    "\033[7m"
#define _SE    "\033[0m"
#define _SObla "\033[7;30m"
#define _SOred "\033[7;31m"
#define _SOgre "\033[7;32m"
#define _SOyel "\033[7;33;46m"
#define _SOblu "\033[7;34m"
#define _SOmag "\033[7;35m"
#define _SOcya "\033[7;36m"
#define _SOwhi "\033[7;37m"
#endif

#if DEBPRTELM > 0
static void
_print_priv_table (const GArray *priv_table, const gchar *fmt, ...)
{
	va_list args;
	va_start (args, fmt);
	gchar *msg = g_strdup_vprintf (fmt, args);
	va_end (args);
	g_printerr ("%s", msg);
	g_free (msg);

	if (priv_table->len == 0)
		g_printerr (_SO"EMPTY TABLE"_SE"\n");
	else for (gint i = 0; i < priv_table->len; i++) {
		gchar *elem = g_array_index (priv_table, gpointer, i);
		g_printerr("% 3d. \"%s\"\n", i, elem);
	}
}

static void
_print_slices (const GArray *slices, const gchar *text, const gchar *fmt, ...)
{
	va_list args;
	va_start (args, fmt);
	gchar *msg = g_strdup_vprintf (fmt, args);
	va_end (args);
	g_printerr ("%s", msg);
	g_free (msg);

	if (slices->len == 0)
		g_printerr (_SO"NO SLICES"_SE"\n");
	else for (gint i = 0; i < slices->len; i++) {
		EggMarkdownSlice *slice = &g_array_index (slices, EggMarkdownSlice, i);
		EggMarkdownSliceType type = slice->type;
		gint start = slice->start;
		gint len   = slice->len;
		gchar data[10]; /* ~itoa */
		sprintf (data, "%d", slice->data);
		g_printerr ("% 3d. % 14s % 3d,%-3d % 3s", i, _slice_type_name[type], start, len, *data == '-' ? "" : data);
		if (text) {
			gchar *p = (gchar*)text + start;
			gchar *q = p + len;
			gchar c = *q;
			*q = '\0';
			g_printerr (" \"%s\"", p);
			*q = c;
		}
		g_printerr ("\n");
	}
}

#endif /* DEBPRTELM */

/**
 * egg_markdown_make_code_ref
 * @id: id of an inline code segment stashed with `egg_markdown_stash_code`.
 * Return:
 * dynamically-allocated string code_ref, which encodes @id
**/
static inline gchar*
egg_markdown_make_code_ref (const guint id) {
	return g_strdup_printf ("%s%dC;%s", sUNIPUA_CODE, id, sUNIPUA_CODE);
}

/**
 * egg_markdown_get_code_id
 * @ref    code_ref text
 * Return: -1   if @text isn't a valid code_ref otherwise return the code id
 **/
static inline gint
egg_markdown_get_code_id (const gchar *ref) {
	gint id = -1;
	if (strncmp (ref, sUNIPUA_CODE, sizeof (sUNIPUA_CODE) - 1) == 0)
		id = atoi (ref + sizeof (sUNIPUA_CODE) - 1);

	return id;
}

/**
 * egg_markdown_stash_code
 * @code: markdown code span or `code_ref`
 * Return:
 * stashed `code_id` - use `egg_markdown_get_code` to retrieve the code text.
 * Line endings are converted to space according to the CM spec.
 * Identical @codes are stashed with the same id.
 **/

static inline guint
egg_markdown_stash_code (EggMarkdown *self, const gchar *code) {
	gint id = egg_markdown_get_code_id (code);
	if (id < 0) {
		gchar *p;
		for (guint i = 0; i < self->priv->code_table->len; i++) {
			p = g_array_index (self->priv->code_table, gchar *, i);
			if (strcmp(code, p) == 0) {
				id = i;
				break;
			}
		}
		if (id < 0) {
			p = g_strdup(code);
			g_array_append_val(self->priv->code_table, p);
			id = self->priv->code_table->len - 1;
		}
	}

	return id;
}

/**
 * egg_markdown_get_code
 * @id: id of a code span previously stashed with `egg_markdown_stash_code`.
 * Return: (the memory is owned by @self) code text with unescaped "\`".
 **/
static gchar *
egg_markdown_get_code(EggMarkdown *self, const gint id)
{
	gchar *ret = NULL;

	g_return_val_if_fail(id < self->priv->code_table->len, NULL);
	return g_array_index(self->priv->code_table, gchar *, id);
}

/*
 * sUNIPUA_LE isn't used anymore but I'd rather keep it around as exemplar for
 * (see) egg_markdown_regex_code. sUNIPUA_LE used to be the placeholder for
 * an intra-paragraph input line ending, e.g. \n. CommonMark's cmark calls it
 * a "soft line break", and outputs it as either a new line (default) or as a
 * space when cmark's option --nobreaks is in effect. Mdview has no concept of
 * soft line break.  Should sUNIPUA_LE be revived in its planned role, then
 * egg_markdown_stash_code must consider when to replace space for sUNIPUA_LE,
 * as the CM spec requires.
 */
#define iUNIPUA_LE         0xF48B
#define sUNIPUA_LE         "\357\222\213"
#define rUNIPUA_LE         "\\x{F48B}"

/**
 * egg_markdown_regex_code:
 * Return: GRegex* matcher for markdown code spans
 **/
static GRegex *
egg_markdown_regex_code (EggMarkdown *self)
{
	static GRegex **regex = NULL;

	/*
	https://spec.commonmark.org/0.30/#code-spans
	https://spec.commonmark.org/0.30/#example-328 - 349
	Pattern can match a code segment spanning lines endings (since \X*?
	matches sUNIPUA_BR, which is converted back to line ending in
	release_emerged_line_endings).
	Note: backslash inside code span is treated literally (#example-338).
	Exceptions:
	- see detailed cases in test/script/commonmark_code_span.sh
	*/

	/* see bib_regular_expression */
	/* https://regex101.com tester */
	/*
	/
	(?|((?<!`)`+(?!`))(\h+)((?<!`)\g{1}(?!`))|((?<!`)`+(?!`))([ \x{F48B}])?([^\x{F489}\x{F488}\x{F487}]+?)((?(2)[ \x{F48B}]))((?<!`)\g{1}(?!`)))
	/gmx
	*/

	if (regex == NULL)
		regex = &g_array_index(self->priv->regex_table, GRegex *, EGG_MARKDOWN_REGEX_CODE);
	if (*regex == NULL) {
		GError *err = NULL;
		*regex = g_regex_new (
		/* See EggMarkdownSliceType for meaning of UPPERCASE words. */
		/* The leading number is the capture group id, which is used
		 * for switch{} case labels in egg_markdown_slice_get_codes. */

		/* a CODE span shall not include other protected spans; */
		/* therefore don't match our rUNIPUA_* Unicode code points. */
		/* However, match rUNIPUA_LE and treat it like space. */

		/* alternative match 1: CODE consists of white space only */
		"(?|"
			"((?<!`)`+(?!`))"              /* 1 MID  */
			"()"                           /* 2 TICK */
			"([\\h"rUNIPUA_LE"]+)"         /* 3 CODE */
			"()"                           /* 4 TICK */
			"((?<!`)\\g{1}(?!`))"          /* 5 END  */

		/* alternative match 2: CODE includes non-white space */
		"|"
			"((?<!`)`+(?!`))"              /* 1 MID  (1st capturing group) */
			"([ |"rUNIPUA_LE"])?"          /* 2 TICK (2nd capturing group) */
			                               /* 3 CODE */
			"([^" rUNIPUA_CODE""rUNIPUA_LINK""rUNIPUA_HTML "]+?)"
			"((?(2)[ |"rUNIPUA_LE"]))"     /* 4 TICK (iff 2nd group captured) */
			"((?<!`)\\g{1}(?!`))"          /* 5 END  (what group 1 captured) */
		")", 0, 0, &err);
		if (err != NULL) {
			g_printerr ("code regex: %s\n", err->message);
			g_error_free (err);
		}
	}

	return *regex;
}

/**
 * egg_markdown_regex_code_ref:
 * Return: GRegex* matcher for internal code_refs
 **/
static GRegex *
egg_markdown_regex_code_ref (EggMarkdown *self)
{
	static GRegex **regex = NULL;

	/* code_ref created by egg_markdown_make_code_ref */
	/* and normalized by egg_markdown_slice_join_codes */

	if (regex == NULL)
		regex = &g_array_index(self->priv->regex_table, GRegex *, EGG_MARKDOWN_REGEX_CODE_REF);
	if (*regex == NULL)
		*regex = g_regex_new (sUNIPUA_CODE"(\\d+)C;"sUNIPUA_CODE, 0, 0, NULL);

	return *regex;
}

/**
 * egg_markdown_slice_find_type
 * @slices GArray of @EggMarkdownSlice elements obtained with
 * egg_markdown_slice_get_codes or egg_markdown_slice_get_links.
 * @type:   @EggMarkdownSliceType value to find in @slices
 * @start:  array index where the search starts.
 * @retptr: address of an @EggMarkdownSlice pointer, NULLable.
 * Return:
 * `gint` array index of the first slice of type @type found;
 * a pointer to the value is returned in *@retptr.
 * -1 if no match, and *@retprt is undefined.
 **/
static gint
egg_markdown_slice_find_type (const GArray *slices, const EggMarkdownSliceType type, const gint start, EggMarkdownSlice **retptr)
{
	gint ret = -1, i;
	EggMarkdownSlice *slice;
	for (i = start; i < slices->len; i++) {
		slice = &g_array_index (slices, EggMarkdownSlice, i);
		if (slice->type == type) {
			if (retptr)
				*retptr = slice;
			ret = i;
			break;
		}
	}
	return (ret);
}

/**
 * egg_markdown_slice_get_codes:
 * @text: the markdown text to harvest for inline code slices.
 * @start_position: as in `g_regex_match_full()`.
 * @match_options: `GRegexMatchFlags`.
 * @span_count: `guint` how many code spans to match, -1 for all.
 * @stash_data:  request stashing span's code segment.
 * Return:
 * `GArray` of @EggMarkdownSlice elements.
 * Array length is set to zero on no match.
 *
 * @text is sliced to facilitate extracting the text of markdown code spans
 * e.g. `CODE`.  A slice value of type EGG_MARKDOWN_SLICE_CODE_CODE
 * includes the start and end positions of the CODE component.  Joining all
 * slices together yields the original @text.
 *
 * If @stash_data is TRUE each CODE is stashed into @self's code_table.
 **/
static GArray *
egg_markdown_slice_get_codes (EggMarkdown *self, const gchar *text, const guint start_position, const GRegexMatchFlags match_options, const guint span_count, const gboolean stash_data)
{
	GRegex *regex = egg_markdown_regex_code (self);
	GMatchInfo *match_info = NULL;
	gint i, END, start = 0, end = 0, done_end = 0, n = 0;
	EggMarkdownSlice slice;
	GArray *slices;

	slices = g_array_new (FALSE, FALSE, sizeof (slice));
	g_regex_match_full (regex, text, -1, start_position, 0, &match_info, NULL);
	while (g_match_info_matches (match_info) && (span_count < 0 || ++n <= span_count)) {
		END = g_regex_get_capture_count (regex);
		for (i = 0; i <= END; i++, start = end = 0) {
			(void)g_match_info_fetch_pos (match_info, i, &start, &end);
			if (start < 0)
				continue;
			slice.start = start;
			slice.len   = end - start;
			slice.data  = -1;
			slice.type  = EGG_MARKDOWN_SLICE_UNKNOWN;
			/* switch case label is the capture group number in egg_markdown_regex_code */
			switch(i) {
				case 0:
					slice.type = EGG_MARKDOWN_SLICE_BEFORE;
					slice.start = done_end;
					slice.len = start - done_end;
					done_end = slice.start + slice.len - 1;
					g_array_append_val (slices, slice);

					slice.type = EGG_MARKDOWN_SLICE_MATCH;
					slice.start = start;
					slice.len   = end - start;
					break;
				case 1:
					slice.type = EGG_MARKDOWN_SLICE_CODE_MID;
					break;
				case 2:
				case 4:
					if (slice.len == 0) /* omit empty TICKs */
						continue;
					slice.type = EGG_MARKDOWN_SLICE_CODE_TICK;
					break;
				case 3:
					slice.type = EGG_MARKDOWN_SLICE_CODE_CODE;
					if (stash_data) {
						gchar *code;
						code = g_strndup (text + slice.start, slice.len);
						slice.data = egg_markdown_stash_code (self, code);
						g_free (code);
					}
					break;
				case 5:
					slice.type = EGG_MARKDOWN_SLICE_CODE_END;
					done_end = slice.start + slice.len;
					break;
			}
			g_array_append_val (slices, slice);
		}
		g_match_info_next (match_info, NULL);
	}

	if (slices->len > 0) {
		/* capture the text after the last END */
		slice.type  = EGG_MARKDOWN_SLICE_AFTER;
		slice.start = done_end;
		slice.len   = strlen (text) - done_end;
		slice.data  = -1;
		g_array_append_val (slices, slice);
#if DEBPRTELM > 0
		_print_slices (slices, text,
			_SOgre"%s: CODE SLICES:"_SE" \"%s\"\n", __FUNCTION__, text);
#endif
	}
	g_match_info_free (match_info);

	return (slices);
}

/**
 * egg_markdown_make_link_dest_ref
 * @id: id of a link DEST previously stashed with `egg_markdown_stash_link_dest`.
 * Return:
 * dynamically-allocated string link_dest_ref, which encodes @id
 *
 * This function also applies to markdown image path.
**/
static inline gchar*
egg_markdown_make_link_dest_ref (const guint id) {
	return g_strdup_printf ("%s%dLu;%s", sUNIPUA_LINK, id, sUNIPUA_LINK);
}

/**
 * egg_markdown_make_link_text_ref
 * @id: id of a link text previously stashed with `egg_markdown_stash_link_text`.
 * Return:
 * dynamically-allocated string link_text_ref, which encodes @id
**/
static inline gchar*
egg_markdown_make_link_text_ref (const guint id) {
	return g_strdup_printf ("%s%dLx;%s", sUNIPUA_LINK, id, sUNIPUA_LINK);
}

/**
 * egg_markdown_make_link_title_ref
 * @id: id of a link title previously stashed with `egg_markdown_stash_link_title`.
 * Return:
 * dynamically-allocated string link_title_ref, which encodes @id
**/
static inline gchar*
egg_markdown_make_link_title_ref (const guint id) {
	return g_strdup_printf ("%s%dLt;%s", sUNIPUA_LINK, id, sUNIPUA_LINK);
}

/**
 * egg_markdown_get_link_dest_id
 * @ref    a `link_dest_ref`
 * Return: -1   if @ref isn't a valid link_dest_ref otherwise return the link DEST id
 **/
static inline gint
egg_markdown_get_link_dest_id (const gchar *ref) {
	gint id = -1;
	if (strncmp (ref, sUNIPUA_LINK, sizeof (sUNIPUA_LINK) - 1) == 0)
		id = atoi (ref + sizeof (sUNIPUA_LINK) - 1);

	return id;
}

/**
 * egg_markdown_get_link_text_id
 * @ref    a `link_text_ref`
 * Return: -1   if @ref isn't a valid link_text_ref otherwise return the link text id
 **/
static inline gint
egg_markdown_get_link_text_id (const gchar *ref) {
	return egg_markdown_get_link_dest_id (ref);
}

/**
 * egg_markdown_get_link_title_id
 * @ref    a `link_title_ref`
 * Return: -1   if @ref isn't a valid link_title_ref otherwise return the link title id
 **/
static inline gint
egg_markdown_get_link_title_id (const gchar *ref) {
	return egg_markdown_get_link_dest_id (ref);
}

/**
 * egg_markdown_stash_link_dest
 * @dest: markdown link destination or `link_dest_ref`
 * Return:
 * stashed `link_dest_id` - use `egg_markdown_get_link_dest` to retrieve the link destination.
 * Identical destinations are stashed with the same id.
 *
 * This function also applies to markdown image path.
 **/
static gint
egg_markdown_stash_link_dest (EggMarkdown *self, const gchar *dest) {
	gint id = egg_markdown_get_link_dest_id (dest);
	if (id < 0) {
		gchar *p;
		for (guint i = 0; i < self->priv->link_table->len; i++) {
			p = g_array_index (self->priv->link_table, gchar *, i);
			if (strcmp(dest, p) == 0) {
				id = i;
				break;
			}
		}
		if (id < 0) {
			p = g_strdup(dest);
			g_array_append_val(self->priv->link_table, p);
			id = self->priv->link_table->len - 1;
		}
	}

	return id;
}

/**
 * egg_markdown_stash_link_text
 * @text: markdown link text
 * Return:
 * stashed `link_text_id` - use `egg_markdown_get_link_text` to retrieve the text
 * Identical @{text}s are stashed with the same id.
 **/
static gint
egg_markdown_stash_link_text (EggMarkdown *self, const gchar *text) {
	gint id = egg_markdown_get_link_text_id (text);
	if (id < 0) {
		gchar *p;
		for (guint i = 0; i < self->priv->link_table->len; i++) {
			p = g_array_index (self->priv->link_table, gchar *, i);
			if (strcmp(text, p) == 0) {
				id = i;
				break;
			}
		}
		if (id < 0) {
			p = g_strdup (text);
			g_array_append_val(self->priv->link_table, p);
			id = self->priv->link_table->len - 1;
		}
	}

	return id;
}

/**
 * egg_markdown_stash_link_title
 * @title: markdown link title
 * Return:
 * stashed `link_title_id` - use `egg_markdown_get_link_title` to retrieve the title
 * Identical @{title}s are stashed with the same id.
 **/
static gint
egg_markdown_stash_link_title (EggMarkdown *self, const gchar *title) {
	gint id = egg_markdown_get_link_title_id (title);
	if (id < 0) {
		gchar *p;
		for (guint i = 0; i < self->priv->link_table->len; i++) {
			p = g_array_index (self->priv->link_table, gchar *, i);
			if (strcmp(title, p) == 0) {
				id = i;
				break;
			}
		}
		if (id < 0) {
			p = g_strdup (title);
			g_array_append_val(self->priv->link_table, p);
			id = self->priv->link_table->len - 1;
		}
	}

	return id;
}

/**
 * egg_markdown_get_link_dest
 * @id: link destination id in @self's link_table
 * Return: dynamically-allocated string.
 * This function also returns link URI or image path.
 **/
gchar *
egg_markdown_get_link_dest(EggMarkdown *self, const gint id)
{
	g_return_val_if_fail(id < self->priv->link_table->len, NULL);
	return g_strdup(g_array_index(self->priv->link_table, gchar *, id));
}

/**
 * egg_markdown_get_link_text
 * @id: link text in @self's link_table
 * Return: dynamically-allocated string.
 **/
static gchar *
egg_markdown_get_link_text(EggMarkdown *self, const gint id)
{
	return egg_markdown_get_link_dest (self, id);
}

/**
 * egg_markdown_get_link_title
 * @id: link title in @self's link_table
 * Return: dynamically-allocated string.
 **/
static gchar *
egg_markdown_get_link_title(EggMarkdown *self, const gint id)
{
	return egg_markdown_get_link_dest (self, id);
}

/**
 * egg_markdown_regex_link:
 * Return: GRegex* matcher for markdown link and image syntax
 **/
static GRegex *
egg_markdown_regex_link (EggMarkdown *self)
{
	static GRegex **regex = NULL;

	/*
	* Nearly complete pattern for https://spec.commonmark.org/0.30/#links
	* and https://spec.commonmark.org/0.30/#images
	* Exceptions:
	* - DEST  : https://spec.commonmark.org/0.30/#link-destination
	* - - pointy brackets "(<...>)" are not supported
	* - TITLE : https://spec.commonmark.org/0.30/#link-title
	* - - balanced parenthesis "(...)" not implemented
	* - - escaped quotes not implemented
	* - smart quotes as delimiters : they are not in the CM spec but I added
	*   them to enable support for smart text.
	*/

bib_regular_expression:
	/* https://docs.gtk.org/glib/struct.Regex.html */
	/* https://perldoc.perl.org/perlre */
	/* https://www.regular-expressions.info/unicode.html */
	/* https://www.regular-expressions.info/conditional.html */

	/* https://regex101.com tester */
	/*
	/
	(?<TYPE>\!?\[)(?<TEXT>[^]]+?)?(\]\x28\h*)(?<DEST>[^\x29]+?)?(?:(\h+([“‘]|(?<Q>['\"])))(?<TITLE>.+?)?([’”]|\g{Q}))?(\h*\x29)
	/gmx
	*/

	if (regex == NULL)
		regex = &g_array_index(self->priv->regex_table, GRegex *, EGG_MARKDOWN_REGEX_LINK);
	if (*regex == NULL) {
		GError *err = NULL;
		*regex = g_regex_new (
		/* See EggMarkdownSliceType for meaning of UPPERCASE words. */
		/* The leading number is the capture group id, which is used
		 * for switch{} case labels in egg_markdown_slice_get_links. */
		/* Capture group names are used in _formatter_link_cb. */

			"(?<TYPE>\\!?\\[)"           /* 1 TYPE   */
			"(?<TEXT>[^]]+?)?"           /* 2 TEXT   */
			"(\\]\\x28\\h*)"             /* 3 MID (  */
			"(?<DEST>[^\\x29]+?)?"       /* 4 DEST ) */
		"(?:"
			"(\\h+"                      /* 5 MID, space + open quote */
				"([“‘]|"                   /* 6 smart/straight quotes */
				"(?<Q>['\"]))"             /* 7 straight quote */
			")"
			"(?<TITLE>.+?)?"             /* 8 TITLE  */
			"([’”]|\\g{Q})"              /* 9 END, close quote, merge next ... */
		")?"
			"(\\h*\\x29)"                /* 10 ... END, merge previous  */
		, 0, 0, &err);
		if (err != NULL) {
			g_printerr ("link regex: %s\n", err->message);
			g_error_free (err);
		}
	/*
	* https://spec.commonmark.org/0.30/#example-341
	* Code span backticks have higher precedence than any other inline
	* constructs except HTML tags and autolinks. */
	}

	return *regex;
}

/**
 * egg_markdown_regex_link_dest_ref:
 * Return: GRegex* matcher for internal link_dest_refs
 **/
static GRegex *
egg_markdown_regex_link_dest_ref (EggMarkdown *self)
{
	static GRegex **regex = NULL;

	/* link_dest_refs created by egg_markdown_make_link_dest_ref */

	if (regex == NULL)
		regex = &g_array_index(self->priv->regex_table, GRegex *, EGG_MARKDOWN_REGEX_LINK_DEST_REF);
	if (*regex == NULL)
		*regex = g_regex_new (sUNIPUA_LINK"(\\d+)Lu;"sUNIPUA_LINK, 0, 0, NULL);

	return *regex;
}

/**
 * egg_markdown_regex_link_text_ref:
 * Return: GRegex* matcher for internal link_text_refs
 **/
static GRegex *
egg_markdown_regex_link_text_ref (EggMarkdown *self)
{
	static GRegex **regex = NULL;

	/* link_text_refs created by egg_markdown_make_link_text_ref */

	if (regex == NULL)
		regex = &g_array_index(self->priv->regex_table, GRegex *, EGG_MARKDOWN_REGEX_LINK_TEXT_REF);
	if (*regex == NULL)
		*regex = g_regex_new (sUNIPUA_LINK"(\\d+)Lx;"sUNIPUA_LINK, 0, 0, NULL);

	return *regex;
}

/**
 * egg_markdown_regex_link_title_ref:
 * Return: GRegex* matcher for internal link_title_refs
 **/
static GRegex *
egg_markdown_regex_link_title_ref (EggMarkdown *self)
{
	static GRegex **regex = NULL;

	/* link_title_refs created by egg_markdown_make_link_title_ref */

	if (regex == NULL)
		regex = &g_array_index(self->priv->regex_table, GRegex *, EGG_MARKDOWN_REGEX_LINK_TITLE_REF);
	if (*regex == NULL)
		*regex = g_regex_new (sUNIPUA_LINK"(\\d+)Lt;"sUNIPUA_LINK, 0, 0, NULL);

	return *regex;
}

/**
 * egg_markdown_slice_get_links:
 * @text: the markdown text to harvest for link slices.
 * @start_position: as in `g_regex_match_full()`.
 * @match_options: `GRegexMatchFlags`.
 * @span_count: `guint` how many link spans to match, -1 for all.
 * @stash_data:  request stashing link's destination and title.
 * Return: GArray of EggMarkdownSlice values.
 * Array length is set to zero on no match.
 *
 * @text is sliced to facilitate extracting the components of link and image
 * markdown like `[text](LINK)` and `![alt text](LINK)`.  A slice value of type
 * EGG_MARKDOWN_SLICE_LINK_TYPE_URL or EGG_MARKDOWN_SLICE_LINK_TYPE_DEST
 * includes the start and end positions of the LINK component.  Joining all
 * slices together yields the original @text.
 *
 * If @stash_data is TRUE each LINK and TITLE is stashed into @self's link_table.
 **/
static GArray *
egg_markdown_slice_get_links (EggMarkdown *self, const gchar *text, const guint start_position, const GRegexMatchFlags match_options, const guint span_count, const gboolean stash_data)
{

	GRegex *regex = egg_markdown_regex_link(self);
	GMatchInfo *match_info = NULL;
	gint i, END, start = 0, end = 0, done_end = 0, n = 0;
	EggMarkdownSlice slice;
	GArray *slices;


	slices = g_array_new (FALSE, FALSE, sizeof (slice));
	g_regex_match (regex, text, 0, &match_info);
	while (g_match_info_matches (match_info) && (span_count < 0 || ++n <= span_count)) {
		END = g_regex_get_capture_count (regex);
		for (i = 0; i <= END; i++, start = end = 0) {
			(void)g_match_info_fetch_pos (match_info, i, &start, &end);
			if (start < 0)
				continue;
			slice.start = start;
			slice.len   = end - start;
			slice.data  = -1;
			slice.type  = EGG_MARKDOWN_SLICE_UNKNOWN;
			/* switch case label is the capture group number in egg_markdown_regex_link */
			switch(i) {
				case 0:
					slice.type = EGG_MARKDOWN_SLICE_BEFORE;
					slice.start = done_end;
					slice.len = start - done_end;
					done_end = slice.start + slice.len - 1;
					g_array_append_val (slices, slice);

					slice.type = EGG_MARKDOWN_SLICE_MATCH;
					slice.start = start;
					slice.len   = end - start;
					break;
				case 3:
				case 5:
					slice.type = EGG_MARKDOWN_SLICE_LINK_MID;
					break;
				case 6:
				case 7:
					slice.type = EGG_MARKDOWN_SLICE_IGNORED;
					continue;
				case 1:
					switch (text[start]) {
						case '!': slice.type = EGG_MARKDOWN_SLICE_LINK_TYPE_IMG; break;
						case '[': slice.type = EGG_MARKDOWN_SLICE_LINK_TYPE_URL; break;
					}
					break;
				case 2:
					slice.type = EGG_MARKDOWN_SLICE_LINK_TEXT;
					if (stash_data) {
						gchar *data = g_strndup (text + slice.start, slice.len);
						slice.data = egg_markdown_stash_link_text (self, data);
						g_free (data);
					}
					break;
				case 4:
					slice.type = EGG_MARKDOWN_SLICE_LINK_DEST;
					if (stash_data && slice.len > 0) {
						gchar *link = g_strndup (text + slice.start, slice.len);
						slice.data = egg_markdown_stash_link_dest (self, link);
						g_free (link);
					}
					break;
				case 8:
					slice.type = EGG_MARKDOWN_SLICE_LINK_TITLE;
					if (stash_data) {
						gchar *title = g_strndup (text + slice.start, slice.len);
						slice.data = egg_markdown_stash_link_title (self, title);
						g_free (title);
					}
					break;
				case 9:
					/* start the end segment */
					done_end = slice.start;
					/* merge into next */
					continue;
				case 10:
					slice.type = EGG_MARKDOWN_SLICE_LINK_END;
					/* merge previous */
					slice.len += slice.start - done_end;
					slice.start = done_end;
					done_end = slice.start + slice.len;
					break;
			}
			g_array_append_val (slices, slice);
		}
		g_match_info_next (match_info, NULL);
	}

	if (slices->len > 0) {
		/* capture the text after the last END */
		slice.type  = EGG_MARKDOWN_SLICE_AFTER;
		slice.start = done_end;
		slice.len   = strlen (text) - done_end;
		slice.data  = -1;
		g_array_append_val (slices, slice);
#if DEBPRTELM > 0
		_print_slices (slices, text,
			_SOgre"%s: LINK SLICES:"_SE" %s\n", __FUNCTION__, text);
#endif
	}
	g_match_info_free (match_info);

	return (slices);
}

/**
 * egg_markdown_slice_join_links
 * @slices  @text's link slices from egg_markdown_slice_get_links
 * @text    markdown text
 * Return:  joined markdown text with link_*_ref replacements
 **/
static gchar *
egg_markdown_slice_join_links (const GArray *slices, const gchar* text)
{
	g_warn_if_fail (slices);
	g_warn_if_fail (text);
	EggMarkdownSlice *slice;
	gint done_end = 0, i;
	gchar *yp = NULL;
	GString *buf = g_string_new ("");

	for (i = 0; i < slices->len; i++) {
		slice = &g_array_index (slices, EggMarkdownSlice, i);
		switch (slice->type) {
			case EGG_MARKDOWN_SLICE_LINK_DEST:
				if (slice->start < 0)
					break;
				/* all non-link segments before the link */
				g_string_append_len (buf, text + done_end, slice->start - done_end);

				/* replace link_dest_ref for DEST */
				yp = g_strndup (text + slice->start, slice->len);
				if (slice->data >= 0 && egg_markdown_get_link_dest_id (yp) < 0) { /* yp not yet a link_dest_ref */
					gchar *ref = egg_markdown_make_link_dest_ref (slice->data);
					g_string_append (buf, ref);
					g_free (ref);
				} else {
					g_string_append (buf, yp);
				}
				g_free (yp);

				done_end = slice->start + slice->len;
				break;
			case EGG_MARKDOWN_SLICE_LINK_TEXT:
				if (slice->start < 0)
					break;
				g_string_append_len (buf, text + done_end, slice->start - done_end);

				/* replace link_text_ref for TEXT */
				yp = g_strndup (text + slice->start, slice->len);
				if (slice->data >= 0 && egg_markdown_get_link_text_id (yp) < 0) { /* yp not yet a link_text_ref */
					gchar *ref = egg_markdown_make_link_text_ref (slice->data);
					g_string_append (buf, ref);
					g_free (ref);
				} else {
					g_string_append (buf, yp);
				}
				g_free (yp);

				done_end = slice->start + slice->len;
				break;
			case EGG_MARKDOWN_SLICE_LINK_TITLE:
				if (slice->start < 0)
					break;
				g_string_append_len (buf, text + done_end, slice->start - done_end);

				/* replace link_title_ref for TITLE */
				yp = g_strndup (text + slice->start, slice->len);
				if (slice->data >= 0 && egg_markdown_get_link_title_id (yp) < 0) { /* yp not yet a link_title_ref */
					gchar *ref = egg_markdown_make_link_title_ref (slice->data);
					g_string_append (buf, ref);
					g_free (ref);
				} else {
					g_string_append (buf, yp);
				}
				g_free (yp);

				done_end = slice->start + slice->len;
				break;
		}
	}
	/* non-link text tail */
	g_string_append (buf, text + done_end);

	return g_string_free (buf, FALSE);
}

/**
 * egg_markdown_make_html_ref
 * @id: id of an inline html segment stashed with `egg_markdown_stash_html`.
 * Return:
 * dynamically-allocated string html_ref, which encodes @id
**/
static inline gchar*
egg_markdown_make_html_ref (const guint id) {
	return g_strdup_printf ("%s%dH;%s", sUNIPUA_HTML, id, sUNIPUA_HTML);
}

/**
 * egg_markdown_get_html_id
 * @ref    html_ref text
 * Return: -1   if @text isn't a valid html_ref otherwise return the html id
 **/
static inline gint
egg_markdown_get_html_id (const gchar *ref) {
	gint id = -1;
	if (strncmp (ref, sUNIPUA_HTML, sizeof (sUNIPUA_HTML) - 1) == 0)
		id = atoi (ref + sizeof (sUNIPUA_HTML) - 1);

	return id;
}

/**
 * egg_markdown_stash_html
 * @html: inline html markup or `html_ref`
 * Return:
 * stashed `html_id` - use `egg_markdown_get_html` to retrieve the html markup.
 * Identical @{html}s are stashed with the same id.
 **/
static inline guint
egg_markdown_stash_html (EggMarkdown *self, const gchar *html) {
	gint id = egg_markdown_get_html_id (html);
	if (id < 0) {
		gchar *p;
		for (guint i = 0; i < self->priv->html_table->len; i++) {
			p = g_array_index (self->priv->html_table, gchar *, i);
			if (strcmp(html, p) == 0) {
				id = i;
				break;
			}
		}
		if (id < 0) {
			p = g_strdup(html);
			g_array_append_val(self->priv->html_table, p);
			id = self->priv->html_table->len - 1;
		}
	}

	return id;
}

/* for raw HTML tag literals we want to add_pending, e.g. rule <hr /> */
#define HTML_TAG_REF(self, literal_tag)  egg_markdown_make_html_ref(egg_markdown_stash_html(self, ("!" literal_tag)));

/**
 * egg_markdown_get_html
 * @id: id of html markup previously stashed with `egg_markdown_stash_html`.
 * Return: (the memory is owned by @self) html markup.
 **/
static gchar *
egg_markdown_get_html(EggMarkdown *self, const gint id)
{
	gchar *ret = NULL;

	g_return_val_if_fail(id < self->priv->html_table->len, NULL);
	return g_array_index(self->priv->html_table, gchar *, id);
}

/**
 * egg_markdown_regex_html:
 * Return: GRegex* matcher for inline html markup
 **/
static GRegex *
egg_markdown_regex_html (EggMarkdown *self)
{
	static GRegex **regex = NULL;

	/*
	https://spec.commonmark.org/0.30/#raw-html:
	Text between < and > that looks like an HTML tag is parsed as a raw HTML
	tag and is rendered in HTML without escaping. Custom tag and attribute
	names are allowed.
	https://spec.commonmark.org/0.30/#html-tag:
	An HTML tag consists of an open tag, a closing tag, an HTML comment, a
	processing instruction, a declaration, or a CDATA section.


	PCRE PATTERN FOR HTML INLINES based on commonmark 0.30 above.
	Support open tag (OT), closing tag (CT), HTML comment (HC),
	processing instruction (PI), declaration (DE), CDATA section (CD).
	*/

	/* see bib_regular_expression */
	/* https://regex101.com tester

For study purposes:
/
(?<OT><(?<OTN>[a-zA-Z][-a-zA-Z]*)
  (?<AL>(?<A>\s+(?<AN>[:_a-zA-Z][-.:_a-zA-Z0-9]*)
    (?<AVS>\s*=\s*
      (?<AV>(?|(")([^"]*)"|(')([^']*)'|()([^ \t"'=<>`]+)))
    )?)*)\s*\/?>)
|(?<CT><\/(?<CTN>[a-zA-Z][-a-zA-Z]*)\s*>)
|(?<HC><!--(?!-?>)(?:[^<-]|<(?!!--(?!>))|-(?!-!>))*?(?<!<!-)-->)
|(?<PI><\?[\s\S]*?\?>)
|(?<DE><![a-zA-Z][^>]*>)
|(?<CD><!\[CDATA\[(?!]]>).*?]]>)
/gmx
Credit: HC pattern by https://skeptric.com/html-comment-regexp/index.html

Eliminated capturing groups except #1, which is the name of
the open or closing tag. Element type varies with the second character of
the match as follows: ASCII-letter=OT, "/"=CT, "-"=HC, "?"=PI, "!"=DE/CD.
/
(?|
(?:<([a-zA-Z][-a-zA-Z]*)
  (?:(?:\s+(?:[:_a-zA-Z][-.:_a-zA-Z0-9]*)
    (?:\s*=\s*
      (?|(?:")(?:[^"]*)"|(?:')(?:[^']*)'|(?:)(?:[^ \t"'=<>`]+))
    )?)*)\s*\/?>)
|(?:<\/([a-zA-Z][-a-zA-Z]*)\s*>)
|(?:<!--(?!-?>)(?:[^<-]|<(?!!--(?!>))|-(?!-!>))*?(?<!<!-)-->)
|(?:<\?[\s\S]*?\?>)
|(?:<![a-zA-Z][^>]*>)
|(?:<!\[CDATA\[(?!]]>).*?]]>)
)
/gmx
	*/

	if (regex == NULL)
		regex = &g_array_index(self->priv->regex_table, GRegex *, EGG_MARKDOWN_REGEX_HTML);
	if (*regex == NULL) {
		GError *err = NULL;
		*regex = g_regex_new (
		"(?|"
		"(?:<([a-zA-Z][-a-zA-Z]*)"
		  "(?:(?:\\s+(?:[:_a-zA-Z][-.:_a-zA-Z0-9]*)"
		    "(?:\\s*=\\s*"
		      "(?|(?:\")(?:[^\"]*)\"|(?:')(?:[^']*)'|(?:)(?:[^ \\t\"'=<>`]+))"
		    ")?)*)\\s*\\/?>)"
		"|(?:<\\/([a-zA-Z][-a-zA-Z]*)\\s*>)"
		"|(?:<!--(?!-?>)(?:[^<-]|<(?!!--(?!>))|-(?!-!>))*?(?<!<!-)-->)"
		"|(?:<\\?[\\s\\S]*?\\?>)"
		"|(?:<![a-zA-Z][^>]*>)"
		"|(?:<!\\[CDATA\\[(?!]]>).*?]]>)"
		")"
			, 0, 0, &err);
	}

	return *regex;
}

/**
 * egg_markdown_regex_html_ref:
 * Return: GRegex* matcher for internal html_refs
 **/
static GRegex *
egg_markdown_regex_html_ref (EggMarkdown *self)
{
	static GRegex **regex = NULL;

	/* html_ref created by egg_markdown_make_html_ref */

	if (regex == NULL)
		regex = &g_array_index(self->priv->regex_table, GRegex *, EGG_MARKDOWN_REGEX_HTML_REF);
	if (*regex == NULL)
		*regex = g_regex_new (sUNIPUA_HTML"(\\d+)H;"sUNIPUA_HTML, 0, 0, NULL);

	return *regex;
}

/**
 * egg_markdown_slice_get_htmls:
 * @text: the markdown text to harvest for inline html markup.
 * @start_position: as in `g_regex_match_full()`.
 * @match_options: `GRegexMatchFlags`.
 * @span_count: `guint` how many html segments to match, -1 for all.
 * @stash_data:  request stashing html segment.
 * Return:
 * `GArray` of @EggMarkdownSlice elements.
 * Array length is set to zero on no match.
 *
 * @text is sliced to facilitate extracting html markup inlines
 * e.g. `<tag>text</tag>`.  A slice value of type EGG_MARKDOWN_SLICE_MATCH
 * includes the start and end positions of the HTML inline.  Joining all
 * slices together yields the original @text.
 *
 * If @stash_data is TRUE each HTML inline is stashed into @self's html_table.
 **/
static GArray *
egg_markdown_slice_get_htmls (EggMarkdown *self, const gchar *text, const guint start_position, const GRegexMatchFlags match_options, const guint span_count, const gboolean stash_data)
{
	GRegex *regex = egg_markdown_regex_html (self);
	GMatchInfo *match_info = NULL;
	gint i, END, start = 0, end = 0, done_end = 0, n = 0;
	EggMarkdownSlice slice;
	GArray *slices;

	slices = g_array_new (FALSE, FALSE, sizeof (slice));
	g_regex_match_full (regex, text, -1, start_position, 0, &match_info, NULL);
	while (g_match_info_matches (match_info) && (span_count < 0 || ++n <= span_count)) {
		END = g_regex_get_capture_count (regex);
		for (i = 0; i <= END; i++, start = end = 0) {
			(void)g_match_info_fetch_pos (match_info, i, &start, &end);
			if (start < 0)
				continue;
			slice.start = start;
			slice.len   = end - start;
			slice.data  = -1;
			slice.type  = EGG_MARKDOWN_SLICE_UNKNOWN;
			/* switch case label is the capture group number in egg_markdown_regex_html */
			switch(i) {
				case 0:
					slice.type = EGG_MARKDOWN_SLICE_BEFORE;
					slice.start = done_end;
					slice.len = start - done_end;
					g_array_append_val (slices, slice);

					slice.type = EGG_MARKDOWN_SLICE_MATCH;
					slice.start = start;
					slice.len   = end - start;
					done_end = slice.start + slice.len;

					if (stash_data) {
						gchar *html;
						html = g_strndup (text + slice.start, slice.len);
						slice.data = egg_markdown_stash_html (self, html);
						g_free (html);
					}
					break;
				case 1:
					slice.type = EGG_MARKDOWN_SLICE_HTML_TAG_NAME;
					break;
			}
			g_array_append_val (slices, slice);
		}
		g_match_info_next (match_info, NULL);
	}

	if (slices->len > 0) {
		/* capture the text after the last END */
		slice.type  = EGG_MARKDOWN_SLICE_AFTER;
		slice.start = done_end;
		slice.len   = strlen (text) - done_end;
		slice.data  = -1;
		g_array_append_val (slices, slice);
#if DEBPRTELM > 0
		_print_slices (slices, text,
			_SOgre"%s: HTML SLICES:"_SE" \"%s\"\n", __FUNCTION__, text);
#endif
	}
	g_match_info_free (match_info);

	return (slices);
}

static gchar *egg_markdown_discover_auto_code_spans (const gchar *text);
/* helper for egg_markdown_hold_auto_code_spans */
static void
_append_autocode (GString *buf, gchar *text)
{
	gchar *ac = egg_markdown_discover_auto_code_spans (text);
	if (ac != NULL) {
#if DEBPRTELM > 0
		if (strcmp (text, ac))
			g_print ("discovered code span(s) for \"%s\": \"%s\"\n", text, ac);
#endif
		g_string_append_printf (buf, "%s", ac ? ac : text);
		g_free (ac);
	}
}

/**
 * egg_markdown_slice_join_codes
 * @slices    @text's code slices from egg_markdown_slice_get_codes
 * @text      markdown text
 * Return:  joined markdown text replacing code_ref for code span
 **/
static gchar *
egg_markdown_slice_join_codes (const GArray *slices, const gchar* text)
{
	g_warn_if_fail (slices);
	g_warn_if_fail (text);
	EggMarkdownSlice *slice;
	gint done_end = 0, i, len, match_start;
	gchar *yc = NULL, *nc = NULL;
	GString *buf = g_string_new ("");

	/*
	 * Sequence: START ( MID [TICK] CODE [TICK] )* END
	*/
	gboolean skip_to_END = FALSE;
	for (i = 0; i < slices->len; i++) {
		slice = &g_array_index (slices, EggMarkdownSlice, i);
		if (slice->start < 0)
			continue;
		switch (slice->type) {
			case EGG_MARKDOWN_SLICE_BEFORE:
				match_start = slice->start;
				break;
			case EGG_MARKDOWN_SLICE_CODE_MID:
				if (skip_to_END)
					break;

				/* the match start game */
				if (match_start >= 0) {
					len = slice->start - match_start;
					if (len > 0) {
						nc = g_strndup (text + match_start, len);
						g_string_append (buf, nc);
						g_free (nc);
						done_end = done_end + len;
					}
					match_start = -1;
				}

				len = slice->start - done_end;
				/* flush the non-code (nc) segments before MID */
				if (len > 0) {
					nc = g_strndup (text + done_end, len);
					g_string_append (buf, nc);
					g_free (nc);
				}

				/* go on until the corresponding CODE is found */
				break;

			case EGG_MARKDOWN_SLICE_CODE_CODE:
				/* replace code_ref for CODE */
				yc = g_strndup (text + slice->start, slice->len);
				if (slice->data >= 0 && egg_markdown_get_code_id (yc) < 0) { /* yc isn't a code_ref */
					gchar *ref = egg_markdown_make_code_ref (slice->data);
					g_string_append (buf, ref);
					g_free (ref);
				} else
					g_string_append (buf, yc);
				g_free (yc);

				skip_to_END = TRUE;
				break;

			case EGG_MARKDOWN_SLICE_CODE_END:
				skip_to_END = FALSE;
				done_end = slice->start + slice->len;
				break;
		}
	}

	/* the end game: text's non-code (nc) tail, if any */
	len = strlen (text);
	if (done_end < len) {
		if (slices->len > 0) {
			nc = g_strdup(text + done_end);
			g_string_append_len (buf, nc, len - done_end);
			g_free (nc);
		} else {
			g_string_assign (buf, text);
		}
	}

	return g_string_free (buf, FALSE);
}

/**
 * egg_markdown_strstr_spaces:
 **/
static gchar *
egg_markdown_strstr_spaces (const gchar *haystack, const gchar *needle)
{
	gchar *found;
	const gchar *haystack_new = haystack;

retry:
	/* don't find if surrounded by spaces or preceeded by backslash */
	found = strstr (haystack_new, needle);
	if (found == NULL)
		return NULL;

	/* start of the string, always valid */
	if (found == haystack)
		return found;

	/* end of the string, always valid */
	if (( *(found-1) == ' ' && *(found+1) == ' ') ||
	      *(found-1) == '\\' )
	{
		haystack_new = found+1;
		goto retry;
	}
	return found;
}

/**
 * egg_markdown_to_text_line_formatter:
 * Replace paired occurrences of @formatter with the pair @left and @write
 * This function is only used to replace straight quotes with smart quotes.
 **/
static gchar *
egg_markdown_to_text_line_formatter (const gchar *line, const gchar *formatter, const gchar *left, const gchar *right)
{
	guint len;
	gchar *str1;
	gchar *str2;
	gchar *start = NULL;
	gchar *middle = NULL;
	gchar *end = NULL;
	gchar *copy = NULL;
	gchar *data = NULL;
	gchar *temp;

	/* needed to know for shifts */
	len = strnlen (formatter, EGG_MARKDOWN_MAX_LINE_LENGTH);
	if (len == 0)
		goto out;

	/* find sections */
	copy = g_strdup (line);
	str1 = egg_markdown_strstr_spaces (copy, formatter);
	if (str1 != NULL) {
		*str1 = '\0';
		str2 = egg_markdown_strstr_spaces (str1+len, formatter);
		if (str2 != NULL) {
			*str2 = '\0';
			middle = str1 + len;
			start = copy;
			end = str2 + len;
		}
	}

	/* if we found, replace and keep looking for the same string */
	if (start != NULL && middle != NULL && end != NULL) {
		temp = g_strdup_printf ("%s%s%s%s%s", start, left, middle, right, end);
		/* recursive */
		data = egg_markdown_to_text_line_formatter (temp, formatter, left, right);
		g_free (temp);
	} else {
		/* not found, keep return as-is */
		data = g_strdup (line);
	}
out:
	g_free (copy);
	return data;
}

/**
 * EMPHASIS AND STRONG EMPHASYS
 **/
/*
The CM specification {em} matches several Unicode categories but we don't
support Unicode categories in general. We only support Non-breaking space,
and some sUNIPUA_* code points.  Also, according to {em}, white space is any
code point in the Zs category {2} plus (U+0009), line feed (U+000A), form feed
(U+000C), or carriage return (U+000D). However, we more simply match in set
{' ', '\t', '\n', '\f', '\r', U+00A0, sUNIPUA_BR, sUNIPUA_LE}.
{em} https://spec.commonmark.org/0.30/#emphasis-and-strong-emphasis
{2} https://www.fileformat.info/info/unicode/category/Zs/list.htm
*/
/**********
*  ASCII  *
***********/
#define Span(c,m,M)          ( (m) <= (guchar)(c) && (guchar)(c) <= (M) )
#define Ascii(c)             ( (guint)(c) <= 127 )
#define Cntrl(c)             ( (guchar)(c) <= 31 || (guchar)(c) == 127 )
#define Punct(c)             ( Span(c, 33, 47) || Span(c, 58, 64) || Span(c, 91, 96) || Span(c, 123, 126) )
#define Alpha(c)             ( Span(c, 'a', 'z') || Span(c, 'A', 'Z') )
#define Digit(c)             ( Span(c, '0', '9') )
#define Alnum(c)             ( Alpha(c) || Digit(c) )
#define Space(c)             ( (c)==' ' || (c)=='\n' || (c)=='\t' || (c)=='\r' || (c)=='\f' || (c)=='\v' )

/***********************
*  FUNCTION ARGUMENTS  *
************************/
/* o = MATCH offset in LINE       */
/* h = HAYSTACK length            */
/* f = match START (gchar*)       */
/* n = NEEDLE length              */
/*
LINE contains HAYSTACK somewhere.
NEEDLE can be either "***", "___", "**", "__", "*" or "_".
MATCH length is fixed and equals to NEEDLE length.
MATCH can be flanked on either or both sides by (part of) its NEEDLE, e.g. *"**"***.
A MATCH (either isolated or flanked) is CommonMark's DELIMITER RUN.
We validate MATCH by looking at relations between HAYSTACK and DELIMITER RUN.
Sometimes we need to look at characters before/after the delimiter run. In this
case we assume 3 is the maximum delimiter length. This is totally arbitrary and
results in our failing some pathological conformance tests.
*/

/***************
*  WHITESPACE  *
****************/
/* non-breaking space \u00A0 before match */
static inline guint
_NSpace (gchar *f, guint o) {
	return o > 1 && f[-2]=='\xC2' && f[-1]=='\xA0';
}

/* non-breaking space \u00A0 after match */
static inline guint
NSpace_ (gchar *f, guint o, guint h, guint n) {
	return o+n <= h-1 && f[n]=='\xC2' && f[n+1]=='\xA0';
}

/* sUNIPUA_BR or sUNIPUA_LE before match */
static inline guint
_PuaSpace (gchar *f, guint o) {
	return o > 2 && \
	f[-3]==sUNIPUA_BR[0] &&
	f[-2]==sUNIPUA_BR[1] &&
	( f[-1]==sUNIPUA_BR[2] || f[-1]==sUNIPUA_LE[2] );
}

/* sUNIPUA_BR or sUNIPUA_LE after match */
static inline guint
PuaSpace_ (gchar *f, guint o, guint h, guint n) {
	return o+n <= h-2 &&
		f[n]==sUNIPUA_BR[0] &&
		f[n+1]==sUNIPUA_BR[1] &&
		( f[n+2]==sUNIPUA_BR[2] || (f)[n+2]==sUNIPUA_LE[2] );
}

/* white space before delimiter run -- beginning of line counts as white space */
static inline guint
_Space (gchar *f, guint o) {
	/* expand delimiter as far left as possible */
	while (o > 0 && *(f-1) == *f)
		--f, --o;
	return o == 0 ? TRUE :
		Space (*(f-1)) ||
		_NSpace (f, o) ||
		_PuaSpace (f, o);
}

/* white space after delimiter run -- end of line counts as white space */
static inline guint
Space_ (gchar *f, guint o, guint h, guint n) {
	/* expand delimiter as far right as possible */
	while (o+n < h && *(f+n) == *f)
		++f, ++o;
	return o+n == h ? TRUE :
		Space (*(f+n)) ||
		NSpace_ (f, o, h, n) ||
		PuaSpace_ (f, o, h, n);
}

/****************
*  PUNCTUATION  *
*****************/
/* sUNIPUA_CODE before match akin to Punct because of absorbed "`" */
static inline guint
_PuaPunct (gchar *f, guint o) {
	return o > 2 &&
		f[-3]==sUNIPUA_CODE[0] &&
		f[-2]==sUNIPUA_CODE[1] &&
		f[-1]==sUNIPUA_CODE[2];
}
	/* sUNIPUA_CODE after match */
static inline guint
PuaPunct_ (gchar *f, guint o, guint h, guint n) {
	return o+n <= h-2 &&
		f[n]==sUNIPUA_CODE[0] &&
		f[n+1]==sUNIPUA_CODE[1] &&
		f[n+2]==sUNIPUA_CODE[2];
}

/*
KLUDGE: below we consider '<' and '>' as NON-punctuation because they stand
for open/closing '_' and '*' matched and replaced by {func}. Replacement
angle brackets can't clash with literal brackets in markdown input because,
if markdown includes '<' and '>', {func} replaces them with '&lt;' and '&gt;'
before sending haystack through the emphasis parser call sequence(*).
(*)In html and pango output modes; what happens to other modes? TODO
{func} is egg_markdown_to_text_line_format_sections.
*/
/* punctuation before delimiter run */
static inline guint
_Punct (gchar *f, guint o) {
	/* expand delimiter as far left as possible */
	while (o > 0 && *(f-1) == *f)
		--f, --o;
	return o == 0 ? FALSE :
		*(f-1) == *f || *(f-1) == '>' ? FALSE :
		Punct(*(f-1)) || _PuaPunct(f,o);
}

/* punctuation after delimiter run */
static inline guint
Punct_(gchar *f, guint o, guint h, guint n) {
	/* expand delimiter as far right as possible */
	while (o+n < h && *(f+n) == *f)
		++f, ++o;
	return o+n == h ? FALSE :
		*(f+n) == *f || *(f+n) == '<' ? FALSE :
		Punct(*(f+n)) || PuaPunct_(f, o, h, n);
}

/******************
*  DELIMITER RUN  *
*******************/
/* historical clarifications: https://talk.commonmark.org/t/2134 */

/* match is escaped */
#define Escaped(f,o)         ( o > 0 && f[-1]=='\\' )

/* match is preceded by escaped match[0] */
#define _EscF0(f,o)          ( o > 1 && f[-1]==f[0] && f[-2]=='\\' )

/* match is followed by escaped match[0] */
#define EscF0_(f,o,h,n)      ( o+n < h-1 && f[n]=='\\' && f[n+1]==f[0] )

/* https://spec.commonmark.org/0.30/#delimiter-run */
/* Deviations to consider:
While the CM parser looks for variable-length Run, length in {1,2},
EGG MD's matches a predetermined, fixed-length, Run, n in {1,3}.
*/
#define _Run(f,o,h,n)        ( \
	!Escaped(f,o) && !_EscF0(f,o) \
	)

#define Run_(f,o,h,n)        ( \
	!Escaped(f,o) && !EscF0_(f,o,h,n) \
	)

	/* https://spec.commonmark.org/0.30/#left-flanking-delimiter-run */
//#define DEBEMPH
#ifndef DEBEMPH // release
static gint
LRun (gchar *f, guint o, guint h, guint n) {
	return _Run(f,o,h,n) &&
		!Space_(f,o,h,n) &&
		( Punct_(f,o,h,n) ? (_Space(f,o) || _Punct(f,o)) : TRUE );
}

#else // debug
static gint
LRun (gchar *f, guint o, guint h, guint n) {
	gchar *p;
	guint b0, b1, b2, b3;
	b1 = _Run(f,o,h,n) ;
	b2 = !Space_(f,o,h,n) ;
	b3 = ( Punct_(f,o,h,n) ? (_Space(f,o) || _Punct(f,o)) : TRUE );
	b0 = b1 && b2 && b3;
	fprintf (stderr, "%d r%d s%d p%d| LRun ", b0, b1, b2, b3);
	for (p = f - o; p < f; p++) fputc (*p, stderr);
	fputc ('{', stderr);
	for (p = f; p < f + n; p++) fputc (*p, stderr);
	fprintf (stderr, "{%s|\n", f + n);
	return b0;
}
#endif

	/* https://spec.commonmark.org/0.30/#right-flanking-delimiter-run */
#ifndef DEBEMPH // release
static gint
RRun (gchar *f, gint o, gint h, gint n) {
	return Run_(f,o,h,n) &&
		!_Space(f,o) &&
		( _Punct(f,o) ? (Space_(f,o,h,n) || Punct_(f,o,h,n)) : TRUE );
}
#else // debug
static gint
RRun (gchar *f, gint o, gint h, gint n) {
	gchar *p;
	guint b0, b1, b2, b3;
	b1 = Run_(f,o,h,n) ;
	b2 = !_Space(f,o) ;
	b3 = ( _Punct(f,o) ? (Space_(f,o,h,n) || Punct_(f,o,h,n)) : TRUE );
	b0 = b1 && b2 && b3;
	fprintf (stderr, "%d r%d s%d p%d| RRun ", b0, b1, b2, b3);
	for (p = f - o; p < f; p++) fputc (*p, stderr);
	fputc ('}', stderr);
	for (p = f; p < f + n; p++) fputc (*p, stderr);
	fprintf (stderr, "}%s|\n", f + n);
	return b0;
}
#endif

/*************
*  CM RULES  *
**************/
	/* delimiter run inside a word, CM example 359 */
#ifndef DEBEMPH // release
#define INSIDE_WORD(f,o,h,n) ( \
		'_' != *f || o == 0 || o == h-n ? FALSE : \
		(Alnum(*(f-1)) || !Ascii(*(f-1))) && (Alnum(*(f+n)) || !Ascii(*(f+n))) \
	)
#else // debug
static int INSIDE_WORD(gchar *f, gint o, gint h, gint n) {
	gboolean b0 =
	'_' != f[0] || o == 0 || o == h-n ? FALSE : \
	(Alnum(f[-1]) || !Ascii(f[-1])) && (Alnum(f[n]) || !Ascii(f[n])) \
	;
	g_printerr ("%d| InWd\n", b0);
	return b0;
}
#endif

/***************
*  VALIDATION  *
****************/
#define VALID_LFLANK(f,o,h,n) ( \
	LRun((f),(o),(h),(n)) \
	)

#define VALID_RFLANK(f,o,h,n) ( \
	RRun((f),(o),(h),(n)) \
	)

/**
 * _strstr_flanked: helper for _flanked_formatter
 * Find valid left/right delimiter of markdown (strong) emphasis.
 * @haystack: markdown text.
 * @hlen: @haystack's length.
 * @base: @haystack's base index within its containing line.
 * @needle: the "delimiter run" to look for.
 * @nlen: @needle's length.
 * @lrun: pointer to matching left-flanking delimiter run.
 * Pass NULL to look for left-flanking delimiter.
 *  Pass
 *   NULL  to match @needle<word-to-take-emphasis>,
 *  !NULL  to match <word-to-take-emphasis>@needle.
 * @before: pointers to intra-word flanks, if any, found in @haystack before
 * finding the return value.
 * Return: pointer to valid flank otherwise NULL, and @before.
 **/
static gchar *
_strstr_flanked (
	const gchar  *haystack,
	const guint  hlen,
	const guint  base,
	const gchar  *needle,
	const guint  nlen,
	const gchar  *lrun,
	GQueue       *before)
{
	guint offset, valid;
	gchar *found = (gchar*)haystack - 1;

	do {
		if (found = strstr (found + 1, needle)) {
			/* expand left flank as far right as possible */
			if (lrun == NULL)
				while (*(found + nlen) == found[0])
					++found;

			offset = (ptrdiff_t)(found - haystack) + base;
			valid = lrun
				? VALID_RFLANK(found, offset, hlen, nlen)
				: VALID_LFLANK(found, offset, hlen, nlen);
			if (valid) {
				if (INSIDE_WORD(found, offset, hlen, nlen)) {
					valid = FALSE;
					g_queue_push_tail (before, found);
				}
			}
		}
	} while (found && !valid);

	return found;
}

static inline guint
__escape_flank (GString *str, guint idx, guint nlen) {
	guint added = 0;
	if (nlen == 2)
		if (idx == 0 || str->str[idx] != '\\') {
			g_string_insert_c (str, idx + 1, '\\');
			++added;
		}
	if (idx == 0 || str->str[idx - 1] != '\\') {
		g_string_insert_c (str, idx, '\\');
		++added;
	}
	return added;
}

/**
 * _flanked_formatter:
 * Replace open/closing @flank pair with @left/@write pair (count two replacements).
 * Escaped intra-word flanks (count one replacement).
 * @line: markdown with protected segments
 * @flank: fixed-length format:
 *   (strong) emphasis: "**", "__", "*", "_"; others: "---", "~".
 * @left: formatted replacement for left flanks.
 * @right: formatted replacement for right flanks.
 * @count: (pointer) return replacement count.
 *
 * This function is used for emphasis and strong emphasis {em}, and also
 * for strikethrough.  It and `_strstr_flanked` combine into a simplified,
 * non-conforming implementation of {em} for 8-bit characters with the
 * following limitations:
 * - Unicode White Space and Punctuation categories are treated as content.
 * - {em} rules 13-17 are replaced by ad-hoc methods.
 * - FIXME Some test cases output improperly-nested html/pango markup.
 * {em} https://spec.commonmark.org/0.30/#emphasis-and-strong-emphasis.
 **/
static gchar *
_flanked_formatter (
	const gchar *line,
	const gchar *flank,
	const gchar *left,
	const gchar *right,
	guint *count)
{
	*count = 0;
	const guint nlen = strnlen (flank, EGG_MARKDOWN_MAX_LINE_LENGTH);
	g_return_val_if_fail (nlen > 0, NULL);

	if (strstr (line, flank) == NULL)
		return g_strdup (line);

	/* for flanks contained inside_a__word before the current flank */
	gboolean contained;
	GQueue *before = g_queue_new();

	/* left (fl1, ix1) and right (fl2, ix2) flanks */
	gchar *q, *fl1, *fl2;
	gint ix1, ix2;

	const guint ll = strlen (left);
	const guint lr = strlen (right);
	GString *buf   = g_string_new (line);
	gchar *tail    = buf->str;
	guint tlen     = buf->len;

	while (tlen > nlen) {
		fl1 = _strstr_flanked (tail, buf->len, buf->len - tlen, flank, nlen, NULL, before);
		ix1 = (ptrdiff_t)(fl1 - buf->str); /* don't care if fl1 == NULL */

		while (q = g_queue_pop_tail (before)) {
			ix1 += __escape_flank(buf, (ptrdiff_t)(q - buf->str), nlen);
			*count += 1;
		}

		if (fl1 == NULL) {
			break; /* valid left flank not found */
		}

		tail = buf->str + ix1 + nlen;
		tlen = (ptrdiff_t)((buf->str + buf->len) - tail);

		while (tlen >= nlen) {
			fl2 = _strstr_flanked (tail, buf->len, buf->len - tlen, flank, nlen, fl1, before);
			ix2 = (ptrdiff_t)(fl2 - buf->str); /* don't care if fl2 == NULL */

			while (q = g_queue_pop_tail (before)) {
				guint add = __escape_flank(buf, (ptrdiff_t)(q - buf->str), nlen);
				ix2  += add;
				tlen += add;
				*count += 1;
			}

			if (fl2 == NULL) {
				break; /* valid right flank not found */
			}

			/*** Prevent improperly-nested tags ***/ //FIXME some test cases still fail
			/*** The method employed does not conform to CM{em} rules 13-17 ***/

			/* TRUE if another flank is found in the string between fl1 and fl2 */
			contained = FALSE;
			if (*flank == '_' || *flank == '*') {
				gboolean b; guint n;
				for (q = tail, b = FALSE, n = nlen; !contained && q < fl2; q++) {
					if (*q == '\\') {
						b = !b;
						n = nlen;
						continue;
					}
					else if (*q == *flank) {
						if (--n == 0 && !b) {
							contained = TRUE;
							break;
						}
					}
					b = FALSE;
				}
			}
			if (contained) {
				/* reject <fl1,fl2>, instead recurse on tail */
				guint cnt;
				gchar *temp = _flanked_formatter (tail, flank, left, right, &cnt);
				g_string_truncate (buf, (ptrdiff_t)(tail - buf->str));
				g_string_append (buf, temp);
				g_free (temp);
				*count += cnt;
				goto out;
			}
			else {
				/* accept <fl1,fl2>, then keep looking for the next paired flank */
				g_string_erase  (buf, ix2, nlen);
				g_string_insert (buf, ix2, right);
				g_string_erase  (buf, ix1, nlen);
				g_string_insert (buf, ix1, left);
				*count += 2;
				ix1 = ix2 + ll - nlen + lr;
				tlen = buf->len - ix1;
				tail = buf->str + ix1;
				break;
			}
		}
	}

out:
	g_queue_clear (before);
	g_queue_free (before);
	gchar *ret = buf->str;
	g_string_free (buf, FALSE);
	return ret;
}

#undef Span
#undef Ascii
#undef Cntrl
#undef Punct
#undef Alpha
#undef Digit
#undef Alnum
#undef Space
#undef Escaped
#undef _EscF0
#undef EscF0_
#undef _Run
#undef Run_
#undef INSIDE_WORD
#undef VALID_LFLANK
#undef VALID_RFLANK

/**
 * _resolve_ref:
 * Helper for _formatter_link_cb.
 * Resolve a link_*_ref, if any, fallback on NULL
 * @text `gchar *` link part (destination, text, title) or part's link_*_ref.
 * @get_regex `GRegex *` function returning the regex that can match a reference in @text.
 * @get_ref `gchar *` function returning the value of a matched reference from its id.
 * @ptrid `gint *` to store the id of the matched reference, nullable.
 * Return:
 * a newly-allocated string containing the resolved value otherwise NULL if
 * *@get_regex did not match or there is no value for the matched reference.
 * If @prtid is not NULL return the resolved reference id.
 **/
static gchar *
_resolve_ref (const EggMarkdown *self, const gchar *text, GRegex *get_regex(), gchar *get_ref(), gint *ptrid)
{
	GMatchInfo *info = NULL;
	gchar *ret = NULL, *match = NULL;
	gint start, id = -1;
	GRegex *regex = get_regex (self);

	if (g_regex_match (regex, text, 0, &info)) {
		if (g_match_info_fetch_pos (info, 1, &start, NULL) && start >= 0) {
			id = atoi (text + start);
			ret = get_ref (self, id);
		}
	}
	g_match_info_free (info);
	if (ptrid)
		*ptrid = id;
	return ret;
}

static gchar *egg_markdown_to_text_line_format_sections (EggMarkdown *self, const gchar *text);
/**
 * _formatter_link_cb:
 * Callback to resolve a link_*_ref, if any, and format
 * a link or an image (based on its TYPE match) for the current output surface.
 * This is used for markdown links and images.
 **/
static gboolean
_formatter_link_cb (const GMatchInfo *match_info, GString *buf, gpointer data)
{
	EggMarkdown *self = data;
	static gint TYPE = -2, TEXT = -2, DEST = -2, TITLE = -2;
	gchar *type = NULL, *text = NULL, *dest = NULL, *title = NULL;
	gchar *formatted = NULL, *p;
	GRegex *regex;
	gint id = -1, link_dest_id = -1, start = -1;

	/* get constant group number corresponding to named group capture */
	if (TYPE == -2) {
		regex = egg_markdown_regex_link (self);
		TYPE  = g_regex_get_string_number (regex, "TYPE");
		TEXT  = g_regex_get_string_number (regex, "TEXT");
		DEST  = g_regex_get_string_number (regex, "DEST");
		TITLE = g_regex_get_string_number (regex, "TITLE");
	}

	/* get matched captures, NULL if not captured */
	if (g_match_info_fetch_pos (match_info, TYPE, &start, NULL) && start >= 0)
		type  = g_match_info_fetch (match_info, TYPE);
	if (g_match_info_fetch_pos (match_info, TEXT, &start, NULL) && start >= 0)
		text  = g_match_info_fetch (match_info, TEXT);
	if (g_match_info_fetch_pos (match_info, DEST, &start, NULL) && start >= 0)
		dest  = g_match_info_fetch (match_info, DEST);
	if (g_match_info_fetch_pos (match_info, TITLE, &start, NULL) && start >= 0)
		title = g_match_info_fetch (match_info, TITLE);

	/* resolve link_dest_ref, if any, falling back to dest as is, NULL included */
	if (dest) {
		dest = _resolve_ref (self, p = dest, egg_markdown_regex_link_dest_ref,
			egg_markdown_get_link_dest, &id);
		link_dest_id = id;
		g_free (p);
	}

	/* resolve link_text_ref, if any, falling back to text as is, NULL included */
	if (text) {
		text = _resolve_ref (self, p = text, egg_markdown_regex_link_text_ref,
			egg_markdown_get_link_text, &id);
			g_free (p);
		/* format emphasis and smart text */
		if (text && type && *type == '[') { /* link */
			text = egg_markdown_to_text_line_format_sections (self, p = text);
			g_free (p);
		}
	}

	/* resolve link_title_ref, if any, falling back to title as is, NULL included */
	if (title) {
		title = _resolve_ref (self, p = title, egg_markdown_regex_link_title_ref,
			egg_markdown_get_link_title, &id);
			g_free (p);
	}

	/* finally format the replacement based on TYPE */
	if (type && *type == '[') { /* link */
		formatted = self->priv->tags.link_builder(text, dest, title, link_dest_id);
		g_string_append (buf, formatted);
	}
	else if (type && *type == '!') { /* image */
		formatted = self->priv->tags.image_builder(text, dest, title, link_dest_id);
		g_string_append (buf, formatted);
	}
	else
		/* shouldn't happen */
		g_printerr ("%s: internal error\n", __FUNCTION__);

	g_free (formatted);
	g_free (type);
	g_free (text);
	g_free (dest);
	g_free (title);

	return FALSE;
}

/**
 * egg_markdown_to_text_line_formatter_link:
 * Resolve link_dest_refs and link_title_refs, if any, and format link or image
 * for the current output surface.
 **/
static gchar *
egg_markdown_to_text_line_formatter_link (EggMarkdown *self, const gchar *line)
{
	GRegex *regex;
	GError *err = NULL;

	regex = egg_markdown_regex_link (self);
	gchar *ret = g_regex_replace_eval (regex, line, -1, 0, 0, _formatter_link_cb, self, &err);
	if (err) {
		g_printerr ("%s internal error: %s\n", __FUNCTION__, err->message);
		g_error_free (err);
		ret = g_strdup (line);
	}
#if defined DEBPRTELM
	g_printerr (_SO"%s:"_SE"\t\"%s\"\n", __FUNCTION__, ret);
#endif
	return ret;
}

/**
 * _escape_for_html:
 * Entity-escape @text for HTML output.
 * Adapted from GTK+ gmarkup.c append_escaped_text().
 * @text:
 * @subset: only escape characters in the given subset.
 **/
#define SUBSET_ALL        0x1
#define SUBSET_QUOT       0x2        /* "   */
#define SUBSET_LTGT       0x4        /* < > */
#define SUBSET_AMP        0x8        /* &   */
#define APPEND_TEXT_AND_SEEK(_str, _start, _end)        \
G_STMT_START {                                          \
	if (_end > _start)                                    \
		g_string_append_len (_str, _start, _end - _start);  \
	_start = ++_end;                                      \
} G_STMT_END
static gchar *
_escape_for_html (const gchar *text, const guint subset)
{
	const gssize length = strlen (text);
	GString *str = g_string_sized_new (length);
	const gchar *p, *pending;
	const gchar *end = text + length;
	p = pending = text;
	while (p < end && pending < end) {
		guchar c = (guchar) *pending;
		if (c == '&' && subset & (SUBSET_ALL | SUBSET_AMP)) {
			APPEND_TEXT_AND_SEEK (str, p, pending);
			g_string_append (str, "&amp;");
		}
		else if (c == '<' && subset & (SUBSET_ALL | SUBSET_LTGT)) {
			APPEND_TEXT_AND_SEEK (str, p, pending);
			g_string_append (str, "&lt;");
		}
		else if (c == '>' && subset & (SUBSET_ALL | SUBSET_LTGT)) {
			APPEND_TEXT_AND_SEEK (str, p, pending);
			g_string_append (str, "&gt;");
		}
		else if (c == '"' && subset & (SUBSET_ALL | SUBSET_QUOT)) {
				APPEND_TEXT_AND_SEEK (str, p, pending);
				g_string_append (str, "&quot;");
		}
		else
			pending++;
	}
	if (pending > p)
		g_string_append_len (str, p, pending - p);
	p = str->str;
	g_string_free (str, FALSE);
	return (gchar*)p;
}
#undef APPEND_TEXT_AND_SEEK

/**
 * _formatter_code_span_cb:
 * Callback to resolve a code_ref match, if any, and format the code span for
 * the current output surface.
 **/
static gboolean
_formatter_code_span_cb (const GMatchInfo *match_info, GString *buf, gpointer data)
{
	EggMarkdown *self = data;
	gchar *match = g_match_info_fetch (match_info, 1);
	gboolean ret = TRUE;
	gint id = atoi (match);
	gchar *code = egg_markdown_get_code (self, id);
	gchar *escaped = NULL;
	if (code) {
		if (self->priv->escape)
			escaped = g_markup_escape_text (code, -1);
		else if (self->priv->output == EGG_MARKDOWN_OUTPUT_HTML)
			escaped = _escape_for_html (code, SUBSET_ALL);
		g_string_append_printf (buf, "%s%s%s", self->priv->tags.code_span_start,
				escaped ? escaped : code, self->priv->tags.code_span_end);
		ret = FALSE;
		g_free (escaped);
	}
	g_free (match);

	return ret;
}

/**
 * egg_markdown_to_text_line_formatter_code:
 * Resolve code_refs, if any, and format code span for the current output
 * surface.
 **/
static gchar *
egg_markdown_to_text_line_formatter_code (EggMarkdown *self, const gchar *line)
{
	GRegex *regex = egg_markdown_regex_code_ref (self);
	GError *err = NULL;

	gchar *ret = g_regex_replace_eval (regex, line, -1, 0, 0, _formatter_code_span_cb, self, &err);
	if (err) {
		g_printerr ("%s internal error:\t%s\n", __FUNCTION__, err->message);
		g_error_free (err);
		ret = g_strdup (line);
	}
#if defined DEBPRTELM
	g_printerr (_SO"%s:"_SE"\t\"%s\"\n", __FUNCTION__, ret);
#endif
	return ret;
}

/**
 * _formatter_html_inline_cb:
 * Callback to resolve an html_ref match, if any, and format raw html for
 * the current output surface.
 **/
static gboolean
_formatter_html_inline_cb (const GMatchInfo *match_info, GString *buf, gpointer data)
{
	/* https://spec.commonmark.org/0.30/#raw-html:
	Text between < and > that looks like an HTML tag is parsed as a raw HTML tag and
	is rendered in HTML without escaping. Custom tag and attribute names are allowed.
	*/
	EggMarkdown *self = data;
	if (self->priv->output != EGG_MARKDOWN_OUTPUT_HTML)
		return FALSE;

	gchar *match = g_match_info_fetch (match_info, 1);
	gboolean ret = TRUE;
	gint id = atoi (match);
	gchar *html = egg_markdown_get_html (self, id);
	gboolean emit_raw = self->priv->unsafe_html;
	if (html[0] == '!') { /* HTML_TAG_REF */
		emit_raw = TRUE;
		++html;
	}
	if (html) {
		g_string_append (buf, emit_raw ? html : SAFE_HTML);
		ret = FALSE;
	}
	g_free (match);

	return ret;
}

/**
 * egg_markdown_to_text_line_formatter_html:
 * Resolve html_refs, if any, and format html inlines for the current output
 * surface.
 **/
static gchar *
egg_markdown_to_text_line_formatter_html (EggMarkdown *self, const gchar *line)
{
	GRegex *regex = egg_markdown_regex_html_ref (self);
	GError *err = NULL;

	gchar *ret = g_regex_replace_eval (regex, line, -1, 0, 0, _formatter_html_inline_cb, self, &err);
	if (err) {
		g_printerr ("%s internal error:\t%s\n", __FUNCTION__, err->message);
		g_error_free (err);
		ret = g_strdup (line);
	}
#if defined DEBPRTELM
	g_printerr (_SO"%s:"_SE"\t\"%s\"\n", __FUNCTION__, ret);
#endif
	return ret;
}

static gchar *
egg_markdown_escape_c_str(const gchar *line)
{
	gchar c;
	GString *s = g_string_sized_new (strlen(line)*2);

	for (; (c = *line); line++) {
		switch(c) {
			case '"':
			case '\\':
				g_string_append_c (s,'\\');
			default:
				g_string_append_c (s,c);
		}
	}

	return g_string_free (s, FALSE);
}

/**
 * egg_markdown_to_text_line_formatter_gettext:
	* @line: input text with PUA characters
 **/
static gchar *
egg_markdown_to_text_line_formatter_gettext (EggMarkdown *self, const gchar *line)
{
	/*
	@line contains PENDING_SEPARATOR, which is a space in release builds.
	@ contains sUNIPUA_BR, which we don't want expose outside the source code.
	Therefore replace MSGSTR_BR (ASCII) for sUNIPUA_BR in --po output.
	Translators shall not translate MSGSTR_BR.
	Markdown @line shall not contain MSGSTR_BR.
	*/

	if (self->priv->pot_output)
	{
		if (!self->priv->temporary_disable_pot) {
			gchar *escaped = egg_markdown_escape_c_str(line);

			/* replace sUNIPUA_BR => MSGSTR_BR */
			gchar *msgid = _replace_f1f2 (escaped, sUNIPUA_BR, NULL, MSGSTR_BR, NULL);
			g_free (escaped);

			fprintf(self->priv->pot_output,
			"#: %s:%u\n"
			"msgid \"%s\"\nmsgstr \"\"\n\n",
			self->priv->input_filename, self->priv->line_curr_idx + 1, msgid);
			g_free (msgid);
			return g_strdup(line);
		} else {
			return g_strdup("");
		}
	} else if (self->priv->use_gettext)
	{
		if (!self->priv->temporary_disable_pot) {
			gchar *ret, *msgid, *msgstr;
			GString *br;

			msgid = _replace_f1f2 (line, sUNIPUA_BR, NULL, MSGSTR_BR, NULL);

			msgstr = g_strdup(gettext(msgid));
			g_free (msgid);

			ret = _replace_f1f2 (msgstr, MSGSTR_BR, NULL, sUNIPUA_BR, NULL);
			g_free (msgstr);
			return ret;
		} else {
			return g_strdup(line);
		}
	}
}

/**
 * egg_markdown_to_text_line_formatter_remove_backslash:
 **/
static gchar *
egg_markdown_to_text_line_formatter_remove_backslash (gchar *line)
{
	gchar c;
	gboolean quoted;
	GString *s = g_string_sized_new (strlen(line)*2);

	for (quoted=FALSE; (c = *line); line++) {
		if (quoted) {
			switch(c) {
				case '\\': case '`':
				case '*': case '_' :
				case '{': case '}':
				case '[': case ']':
				case '(': case ')':
				case '#':
				case '+': case '-':
				case '.': case '!':
					// remove backslash
					break;
				default:
					// all other - insert the backslash again
					g_string_append_c (s,'\\');
			}
			g_string_append_c (s,c);
			quoted = FALSE;
		} else {
			if (c=='\\') quoted=TRUE;
			else
				g_string_append_c (s,c);
		}
	}

	return g_string_free (s, FALSE);
}

void
egg_markdown_soft_init(EggMarkdown *self)
{
	int i;

	for (i = 0; i < self->priv->link_table->len; i++) {
		g_free(g_array_index(self->priv->link_table, gchar *, i));
	}

	g_array_free(self->priv->link_table, TRUE);
	self->priv->link_table = g_array_new(FALSE, FALSE, sizeof(gchar *));

	for (i = 0; i < self->priv->code_table->len; i++) {
		g_free(g_array_index(self->priv->code_table, gchar *, i));
	}

	g_array_free(self->priv->code_table, TRUE);
	self->priv->code_table = g_array_new(FALSE, FALSE, sizeof(gchar *));

	for (i = 0; i < self->priv->html_table->len; i++) {
		g_free(g_array_index(self->priv->html_table, gchar *, i));
	}

	g_array_free(self->priv->html_table, TRUE);
	self->priv->html_table = g_array_new(FALSE, FALSE, sizeof(gchar *));

	if (self->priv->output == EGG_MARKDOWN_OUTPUT_HTML) {
		self->priv->tags.rule = HTML_TAG_REF(self, "<hr />");
	}
}

/**
 * egg_markdown_to_text_line_format_sections:
	* @text: markdown text
.*
.* Format @text for the current output surface.
 * Call this function after egg_markdown_hold_protected.
 **/
static gchar *
egg_markdown_to_text_line_format_sections (EggMarkdown *self, const gchar *text)
{
	gchar *temp, *data;
	const gboolean escaping = self->priv->escape || self->priv->output == EGG_MARKDOWN_OUTPUT_HTML;

	if (escaping)
		data = _escape_for_html (text, SUBSET_LTGT | SUBSET_AMP);
	else
		data = g_strdup (text);

	/*
	FIXME (future):
	The current emphasis format parsing strategy has a number of shortcomings.
	Most importantly, to be able to properly parse nested formats conforming to
	{em}, we must parse variable-length formats all together in a single pass,
	instead of parsing fixed-length formats in multiple passes as we do now.
	{em} https://spec.commonmark.org/0.30/#emphasis-and-strong-emphasis
	{2} https://github.com/commonmark/commonmark-spec/blob/d5706b9553d4665ca730524323d484880440186c/spec.txt#L9697
	*/

	/* z is a kludge to improve CM conformance */
	/* covering up for our lack of a truly recursive emphasis parser */
	guint count;
	for (guint z = 2; z > 0; z--) {

		/* bold1 */
		do {
			data = _flanked_formatter (temp = data, "**", self->priv->tags.strong_start, self->priv->tags.strong_end, &count);
			g_free (temp);
		} while (count);

		/* bold2 */
		do {
			data = _flanked_formatter (temp = data, "__", self->priv->tags.strong_start, self->priv->tags.strong_end, &count);
			g_free (temp);
		} while (count);

		/* italic1 */
		data = _flanked_formatter (temp = data, "*", self->priv->tags.em_start, self->priv->tags.em_end, &count);
		g_free (temp);

		/* italic2 */
		data = _flanked_formatter (temp = data, "_", self->priv->tags.em_start, self->priv->tags.em_end, &count);
		g_free (temp);
	}

	/* strikethrough1 */
	data = _flanked_formatter (temp = data, "---", self->priv->tags.strikethrough_start, self->priv->tags.strikethrough_end, &count);
	g_free (temp);

	/* strikethrough2 cmark(-gfm) strikethrough */
	data = _flanked_formatter (temp = data, "~", self->priv->tags.strikethrough_start, self->priv->tags.strikethrough_end, &count);
	g_free (temp);

	/* smart text goes AFTER _flanked_formatter */

	/* smart quotes */
	if (self->priv->smart_text) {
		data = egg_markdown_to_text_line_formatter (temp = data, "\"", "“", "”");
		g_free (temp);

		data = egg_markdown_to_text_line_formatter (temp = data, "'", "‘", "’");
		g_free (temp);

		data = _escape_for_html (temp = data, SUBSET_QUOT);
		g_free (temp);
	}
	else if (escaping) {
		data = _escape_for_html (temp = data, SUBSET_QUOT);
		g_free (temp);
	}

	/* smart em-dash */
	if (self->priv->smart_text && strstr (data, " -- ")) {
		GString *buf = g_string_new (temp = data);
		g_free (temp);
		g_string_replace (buf, " -- ", " — ", 0);
		data = buf->str;
		g_string_free (buf, FALSE);
	}

	/* remove redundant backslash */
	data = egg_markdown_to_text_line_formatter_remove_backslash (temp = data);
	g_free (temp);

	return data;
}

/**
 * egg_markdown_add_pending:
 * @text: markdown
 *
 * Add inlines to @self's pending memory to form a paragraph{1}.
 * Inlines include sUNIPUA_* code points.
 * {1} https://spec.commonmark.org/0.30/#paragraph.
 **/
static gboolean
egg_markdown_add_pending (EggMarkdown *self, const gchar *text)
{
	if (self->priv->line_count >= self->priv->max_lines)
		return FALSE;
	/*
	According to {1} initial and final spaces (' ') and tabs must
	be removed.  Note Glib macro g_strstrip determines white space
	with g_ascii_isspace, which returns %TRUE for ' ', '\t', _and_
	'\n', '\r', '\f'. Conclusion: don't use g_strstrip, g_strchug and
	g_strchomp. Instead use our functions named without the g_ prefix.
	*/
	/*
	Note about test output compatibility: while building a paragraph,
	cmark preserves interior line endings (a.k.a soft line breaks) and by
	default outputs then as line feeds. EGG MD does not preserve input line
	endings. Instead, while accumulating the pending buffer, it replaces
	them with PENDING_SEPARATOR, which outputs as a space (' '). Thus, the
	output paragraph is a single line (possibly with hard-line breaks).
	Representing paragraphs as single lines is a pre-condition of mdview's
	gettext feature. In other words, if EGG MD imitated cmark's soft line
	breaks, mdview's gettext feature could not work.
	*/

	if (self->priv->mode == EGG_MARKDOWN_MODE_CODEBLOCK || self->priv->mode == EGG_MARKDOWN_MODE_COMMENTBLOCK)
		g_string_append (self->priv->pending, text);
	else {
		gchar *temp = _strstrip (g_strdup (text));
		g_string_append_printf (self->priv->pending, "%s%s", temp, PENDING_SEPARATOR);
		g_free (temp);
	}
	return TRUE;
}

/**
 * egg_markdown_add_pending_br:
 **/
static gboolean
egg_markdown_add_pending_br (EggMarkdown *self, const gchar *line)
{
	GString *s;
	gboolean ret;

	s = g_string_new (line);
	_strchomp (s->str); /* strip dangling spaces */
	g_string_set_size (s, strlen (s->str));

	/* erase pending separator and append hard-line-break placeholder */
	if (g_str_has_suffix (s->str, PENDING_SEPARATOR))
		g_string_truncate (s, s->len - sizeof (PENDING_SEPARATOR) + 1);
	g_string_append (s, sUNIPUA_BR);
	ret = egg_markdown_add_pending (self, s->str);

	g_string_free (s, TRUE);
	return ret;
}

/**
 * egg_markdown_add_pending_header:
 **/
static gboolean
egg_markdown_add_pending_header (EggMarkdown *self, const gchar *text)
{
	gchar *copy;
	gboolean ret;
	gchar *p;
	/*
	* https://spec.commonmark.org/0.30/#atx-heading
	*/

	/* right-strip #s iff preceded by space */
	p = copy = g_strdup (text);
	p += strlen(p) -1;
	while (p != copy && (*p == ' ' || *p == '\t')) /* optional trailing space */
		p--;
	while (p != copy && *p == '#')
		p--;
	if (*p == ' ' || *p == '\t')
		*p = '\0';

	ret = egg_markdown_add_pending (self, copy); /* strips spaces */
	g_free (copy);
	return ret;
}

/**
 * egg_markdown_count_chars_in_word:
 **/
static guint
egg_markdown_count_chars_in_word (const gchar *text, gchar find)
{
	guint i;
	guint len;
	guint count = 0;

	/* get length */
	len = strnlen (text, EGG_MARKDOWN_MAX_LINE_LENGTH);
	if (len == 0)
		goto out;

	/* find matching chars */
	for (i=0; i<len; i++) {
		if (text[i] == find)
			count++;
	}
out:
	return count;
}

/**
 * egg_markdown_word_boundary_start:
 * Find the word after the formatting prefix of input string.
 * Prefix is a possibly-empty sequence of formatting characters accepted
 * by egg_markdown_to_text_line_format_sections, e.g., '_', '*', '-'.
 * Prefix is the longest sequence of unescaped (with '\') formatting characters.
 * At most `len` characters of `s` are examined.
 * Return value: pointer to the string after prefix.
**/
static inline gchar *
egg_markdown_word_boundary_start (const gchar *str, size_t len)
{
	gboolean f;
	gint i;
	gchar *s = (gchar *) str;

	for (i = 0; i < len; i++, s++) {
		f = *s == '_' || *s == '*' || *s == '-';
		if (!f)
			break;
	}

	return s;
}

/**
 * egg_markdown_word_boundary_end:
 * Find the end of the word before the formatting suffix of input string.
 * Suffix is defined as in egg_markdown_word_boundary_start.
 * At most `len` characters of `s` are examined. Unicode PUA character
 * sUNIPUA_BR is treated as a word boundary.
 * Return value: pointer to 1-character string before suffix.
 * If suffix is empty the return value points to the last character of `s`.
**/
static inline gchar *
egg_markdown_word_boundary_end (const gchar *str, size_t len)
{
	gboolean f, b; /* previous char was backslash */
	gint i;
	gchar *e = NULL, *s = (gchar *) str;

	for (b = FALSE, i = 0; i < len; i++, s++) {
		f = *s == '_' || *s == '*' || *s == '-';
		if (f && !b) {
			if (e == NULL)
				e = s;
			continue;
		}
		if (*s == sUNIPUA_BR[0]) {
			if (e == NULL)
				e = s;
			break;
		}
		e = NULL;
		if (*s == '\\')
			b = !b;
		else
			b = FALSE;
	}

	if (e)
		return e - 1;
	return (gchar *) str + (ptrdiff_t) (len - 1);
}

/**
 * egg_markdown_to_text_line_match_pair:
 * @s subject string
 * @len string length, input/output
 * Match any pair of characters, typically "\"\"", "''", "<>", "()" at the start
 * and end of a string, ignoring punctuation before the matching end.
 * Return:
 * TRUE and the length of the contained string in @len if the pair matches.
 */
static inline gboolean
egg_markdown_to_text_line_match_pair (const gchar *s, gint *len, const gchar *pair)
{
	gboolean b = FALSE;
	gchar *e;

	if (*s == pair[0]) {
		for (e = (gchar *) s + *len - 1; e != s && *e != pair[1] && ispunct (*e); e--)
			;
		if (b = (e != s && *e == pair[1]))
			*len = (ptrdiff_t) (e - s) - 1;
	}
	return b;
}

/**
 * egg_markdown_word_is_ident:
 * @up: option uppercase only
 * Return: length of matched indentifier [[:alnum:]_]+
 * - first character not digit
 * - underscore can be escaped with backslash
**/
static inline gint
egg_markdown_word_is_ident (const gchar *s, size_t len, gboolean up)
{
	gint i;

	if (isdigit (*s))
		return 0;

	for (i = 0; i < len; i++) {
		switch (s[i]) {
			case '\\':
			case '_' :
				continue;
		}
		if (up ? isdigit (s[i]) || isupper (s[i]) : isalnum (s[i]))
			continue;
		break;
	}
	return i;
}

/**
 * egg_markdown_word_type:
 * @text: markdown text tighly wrapping a word.
 * @start: pointer to starting position in @text; also a return value, nullable.
 * @len: pointer to length of segment in bytes; also a return value, nullable.
 * Input: If *@start is >= 0 it is the starting position of the segment to be assessed.
 * Input: If *@len is >= 0 it is the length of the segment to be assessed.
 * Return value: `EggMarkdownWordType` type of the matched word.
 * Output: If @start is not NULL return the starting position of the matched word.
 * Output: If @len is not NULL return the length of the match.
 * Formatting characters (cf. egg_markdown_to_text_line_format_sections) at the
 * start and end of @text are not included in the matched word.
**/
static EggMarkdownWordType
egg_markdown_word_type (const gchar *text, gint *start, gint *length)
{
	const gint len = strlen (text);
	EggMarkdownWordType ret = EGG_MARKDOWN_WORD_UNKNOWN;
	gchar *s, *e; /* return-word *s(tart) and *e(nd) */
	gint i, cwl;  /* return-word length */

	/* find *s(tart) of return-word candidate after formatting prefix */
	const gchar *s0 = (start && *start >= 0 ? text + *start : egg_markdown_word_boundary_start (text, len));
	/* find *e(nd) of return-word candidate before formatting suffix */
	const gchar *e0 = (length && *length >= 0 ? text + *start + *length - 1 : egg_markdown_word_boundary_end (text, len));

	s   = (gchar *)s0;
	e   = (gchar *)e0;
	cwl = (ptrdiff_t) (e - s) + 1;

/* Invariant: the return-word candidate <s>...<e> will be cwl chars long */
/* range of s [s0..e]; range of e [s..e0] */
/* Test Nth character before / after current return-word start / end */
#define LFLANK(N, C) (((ptrdiff_t)(s - s0) > 0) && (*(s - N) == (C)))
#define RFLANK(N, C) (((ptrdiff_t)(e0 - e) > 0) && (*(e + N) == (C)))

	/* word is empty or too short */
	if (cwl <= 1)
		goto out;

	/***********/
	/* dequote */
	/***********/
	if (egg_markdown_to_text_line_match_pair (s, &cwl, "\"\"") && cwl ||
		egg_markdown_to_text_line_match_pair (s, &cwl, "''") && cwl) {
			s++;
			e = s + cwl - 1;
	}
	if (cwl == 0)
		goto out;

	/* URI can be non-ASCII */
	/* https://en.wikipedia.org/wiki/Internationalized_domain_name */
	{
		gchar **x, *ext[] = { "https://", "http://", "ftp://", NULL };
		guint len;
		for (x = ext; *x; x++) {
			len = strlen (*x);
			if (!strncmp (s, *x, len)) {
				ret = EGG_MARKDOWN_WORD_URI;
				goto out;
			}

			/* [(<]http://...[>)] */
			if (cwl >= len + 2 && !strncmp (s + 1, *x, len - 1) &&
				(egg_markdown_to_text_line_match_pair (s, &cwl, "<>") ||
				egg_markdown_to_text_line_match_pair (s, &cwl, "()"))) {
					s++;
					e = s + cwl - 1;
					ret = EGG_MARKDOWN_WORD_URI_FLANKED;
					goto out;
				}
		}
	}

	/*******************************/
	/* ignore trailing punctuation */
	/*******************************/
	if (e[-1] != '\\') {
		for (i = cwl; i > 0 && ispunct (*e); )
			e--, i--;
		cwl = i;
	}

	/* absolute path len >= 4 bytes */
	if (cwl >= 4 && *s == '/') {
		if (RFLANK(1, '/'))
			e++, cwl++;
		ret = EGG_MARKDOWN_WORD_ABS_PATH;
		goto out;
	}

	/******************************/
	/* ignore leading punctuation */
	/******************************/
	if (*s != '\\') {
		for (i = cwl; i > 0 && ispunct (*s); )
			s++, i--;
		cwl = i;
	}

	if (cwl <= 0) /* nothing to do */
		goto out;

	/* match file name != "" + extension */
	{
		gchar **x, *ext[] = { ".patch", ".diff", NULL };
		guint len;
		for (x = ext; *x; x++) {
			len = strlen (*x);
			if (cwl > len && !strncmp (e - len + 1, *x, len)) {
				ret = EGG_MARKDOWN_WORD_FILE_DIFF;
				goto out;
			}
		}
	}

	/*************************************************************
	*                  ASCII only from here on                   *
	*************************************************************/

	for (i = 0; i < cwl; i++)
		if (!isascii(s[i]))
			goto out;

	/* bugzillas */
	if (LFLANK(1, '#')) {
		s--, cwl++;
		ret = EGG_MARKDOWN_WORD_BUGZILLA;
		goto out;
	}

	/* function name followed by "()" no spaces */
	if (RFLANK(1, '(') && RFLANK(2, ')') &&
		egg_markdown_word_is_ident (s, cwl, FALSE) == cwl) {
		e = e + 2;
		cwl += 2;
		ret = EGG_MARKDOWN_WORD_FUNCNAME;
		goto out;
	}

	/* email addresses */
	if (*s != '@' && *s != '.' && *e != '@') {
		guint dot = 0, at = 0;
		for (i = 0; i < cwl; i++)
			switch (s[i]) {
				case '.': ++dot; break;
				case '@': ++at;  break;
			}
		if (at == 1 && dot > 0) {
			/* kludge to prevent entangling an autolink */
			if (LFLANK(1, '[') && RFLANK(1,']'))
				goto out;
			ret = EGG_MARKDOWN_WORD_EMAIL;
			goto out;
		}
	}

	/* uppercase identifier that includes at least one '_' */
	if (egg_markdown_word_is_ident (s, cwl, TRUE) == cwl && memchr (s, '_', cwl)) {
		if (LFLANK(1, '$'))
			s--, cwl++;
		ret = EGG_MARKDOWN_WORD_UIDENT;
		goto out;
	}

out:
	if (start)
		*start = (ptrdiff_t) (s - text);
	if (length)
		*length = cwl;
	return ret;
}

/**
 * egg_markdown_discover_auto_code_spans:
 * @text: markdown text to be searched for auto code spans
 * Return:
 * newly-allocated markdown text with auto code spans or NULL if no auto code
 * spans were discovered.
**/
static gchar *
egg_markdown_discover_auto_code_spans (const gchar *text)
{
	gchar *span = NULL;
	gchar **words = g_strsplit (text, " ", -1);

	for (gint j = 0; words[j] != NULL; j++) {
		gint start = -1, cwl = -1;
		if (egg_markdown_word_type (words[j], &start, &cwl) != EGG_MARKDOWN_WORD_UNKNOWN) {
			gint i, len = strlen (words[j]);
			gchar *s, *e, *c = words[j]; /* word *s(tart) and *e(nd) chars within text */
			s = c + start;
			e = s + cwl - 1;

			if (cwl > 0 && (cwl != len || strchr (s, '\\'))) {
				span = g_malloc0 ((len + 3) * sizeof *c);

			 /* Allow backlash to escape a formatting character or itself. */
			 /* Unescaped backslash in the word itself is removed. */

				/* formatting prefix */
				for (i = 0; c != s; c++)
					span[i++] = *c;
				span[i++] = '`';
				/* word - chop unescaped backslash */
				for (gboolean b = FALSE; c != e; c++)
					switch (*c) {
						case '\\':
							b = !b;
							if (b)
								continue;
						default:
							span[i++] = *c;
							b = FALSE;
					}
				/* word end */
				span[i++] = *e;
				span[i++] = '`'; c++;
				/* formatting suffix */
				for (; *c; c++)
					span[i++] = *c;
				/* span[i] = '\0'; */
			}
			else
				span = g_strdup_printf ("`%s`", words[j]);

			g_free (words[j]);
			words[j] = span;
		}
	}
	gchar *ret = g_strjoinv (" ", words);
	g_strfreev (words);
	return ret;
}

/**
 * egg_markdown_hold_code_spans:
 * @text:  markdown text with some protected segments, typically @self->pending
 *
 * Replace each matching code span in @text with a `code_ref`.
 **/
static gchar *
egg_markdown_hold_code_spans (EggMarkdown *self, const gchar *text)
{
	gchar *ret = NULL;
	GArray *slices = NULL;

#if DEBPRTELM > 1
	g_printerr (_SO"%s"_SE":\t\t\"%s\"\n", __FUNCTION__, text);
#endif
	/* identify inline code slices, and stash code text */
	slices = egg_markdown_slice_get_codes (self, text, 0, 0, -1, TRUE);
	if (slices->len > 0) {
#if DEBPRTELM > 0
		_print_priv_table (self->priv->code_table,
			_SOyel"%s: CODE TABLE:"_SE" \"%s\"\n", __FUNCTION__, text);
#endif
		/* replace each CODE segment with code_ref to its stashed value */
		ret = egg_markdown_slice_join_codes (slices, text);
	}
	else
		ret = g_strdup (text);
	g_array_free (slices, TRUE);

	return ret;
}

/**
 * egg_markdown_hold_auto_code_spans:
 * @text:  markdown text with some protected segments, typically @self->pending
 * @value: pointer to newly-allocated string holding a modified copy of @text.
 *
 * Discover auto code spans in @text and replace each with a `code_ref`.
 * Return TRUE and set *@value if auto code spans added, otherwise return FALSE and leave *@value unchanged.
 **/
static gboolean
egg_markdown_hold_auto_code_spans (EggMarkdown *self, gchar *text, gchar **value)
{
#if DEBPRTELM > 1
	g_printerr (_SO"%s"_SE":\t\t\"%s\"\n", __FUNCTION__, text);
#endif
	gboolean ret = FALSE;
	GArray *slices = NULL;
	GString *buf = g_string_new ("");

	/* discover auto code spans */
		_append_autocode (buf, text);

	/* identify discovered auto codes, if any, and stash them */
	slices = egg_markdown_slice_get_codes (self, buf->str, 0, 0, -1, TRUE);
	if (slices->len > 0) {
		/* replace with code_ref */
		*value = egg_markdown_slice_join_codes (slices, buf->str);
		ret = TRUE;
	}
	g_array_free (slices, TRUE);
	g_string_free (buf, TRUE);

#if DEBPRTELM > 0
	if (ret)
		_print_priv_table (self->priv->code_table,
			_SOyel"%s: CODE TABLE:"_SE" \"%s\"\n", __FUNCTION__, *value);
#endif

	return ret;
}

/**
 * egg_markdown_hold_protected:
 * @text: markdown text obtained from joining multiple input lines
 * Return:
 * newly-allocated string of markdown text in which some segments are
 * encoded to protect their content from mangling by character formatters.
 * For example, since the priority of markdown inline image is higher
 * than that of inline character formatting's, a markdown segment such as
 *    `![*image alt text*](image.png)`
 * must not be formatted as HTML
 *    `<src alt="<strong>alt text</strong>" src="image.png">`
 * also not as
 *    `<src alt="&lt;strong&gt;alt text&lt;/strong&gt;" src="image.png">`
 *
 * The return string encodes high priority segments as follows:
 * - includes PENDING_SEPARATOR and Unicode PUA characters.
 * - code_refs replace code spans (inline code segments of "`CODE`" markdown)
 * - code_refs replace auto code spans (e.g. discovered "/path/file")
 * - link_dest_refs replace link paths ("![...](DEST)" and "[...](DEST ...)")
 * - link_text_refs replace link texts ('![TEXT](...)' and '[TEXT](...)')
 * - link_title_refs replace link titles ('![...](... "TITLE")' and '[...](... "TITLE")')
 * - link_dest_refs replace autolinks ("<URI>")
 * - html_refs replace HTML inlines
 *
 * egg_markdown_to_text_line_formatter_code resolves code_refs
 * egg_markdown_to_text_line_formatter_link resolves link_dest_refs, link_text_ref and link_title_refs.
 **/
static gchar *
egg_markdown_hold_protected (EggMarkdown *self, const gchar *text)
{
#ifdef DEBPRTELM
	g_printerr (_SOblu"%s: INPUT TEXT"_SE"\t\t\"%s\"\n", __FUNCTION__, text);
#endif
	guint start, len, id;
	gchar *p, *w, *temp = NULL, *atext = NULL;
	GArray *slices = NULL;
	EggMarkdownSlice *slice;

	/**********************************************************************
	*                 CODE SPANS, HTML TAGS & AUTOLINKS                  *
	**********************************************************************/

	/* https://spec.commonmark.org/0.30/#example-341
	* Code span backticks have higher precedence than any other inline
	* constructs except HTML tags and autolinks. */
	/* https://spec.commonmark.org/0.30/#example-343 - 346
	* Code spans, HTML tags, and autolinks have the same precedence. */

	for (w = atext = g_strdup (text); w && *w; w++) {
		switch (*w) {
			case '`':
				/**** potential code span ****/
				{
					/* identify anchored code span, and stash the CODE text */
					slices = egg_markdown_slice_get_codes (self, atext, w - atext, G_REGEX_MATCH_ANCHORED, 1, TRUE);
					if (slices->len > 0) {
						guint match_end, before_code;
						for (guint i = 0; i < slices->len; i++) {
							slice = &g_array_index (slices, EggMarkdownSlice, i);
							if (slice->type == EGG_MARKDOWN_SLICE_BEFORE)
								before_code = slice->len;
							else if (slice->type == EGG_MARKDOWN_SLICE_MATCH)
								match_end = slice->start + slice->len;
							else if (slice->type == EGG_MARKDOWN_SLICE_CODE_CODE)
								break;
						}
						gchar *code = g_strndup (atext + slice->start, slice->len);
						if ((id = egg_markdown_stash_code (self, code)) >= 0) {
							gchar *ref = egg_markdown_make_code_ref (id);
							/* head */
							len = (ptrdiff_t)(w - atext);
							GString *buf = g_string_new_len (atext, len);
							/* before CODE */
							if (before_code > len)
								g_string_append_len (buf, w, before_code - len);
							/* code_ref */
							g_string_append (buf, ref);
							g_free (ref);
							len = buf->len;
							/* tail */
							if (*(atext + match_end))
								g_string_append (buf, atext + match_end);
							/* swap loop subject */
							g_free (atext);
							atext = buf->str;
							g_string_free (buf, FALSE);
							/* fast forward past the "`code_ref`" just inserted */
							w = atext + len - 1;
						}
						g_free (code);
					}
					g_array_free (slices, TRUE);
				}
				break;

			case '<':
				/**** potential autolink or HTML inline ****/
				p = strchr (w, '>');
				if (p == NULL)
					break; /* whut?! autolink and HTML inline want a closing bracket */

				/* try autolink first */
				start = 0; len = (ptrdiff_t)(p - w) + 1;
				EggMarkdownWordType type = egg_markdown_word_type (w, &start, &len);
				if (type == EGG_MARKDOWN_WORD_URI_FLANKED || type == EGG_MARKDOWN_WORD_EMAIL) {
					gchar *txt = g_strndup (w + start, len);
					gchar *uri = g_strconcat (type == EGG_MARKDOWN_WORD_EMAIL ?"mailto:" :"", txt, NULL);
					uri = g_uri_escape_string (temp = uri,
						(G_URI_RESERVED_CHARS_ALLOWED_IN_PATH
						G_URI_RESERVED_CHARS_GENERIC_DELIMITERS), TRUE);
					g_free (temp);
					id = egg_markdown_stash_link_dest (self, temp = uri);
					g_free (uri);
					if (id >= 0) {
						/* head */
						GString *buf = g_string_new_len (atext, (ptrdiff_t)(w - atext));
						/* [TEXT](link_dest_ref) */
						gchar *ref = egg_markdown_make_link_dest_ref (id);
						g_string_append (buf, temp = g_strdup_printf ("[%s](%s)", txt, ref));
						g_free (temp);
						g_free (ref);
						g_free (txt);
						len = buf->len;
						/* tail */
						if (*p)
							g_string_append (buf, p + 1);
						/* swap loop subject */
						g_free (atext);
						atext = buf->str;
						g_string_free (buf, FALSE);
						/* fast forward past the autolink just inserted  */
						w = atext + len - 1;
					}
				}
				else {
				/* try HTML inline */

					/* identify anchored html segment, and stash it (even if unsafe_html is disabled) */
					slices = egg_markdown_slice_get_htmls (self, atext, w - atext, G_REGEX_MATCH_ANCHORED, 1, TRUE);
					if (slices->len > 0) {
						for (guint i = len = 0; i < slices->len; i++) {
							slice = &g_array_index (slices, EggMarkdownSlice, i);
							if (slice->type == EGG_MARKDOWN_SLICE_MATCH)
								break;
						}
						gchar *html = g_strndup (atext + slice->start, slice->len);
						if ((id = egg_markdown_stash_html (self, html)) >= 0) {
							gchar *ref = egg_markdown_make_html_ref (id);
							/* head */
							GString *buf = g_string_new_len (atext, (ptrdiff_t)(w - atext));
							/* html_ref */
							g_string_append (buf, ref);
							g_free (ref);
							len = buf->len;
							/* tail */
							if (*(p = atext + slice->start + slice->len))
								g_string_append (buf, p);
							/* swap loop subject */
							g_free (atext);
							atext = buf->str;
							g_string_free (buf, FALSE);
							/* fast forward past the html_ref just inserted */
							w = atext + len - 1;
						}
						g_free (html);
					}
					g_array_free (slices, TRUE);
				}
				break;
			default:
				w += strcspn (w, "`<") - 1;
		}
	}

	/**********************************************************************
	*                           LINKS AND IMAGES                         *
	**********************************************************************/

	/* identify link positions, and stash link destinations */
	slices = egg_markdown_slice_get_links (self, atext, 0, 0, -1, TRUE);
	if (slices->len > 0) {
		/* replace link destination with ref to its stashed value */
		atext = egg_markdown_slice_join_links (slices, temp = atext);
		g_free (temp);
#if DEBPRTELM > 0
		_print_priv_table (self->priv->link_table,
			_SOyel"%s: LINK TABLE:"_SE" \"%s\"\n", __FUNCTION__, atext);
#endif
	}
	g_array_free (slices, TRUE);

	/**********************************************************************
	*                      AUTO CODE SPAN DISCOVERY                      *
	**********************************************************************/

	/* "auto code span" is our own extension and not a commonmark thing,
	* so we get to choose its precedence vs other inlines; let it be now */

	if (egg_markdown_extension_enabled (self, EGG_MARKDOWN_EXTENSION_AUTOCODE) && self->priv->autocode) {
		if (egg_markdown_hold_auto_code_spans (self, temp = atext, &atext))
			g_free (temp);
	}

	return atext;
}

/**
 * _flush_codeblock_line:
 * @line: markdown codeblock line, including codeblock start and end delimiters.
 * @line is `NULL` at the end of the codeblock.
 *
 * Format codeblock @line for the current output surface, and append the result
 * to @self's processed buffer. Note that @line doesn't ever transit through
 * @self's pending buffer.
 * See also egg_markdown_flush_pending.
 **/
static void
_flush_codeblock_line (EggMarkdown *self, const gchar *line)
{
/* Commonmark 0.30 adds a line feed before the </code> closing tag, which
* creates an empty line at the end of the block. I don't like the empty
* line. I made it a compile-time option (default -U #undef), which, when
* undefined, also adds run-time-option --cm-block-end to assist with CM
* conformance tests.
*/
#ifndef COMMONMARK_FENCED_CODEBLOCK_LINE_ENDING
#define FMT_CMCB1    "%s"
#define FMT_CMCB2    "\n%s"
#else
#define FMT_CMCB1    "%s\n"
#define FMT_CMCB2    "%s\n"
#endif
#define TEST_FMT_CMCB1 "%s\n"
#define TEST_FMT_CMCB2 "%s\n"
	static gboolean did_end = FALSE;
	gchar *copy = NULL, *temp = NULL;

	if (self->priv->line_curr_idx == self->priv->codeblock_start_line_idx) {
		/* start block */

		self->priv->codeblock_first_line_idx = self->priv->codeblock_start_line_idx;
		switch (self->priv->codeblock_mode) {
			case EGG_MARKDOWN_CODEBLOCK_TABLE:
				g_string_append (self->priv->processed, self->priv->tags.codeblock_table_start);
				break;
			case EGG_MARKDOWN_CODEBLOCK_SHEBANG:
				g_string_append (self->priv->processed, self->priv->tags.codeblock_shebang_start);
				break;
			case EGG_MARKDOWN_CODEBLOCK_STD:
				g_string_append (self->priv->processed, self->priv->tags.codeblock_start);
				break;
			case EGG_MARKDOWN_CODEBLOCK_FENCED:
				++self->priv->codeblock_first_line_idx;
				temp = self->priv->tags.fence_builder (self);
				g_string_append (self->priv->processed, temp);
				g_free (temp);
				break;
			default:
				g_printerr ("%s: internal error\n", __FUNCTION__);
				return;
		}
		did_end = FALSE;
	}
	if (line && !self->priv->eof) {
		/* first and subsequent lines */

		guint offset = 0;
		if (self->priv->codeblock_mode == EGG_MARKDOWN_CODEBLOCK_FENCED &&
			self->priv->codeblock_fence_indent > 0) {
			offset = strspn (line, " ");
			if (offset > self->priv->codeblock_fence_indent)
				offset = self->priv->codeblock_fence_indent;
		}
		else if (self->priv->codeblock_mode == EGG_MARKDOWN_CODEBLOCK_STD)
			offset = *line == '\t' ? 1 : 4;

		copy = g_strdup (line + offset);

		/* check for translation */
		/* all codeblock modes are literal except fenced ```gettext */
		if (self->priv->codeblock_mode == EGG_MARKDOWN_CODEBLOCK_FENCED &&
			self->priv->line_curr_idx >= self->priv->codeblock_first_line_idx &&
			(self->priv->pot_output || self->priv->use_gettext)) {
			copy = egg_markdown_to_text_line_formatter_gettext (self, temp = copy);
			g_free (temp);
		}

		if (self->priv->output == EGG_MARKDOWN_OUTPUT_HTML) {
			copy = _escape_for_html (temp = copy, SUBSET_ALL);
			g_free (temp);
		}
		else if (self->priv->output == EGG_MARKDOWN_OUTPUT_PANGO) {
			copy = g_markup_escape_text (temp = copy, -1);
			g_free (temp);
		}

		if (self->priv->line_curr_idx == self->priv->codeblock_first_line_idx)
				/* first line only */
				g_string_append_printf (self->priv->processed,
					self->priv->cm_block_end ? TEST_FMT_CMCB1 : FMT_CMCB1, copy);
		else if (self->priv->line_curr_idx > self->priv->codeblock_first_line_idx)
				/* subsequent lines */
				g_string_append_printf (self->priv->processed,
					self->priv->cm_block_end ? TEST_FMT_CMCB2 : FMT_CMCB2, copy);
	}

	if (!did_end && (line == NULL || self->priv->eof)) {
		/* end properly closed block or run-away block */

		switch (self->priv->codeblock_mode) {
			case EGG_MARKDOWN_CODEBLOCK_TABLE:
				g_string_append_printf (self->priv->processed, "%s\n", self->priv->tags.codeblock_table_end);
				break;
			case EGG_MARKDOWN_CODEBLOCK_SHEBANG:
				g_string_append_printf (self->priv->processed, "%s\n", self->priv->tags.codeblock_shebang_end);
				break;
			case EGG_MARKDOWN_CODEBLOCK_STD:
				g_string_append_printf (self->priv->processed, "%s\n", self->priv->tags.codeblock_end);
				break;
			case EGG_MARKDOWN_CODEBLOCK_FENCED:
				g_string_append_printf (self->priv->processed, "%s\n", self->priv->tags.codeblock_end);
				break;
			default:
				g_printerr ("%s: internal error\n", __FUNCTION__);
				return;
		}
		did_end = TRUE;
	}

	if (copy != line)
		g_free (copy);
}

/**
 * egg_markdown_flush_pending:
 * Format @self's pending for the current output surface, and append the result
 * to @self's processed.
 **/
static void
egg_markdown_flush_pending (EggMarkdown *self)
{
	gchar *copy = NULL;
	gchar *temp = NULL;

	/* no data yet */
	if (self->priv->mode == EGG_MARKDOWN_MODE_UNKNOWN)
		return;

	/* flush HTML comment block early */
	if (self->priv->output == EGG_MARKDOWN_OUTPUT_HTML &&
		self->priv->mode == EGG_MARKDOWN_MODE_COMMENTBLOCK) {
		g_string_append_printf (self->priv->processed, "%s\n", self->priv->pending->str);
		goto clear_pending;
	}

	/* Trim the final separator that the last add_pending added */
	if (self->priv->mode != EGG_MARKDOWN_MODE_CODEBLOCK &&
		g_str_has_suffix (self->priv->pending->str, PENDING_SEPARATOR))
		g_string_truncate (self->priv->pending, self->priv->pending->len - sizeof (PENDING_SEPARATOR) + 1);

	/* nothing to flush */
	if (self->priv->pending->len == 0 &&
		self->priv->mode != EGG_MARKDOWN_MODE_CODEBLOCK &&
		self->priv->mode != EGG_MARKDOWN_MODE_BULLET_FIRST)
		return;

	copy = g_strdup (self->priv->pending->str);

	/* check for translation first */
	if (self->priv->pot_output || self->priv->use_gettext) {
		copy = egg_markdown_to_text_line_formatter_gettext (self, temp = copy);
		g_free(temp);
	}

	if (self->priv->mode != EGG_MARKDOWN_MODE_CODEBLOCK) {
		guint count = 0;

		/* protect code spans, link and image segments, HTML inlines */
		copy = egg_markdown_hold_protected (self, temp = copy);
		g_free (temp);

		/* We're left with a mix of text and protected elements.      */
		/* Text must be escaped for pango and HTML output, which will */
		/* take place in egg_markdown_to_text_line_format_sections.   */

		/*
		* format inline elements
		*/

		/* last chance to match multi-line code spans, per commonmark spec */
		/* since line endings are still protected as sUNIPUA_BR.       */
			copy = egg_markdown_hold_code_spans (self, temp = copy);
			g_free (temp);

	format_styled_text_and_smart_text: /* and escape text for pango and HTML output */
		copy = egg_markdown_to_text_line_format_sections (self, temp = copy);
		g_free (temp);

	release_line_endings:
		/* replace protected hard-line breaks with output-surface line endings */
		copy = _replace_f1f2 (temp = copy, sUNIPUA_BR, PENDING_SEPARATOR, self->priv->tags.br, &count);
		g_free (temp);
		self->priv->line_count += count; /* output line count - not used */

#ifdef DEBPRTELM
		g_printerr (_SO"%s releasing:"_SE"\t\"%s\"\n", __FUNCTION__, copy);
#endif
	/* format inline protected links (including images) */
	release_links_and_images:
		copy = egg_markdown_to_text_line_formatter_link (self, temp = copy);
		g_free(temp);

	/* drop HTML inlines unless HTML output is active */
	release_html_inlines:
		copy = egg_markdown_to_text_line_formatter_html (self, temp = copy);
		g_free (temp);

	/* format inline protected code spans */
	release_code_spans:
		copy = egg_markdown_to_text_line_formatter_code (self, temp = copy);
		g_free (temp);

	/* replace spaces for line endings that emerged from releasing code spans */
	/* per commonmark spec */
	release_emerged_line_endings:
		copy = _replace_f1f2 (temp = copy, sUNIPUA_BR, NULL, "  ", NULL);
		g_free (temp);

	}

	temp = NULL;
	DEBUG ("adding '%s'", copy);

	/*
	* format block elements
	*/

switch (self->priv->mode) { //TODO indent
	case EGG_MARKDOWN_MODE_BULLET_FIRST:
		g_string_append_printf (self->priv->processed, self->priv->tags.ulist_start);
		self->priv->mode = EGG_MARKDOWN_MODE_BULLET;
		break;
	case EGG_MARKDOWN_MODE_BULLET:
		if (self->priv->output != EGG_MARKDOWN_OUTPUT_HTML || !self->priv->ulist_loose)
			g_string_append_printf (self->priv->processed, "%s%s%s\n",
				self->priv->tags.bullet_start, copy, self->priv->tags.bullet_end);
		else
			g_string_append_printf (self->priv->processed, "%s\n%s%s%s\n%s\n",
				self->priv->tags.bullet_start,
				self->priv->tags.para_start, copy, self->priv->tags.para_end,
				self->priv->tags.bullet_end);

		++self->priv->line_count;
		--self->priv->ulist_length;
		if (self->priv->ulist_length == 0) {
			g_string_append (self->priv->processed, self->priv->tags.ulist_end);
			self->priv->ulist_loose = FALSE;
		}
	break;

	case EGG_MARKDOWN_MODE_H1:
		g_string_append_printf (self->priv->processed, "%s%s%s\n", self->priv->tags.h1_start, copy, self->priv->tags.h1_end);
		break;
	case EGG_MARKDOWN_MODE_H2:
		g_string_append_printf (self->priv->processed, "%s%s%s\n", self->priv->tags.h2_start, copy, self->priv->tags.h2_end);
		break;
	case EGG_MARKDOWN_MODE_H3:
		g_string_append_printf (self->priv->processed, "%s%s%s\n", self->priv->tags.h3_start, copy, self->priv->tags.h3_end);
		break;
	case EGG_MARKDOWN_MODE_H4:
		g_string_append_printf (self->priv->processed, "%s%s%s\n", self->priv->tags.h4_start, copy, self->priv->tags.h4_end);
		break;
	case EGG_MARKDOWN_MODE_H5:
		g_string_append_printf (self->priv->processed, "%s%s%s\n", self->priv->tags.h5_start, copy, self->priv->tags.h5_end);
		break;
	case EGG_MARKDOWN_MODE_H6:
		g_string_append_printf (self->priv->processed, "%s%s%s\n", self->priv->tags.h6_start, copy, self->priv->tags.h6_end);
		break;
	case EGG_MARKDOWN_MODE_CODEBLOCK:
		_flush_codeblock_line (self, copy);
		self->priv->line_count++;
		break;
	case EGG_MARKDOWN_MODE_PARA:
		g_string_append_printf (self->priv->processed, "%s%s%s\n", self->priv->tags.para_start, copy, self->priv->tags.para_end);
		self->priv->line_count++;
		break;
	case EGG_MARKDOWN_MODE_RULE:
		g_string_append_printf (self->priv->processed, "%s\n", copy);
		self->priv->line_count++;
		break;
}
	g_free (copy);

clear_pending:
	g_string_truncate (self->priv->pending, 0);
}

/**
 * egg_markdown_enter_codeblock_mode:
**/
static inline void
egg_markdown_enter_codeblock_mode (EggMarkdown *self, EggMarkdownCodeblockMode codeblock_mode)
{
	self->priv->mode = EGG_MARKDOWN_MODE_CODEBLOCK;
	self->priv->codeblock_mode = codeblock_mode;
	if (self->priv->codeblock_syntax != EGG_MARKDOWN_CODEBLOCK_SYNTAX_GETTEXT) {
		self->priv->codeblock_prev_pot = self->priv->temporary_disable_pot;
		self->priv->temporary_disable_pot = TRUE;
	}
	self->priv->codeblock_start_line_idx = self->priv->line_curr_idx;
}

/**
 * egg_markdown_exit_codeblock_mode:
 * Return: `TRUE`, which means that we have set `self->priv->mode`.
**/
static inline gboolean
egg_markdown_exit_codeblock_mode (EggMarkdown *self)
{
	self->priv->mode = EGG_MARKDOWN_MODE_UNKNOWN;
	self->priv->codeblock_mode = EGG_MARKDOWN_CODEBLOCK_NONE;
	if (self->priv->codeblock_syntax != EGG_MARKDOWN_CODEBLOCK_SYNTAX_GETTEXT)
		self->priv->temporary_disable_pot = self->priv->codeblock_prev_pot;
	self->priv->codeblock_syntax = EGG_MARKDOWN_CODEBLOCK_SYNTAX_NONE;
	self->priv->codeblock_start_line_idx = -1;
	return TRUE;
}

/**
 * egg_markdown_is_codeblock_mode:
 **/
static inline gboolean
egg_markdown_is_codeblock_mode (EggMarkdown *self, EggMarkdownCodeblockMode codeblock_mode)
{
	return self->priv->mode == EGG_MARKDOWN_MODE_CODEBLOCK &&
		self->priv->codeblock_mode == codeblock_mode;
}

/**
 * _is_bullet_continuation:
 * @line: markdown text.
 * Fast test if @line can be the continuation of a bullet line.
 * This function should not return false positives.
 *
 * IMPORTANT: for the sake of speed, this function unpacks some of the
 * egg_markdown_to_text_line_is_* functions into inline statements. Should those
 * functions change, the changes should be reflected here.  The more robust but
 * slower implementation is to call each individual function in this function.
 **/
static inline gboolean
_is_bullet_continuation (EggMarkdown *self, gchar *line)
{
	return line[0] && !(
		/* atx header and shebang */
	line[0] == '#' ||
		/* egg_markdown extension: table codeblock */
	(line[0] == '|' && self->priv->extensions & EGG_MARKDOWN_EXTENSION_TABLE) ||
		/* setext header 1 || 2 */
	line[0] == '=' || (line[0] == '-' && line[1] == '-') ||
		/* unpacks directives (lame: ignores non-existent directives) */
	(line[0] == '%' && line[1] == '%') ||
		/* unpacks codeblock fence start/end (lame: indenting is ignored) */
	(line[0] == '`' && line[1] == '`' && line[2] == '`'  ||
	 line[0] == '~' && line[1] == '~' && line[2] == '~') ||
		/* blockcomment start */
	(line[0] == '<' && line[1] == '!' && line[2] == '-' && line[3] == '-') ||
		/* unpacks standard codeblock */
	(line[0] == '\t' || strspn (line, " ") >= 4) ||
		/* redundant for the only use we make of this function */
	egg_markdown_to_text_line_is_blank (line) ||
		/* rule */
	egg_markdown_to_text_line_is_rule (line)
	);
}

/**
 * _preflight_ulist:
 * Test if the bullet list starting at @self's `lines[self->priv->line_curr_idx]`
 * is "loose" (a CM concept) or tight, and return the number of list items.
 * Must be called with @self's  at the (first) bullet.
 * Return: the number of list items and set @self's `ulist_loose`.
 **/
static guint
_preflight_ulist (EggMarkdown *self) {
	guint i = self->priv->line_curr_idx;
	gchar *line = self->priv->lines[i];
	gboolean is_loose = FALSE;
	guint bullet_count = 0;
	g_return_val_if_fail (egg_markdown_to_text_line_is_bullet(line), 0);
	/*
	CM: A list is loose if any of its constituent list items are separated
	by blank lines, or if any of its constituent list items directly contain
	two block-level elements with a blank line between them. Otherwise a
	list is tight. (The difference in HTML output is that paragraphs in a
	loose list are wrapped in <p> tags, while in a tight list they are not.)
	*/
	do {
		++bullet_count;
		line = self->priv->lines[++i];
		while (line) {
			if(egg_markdown_to_text_line_is_bullet (line))
				++bullet_count;
			else if (_is_bullet_continuation (self, line))
			 ;
			else
				break;
			line = self->priv->lines[++i];
		}

		while (line && egg_markdown_to_text_line_is_blank (line)) {
			line = self->priv->lines[++i];
			if (line == NULL)
				goto out;
			if (egg_markdown_to_text_line_is_bullet (line))
				is_loose = TRUE;
		}
	} while (line && egg_markdown_to_text_line_is_bullet (line));

out:
	self->priv->ulist_loose = is_loose;
	return self->priv->ulist_length = bullet_count;
}

/**
 * egg_markdown_to_text_line_process:
 * @self: the EggMarkdown instance
 **/
static gboolean
egg_markdown_to_text_line_process (EggMarkdown *self)
{
	g_return_val_if_fail (self->priv->line_curr_idx <= self->priv->line_last_idx, FALSE);
	const gchar *line      = self->priv->lines[self->priv->line_curr_idx];
	const gchar *next_line = self->priv->lines[self->priv->line_curr_idx + 1]; /* can be NULL */
	gboolean ret = FALSE;

/* COMMENT BLOCK AND CODE BLOCKS {{{ */

	/* while in HTML comment block */
	if (self->priv->mode == EGG_MARKDOWN_MODE_COMMENTBLOCK) {
		/* if safe handling (default) skip lines until end of comment or eof */
		/* otherwise: */
		if (self->priv->unsafe_html && self->priv->output == EGG_MARKDOWN_OUTPUT_HTML) {
			egg_markdown_add_pending (self, line);
			egg_markdown_flush_pending (self);
		}

		if (egg_markdown_to_text_line_is_blockcomment_end(line)) {
			/*
			https://spec.commonmark.org/0.30/#html-blocks type 2
			CM: "All types of HTML blocks except type 7 may interrupt a paragraph."
			*/
			if (self->priv->mode != EGG_MARKDOWN_MODE_BULLET &&
				self->priv->mode != EGG_MARKDOWN_MODE_PARA)
				self->priv->mode = EGG_MARKDOWN_MODE_UNKNOWN;
			else
				self->priv->mode = self->priv->blockcomment_prev_mode;
		}
		ret = TRUE;
		goto out;
	}

	/* extension: shebang code block (whole file) */
	if (egg_markdown_extension_enabled (self, EGG_MARKDOWN_EXTENSION_SHEBANG)) {
		if (self->priv->autocode_shebang || // reset on (re-)parsing rendered file
			( (self->priv->line_curr_idx == 0) && // only look for shebang on first line
			  egg_markdown_to_text_line_is_shebang (line) )
			) {
			if (!self->priv->autocode_shebang)
				egg_markdown_enter_codeblock_mode (self, EGG_MARKDOWN_CODEBLOCK_SHEBANG);
			ret = self->priv->autocode_shebang = TRUE;
			self->priv->mode = EGG_MARKDOWN_MODE_CODEBLOCK;
			self->priv->codeblock_mode = EGG_MARKDOWN_CODEBLOCK_SHEBANG;

			/* render codeblock */
			egg_markdown_add_pending (self, line);
			egg_markdown_flush_pending (self);
			goto out;
		}
	}

	/* extension: pipe table as code block - one-line lookahead */
	if (egg_markdown_extension_enabled (self, EGG_MARKDOWN_EXTENSION_TABLE)) {
		if (egg_markdown_is_codeblock_mode (self, EGG_MARKDOWN_CODEBLOCK_TABLE) &&
			egg_markdown_to_text_line_is_pipe_table_end (next_line, self->priv->table_columns))
		{
		/* end table block */
			_flush_codeblock_line (self, line);
			_flush_codeblock_line (self, NULL);
			ret = egg_markdown_exit_codeblock_mode (self);
			goto out;
		}
		if (self->priv->mode != EGG_MARKDOWN_MODE_CODEBLOCK && !egg_markdown_is_codeblock_mode(self, EGG_MARKDOWN_CODEBLOCK_TABLE)) {
			if (ret = egg_markdown_to_text_line_is_pipe_table_start (
				line, next_line, &self->priv->table_columns))
			{
		/* begin table block */
				egg_markdown_flush_pending (self);
				egg_markdown_enter_codeblock_mode (self, EGG_MARKDOWN_CODEBLOCK_TABLE);
				/* don't _flush_codeblock_line (self, line); */
			}
		}
		/* flush each line while EGG_MARKDOWN_CODEBLOCK_TABLE */
		if (ret = egg_markdown_is_codeblock_mode (self, EGG_MARKDOWN_CODEBLOCK_TABLE))
		{
			_flush_codeblock_line (self, line);
			goto out;
		}
	}

	/* CM: fenced code block */
	{
		if (self->priv->codeblock_mode != EGG_MARKDOWN_CODEBLOCK_TABLE) {
			if (*line) {
				if (self->priv->codeblock_mode != EGG_MARKDOWN_CODEBLOCK_FENCED &&
					(ret = egg_markdown_to_text_line_is_codeblock_fence_start (self, line))) {
					// begin code block
					egg_markdown_flush_pending (self);
					egg_markdown_enter_codeblock_mode (self, EGG_MARKDOWN_CODEBLOCK_FENCED);
					_flush_codeblock_line (self, line);
					goto out;
				}
				else if (self->priv->codeblock_mode == EGG_MARKDOWN_CODEBLOCK_FENCED &&
					egg_markdown_to_text_line_is_codeblock_fence_end (self, line)) {
					// end code block
					_flush_codeblock_line (self, NULL);
					ret = egg_markdown_exit_codeblock_mode (self);
					goto out;
				}
			}
			/* flush each line while EGG_MARKDOWN_CODEBLOCK_FENCED */
			if (ret = egg_markdown_is_codeblock_mode (self, EGG_MARKDOWN_CODEBLOCK_FENCED))
			{
				_flush_codeblock_line (self, line);
				goto out;
			}
		}
	}

	/* indented code block */
	if (self->priv->mode != EGG_MARKDOWN_MODE_PARA &&
		(ret = egg_markdown_to_text_line_is_indented_code (self, line))) {
		if (self->priv->mode != EGG_MARKDOWN_MODE_CODEBLOCK) {
			// begin indented code block
			egg_markdown_flush_pending (self);
			egg_markdown_enter_codeblock_mode (self, EGG_MARKDOWN_CODEBLOCK_STD);
			/* don't _flush_codeblock_line (self, line); */
		}
		/* flush each line while EGG_MARKDOWN_CODEBLOCK_STD */
		_flush_codeblock_line (self, line);
		goto out;
	}
	else if (self->priv->mode == EGG_MARKDOWN_MODE_CODEBLOCK) {
		/* end indented code block */
		_flush_codeblock_line (self, NULL);
		ret = egg_markdown_exit_codeblock_mode (self);
	}

	/* HTML comment block */
	/* CM:
	"[An HTML] block begins with a line that meets a start condition [...and]
	ends with the first ?subsequent? line that meets a matching end condition
	[...]. Any HTML within an HTML block that might otherwise be recognised
	as a start condition [is ignored and] and passed through as-is.{1}
	{1} https://spec.commonmark.org/0.30/#html-blocks
	*/
	if (ret = egg_markdown_to_text_line_is_blockcomment_start(line)) {
		egg_markdown_flush_pending (self);

		/* saved mode will be restored when blockcomment ends */
		self->priv->blockcomment_prev_mode = self->priv->mode;
		self->priv->mode = EGG_MARKDOWN_MODE_COMMENTBLOCK;

		if (self->priv->output == EGG_MARKDOWN_OUTPUT_HTML) {
			if (self->priv->unsafe_html)
				egg_markdown_add_pending (self, line);
			else
				egg_markdown_add_pending (self, SAFE_HTML);
			egg_markdown_flush_pending (self);
		}

		if (egg_markdown_to_text_line_is_blockcomment_end(line)) {
			/* line is the whole comment block */
			/* https://github.com/commonmark/commonmark-spec/issues/757 */
			/* reset mode and be done with this comment block */
			self->priv->mode = self->priv->blockcomment_prev_mode;
		}

		goto out;
	}
//}}}

	/* blank line */
	ret = egg_markdown_to_text_line_is_blank (line);
	if (ret) {
		DEBUG ("blank: '%s'", line);
		/* https://spec.commonmark.org/0.30/#blank-lines */
		/*
		CM Blank lines at the beginning and end of the document are ignored.
		CM: Blank lines between block-level elements are ignored, except for
		the role they play in determining whether a list is tight or loose.

		step: see _preflight_ulist
		step: It's worth noting that blank lines between bullets don't
		split the list, nor insert line endings or start a new block.
		*/

		egg_markdown_flush_pending (self);
		self->priv->mode = EGG_MARKDOWN_MODE_BLANK;
		goto out;
	}

	/* extension: heading continuation */
	if (!egg_markdown_extension_enabled (self, EGG_MARKDOWN_EXTENSION_HEADCONT) &&
			self->priv->mode & EGG_MARKDOWN_MODE_HEADING) {
		/* flush pending heading; start new block; carry on current line */
		egg_markdown_flush_pending (self);
		self->priv->mode = EGG_MARKDOWN_MODE_UNKNOWN;
	}

	/* setext heading 1 */
	ret = egg_markdown_to_text_line_is_heading1_setext (line);
	if (ret && self->priv->mode == EGG_MARKDOWN_MODE_PARA) {
		DEBUG ("heading1_setext: '%s'", line);
		self->priv->mode = EGG_MARKDOWN_MODE_H1;
		goto out;
	}

	/* setext heading 2 */
	ret = egg_markdown_to_text_line_is_heading2_setext (line);
	if (ret && self->priv->mode == EGG_MARKDOWN_MODE_PARA) {
		DEBUG ("heading2_setext: '%s'", line);
		self->priv->mode = EGG_MARKDOWN_MODE_H2;
		goto out;
	}

	/* rule */
	ret = egg_markdown_to_text_line_is_rule (line);
	if (ret) {
		DEBUG ("rule: '%s'", line);
		egg_markdown_flush_pending (self);
		self->priv->mode = EGG_MARKDOWN_MODE_RULE;
		if (!self->priv->pot_output) /* don't create gettext msgid "-----" */
			ret = egg_markdown_add_pending (self, self->priv->tags.rule);
		goto out;
	}

	/* bullet */
	ret = egg_markdown_to_text_line_is_bullet (line);
	if (ret) {
		DEBUG ("bullet: '%s'", line);
		egg_markdown_flush_pending (self);

		if (self->priv->mode != EGG_MARKDOWN_MODE_BULLET && self->priv->ulist_length == 0) {
			self->priv->ulist_length = _preflight_ulist (self);
			/* flush ulist start ... */
			self->priv->mode = EGG_MARKDOWN_MODE_BULLET_FIRST;
			/* ...and rewind parse loop to properly press this bullet */
			self->priv->line_delta = -1;
		}
		else {
			self->priv->mode = EGG_MARKDOWN_MODE_BULLET;
			if (egg_markdown_to_text_line_has_line_break (line))
				ret = egg_markdown_add_pending_br (self, &line[2]);
			else
				ret = egg_markdown_add_pending (self, &line[2]);
		}
		goto out;
	}

	/* ATX-style heading 1 through 6 */
	{
		EggMarkdownSlice slice;
		EggMarkdownMode mode = egg_markdown_to_text_line_is_heading_atx (line, &slice);
		if (mode) {
			egg_markdown_flush_pending (self);
			self->priv->mode = mode;
			if (egg_markdown_to_text_line_has_line_break (line))
				ret = egg_markdown_add_pending_br (self, line + slice.start);
			else
				ret = egg_markdown_add_pending_header (self, line + slice.start);
			goto out;
		}
	}

	/* directives */
	ret = egg_markdown_to_text_line_is_textdomain (line);
	if (ret) {
		egg_markdown_flush_pending (self); // end line
		gchar *domain = g_strdup(&line[sizeof(TEXTDOMAIN_DIRECTIVE)-1]);
		_strstrip (domain);
		if (domain[0]) {
			g_debug ("text_domain: '%s'\n", domain);
			gchar *dir = bindtextdomain(textdomain(NULL),NULL);
			textdomain(domain);
			bindtextdomain(textdomain(NULL), dir);
		}
		g_free (domain);
		goto out;
	}

	ret = egg_markdown_to_text_line_is_nopot (line);
	if (ret) {
		egg_markdown_flush_pending (self); // end line
		gchar *tmp = g_strdup(&line[sizeof(NOPOT_DIRECTIVE)-1]);
		_strstrip (tmp);
		if (tmp[0]) {
			g_debug ("nopot: '%s'\n", tmp);
			if (g_str_has_prefix(tmp, "yes"))
				self->priv->temporary_disable_pot = TRUE;
			else
				self->priv->temporary_disable_pot = FALSE;
		}
		g_free (tmp);
		goto out;
	}


para:
	/* paragraph */
{
	DEBUG ("continue: '%s'", line);
	gboolean new_paragraph = self->priv->mode == EGG_MARKDOWN_MODE_BLANK || self->priv->mode == EGG_MARKDOWN_MODE_UNKNOWN;
	gboolean continuation  = self->priv->mode == EGG_MARKDOWN_MODE_PARA  || self->priv->mode == EGG_MARKDOWN_MODE_BULLET;

	/* hard line-break */
	ret = egg_markdown_to_text_line_has_line_break (line);
	if (ret || new_paragraph) {
		if (!continuation) {
			egg_markdown_flush_pending (self);
			if (new_paragraph)
				self->priv->mode = EGG_MARKDOWN_MODE_PARA;
		}
		if (ret) {
			ret = egg_markdown_add_pending_br (self, line);
			goto out;
		}
	}
	/* grow new/continued paragraph or continued bullet */
	ret = egg_markdown_add_pending (self, line);
}

out:
	/* if we failed to add, we don't know the mode */
	if (!ret)
		self->priv->mode = EGG_MARKDOWN_MODE_UNKNOWN;
	return ret;
}

/**
 * egg_markdown_linkbuilder_pango:
 * @text: can be NULL
 * @dest: can be NULL, forwarded as font span.
 * @title: can be NULL
 * @link_dest_id: can be -1;
 **/
static gchar *
egg_markdown_linkbuilder_pango (const gchar *text, const gchar *dest, const gchar *title, const gint link_dest_id)
{
	/*
	* Here and in egg_markdown_imagebuilder_pango(), We need to pass some
	* extra data directly to the output surface renderer, in this case the
	* code that uses GtkTextView in markdown-text-view.c.  Since extending
	* Pango markup to allow new tags is too complicated, our extra data will
	* piggyback the Pango markup "font" attribute. The renderer will parse
	* the font string and call egg_markdown_get_link_dest() et al. to
	* obtain the link DEST.
	*
	* link_table index embedded in pango font description attribute:
	*   font="@dest=<type><id>",   with
	*   <type> "Lu"(URI), "Ip"(image path); <id> int(link table index aka id)
	*
	* Is it safe to piggyback the font description attribute{1}?
	* Test <span>s with pango-view --markup -t, e.g.
	*   font="Serif Bold Italic 44 @x=1,y=3,z=4"
	* and play with "Serif", "Bold", "Italic", "44"...
	*
	* @..., if any, is called VARIATIONS and must start with "@" in tail position.
	* I tested made-up variations; they are allowed and preserve the font.
	*
	* {1} https://docs.gtk.org/Pango/type_func.FontDescription.from_string.html#description
	* {2} https://gitlab.gnome.org/GNOME/pango/-/blob/main/pango/fonts.c#L1252
	*     is the "[src]" link in {1}
	*/
	gchar *ret = NULL, *esc_title = NULL;
	gchar *style = egg_markdown_style_pango[EGG_MARKDOWN_STYLE_URL_PANGO];

	if (title)
		esc_title = g_markup_printf_escaped (" (%s)", title);
	if (dest && link_dest_id >= 0)
		ret = g_strdup_printf (
			"<span font=\"@%s%d\" %s>" "%s</span>%s",
			_tag_info[EGG_MARKDOWN_TAG_DEST_LINK_URI_ID],
			link_dest_id, style, text ? text : "⯅⯅", esc_title ? esc_title : "");

	else /* valid commonmark markdown w/o dest, e.g. [text]() - I know... */
		ret = g_strdup_printf (
			"<span %s>%s</span>%s", style, text ? text : "⯅⯅", esc_title ? esc_title : "");
	g_free (esc_title);
	return ret;
}

/**
 * egg_markdown_linkbuilder_html
 * @text: can be NULL
 * @dest: can be NULL
 * @title: can be NULL
 * @link_dest_id: ignored
 **/
static gchar *
egg_markdown_linkbuilder_html (const gchar *text, const gchar *dest, const gchar *title, const gint link_dest_id)
{
	gchar *ret, *esc_title = NULL;

	if (title)
		esc_title = g_markup_escape_text (title, -1);
	if (esc_title)
		ret = g_strdup_printf("<a href=\"%s\" title=\"%s\">%s</a>", dest ? dest : "", esc_title, text ? text : "");
	else
		ret = g_strdup_printf("<a href=\"%s\">%s</a>", dest ? dest : "", text ? text : "");
	g_free (esc_title);
	return ret;
}

/**
 * egg_markdown_linkbuilder_text
 * @text: can be NULL
 * @dest: can be NULL
 * @title: can be NULL
 * @link_dest_id: ignored
 **/
static gchar *
egg_markdown_linkbuilder_text (const gchar *text, const gchar *dest, const gchar *title, const gint link_dest_id)
{
	if (title)
		return g_strdup_printf ("%s%s%s%s%s%s(%s)", text ? text : "", text && dest ? " " : "",
				dest ? "<" : "", dest ? dest : "", dest ? ">" : "",
				text || dest ? " " : "", title);
	else
		return g_strdup_printf ("%s%s%s%s%s", text ? text : "", text && dest ? " " : "",
				dest ? "<" : "", dest ? dest : "", dest ? ">" : "");
}

/**
 * egg_markdown_linkbuilder_ansi
 * @text: can be NULL
 * @dest: can be NULL
 * @title: can be NULL
 * @link_dest_id: ignored
 **/
static gchar *
egg_markdown_linkbuilder_ansi (const gchar *text, const gchar *dest, const gchar *title, const gint link_dest_id)
{
#define ESCP "\033"
	if (title)
		return g_strdup_printf (ESCP"[1;93m%s"ESCP"[0m" "%s" ESCP"[4;32m%s%s%s"ESCP"[0m" "%s" ESCP"[37m(%s)"ESCP"[0m",
				text ? text : "", text ? " " : "",
				dest ? "<" : "", dest ? dest : "", dest ? ">" : "",
				text || dest ? " " : "", title);
	else
		return g_strdup_printf (ESCP"[1;93m%s"ESCP"[0m" "%s" ESCP"[4;32m%s%s%s"ESCP"[0m",
			text ? text : "", text ? " " : "",
			dest ? "<" : "", dest ? dest : "", dest ? ">" : "");
#undef ESCP
}

/**
 * egg_markdown_imagebuilder_pango:
 * @text: can be NULL, image alternate text
 * @dest: can be NULL, image path, forwarded as font span
 * @title: can be NULL
 * @link_dest_id: can be -1
 **/
static gchar *
egg_markdown_imagebuilder_pango (const gchar *text, const gchar *dest, const gchar *title, const gint link_dest_id)
{
	gchar *ret, *esc_title = NULL;
	gchar *style = egg_markdown_style_pango[EGG_MARKDOWN_STYLE_IMAGE_PANGO];
	char alt[16] = {0};
	/* Similar comments here as in egg_markdown_linkbuilder_pango. */

	/*
	* Text in a GtkTextBuffer can be marked with tags. A tag is an attribute that
	* can be applied to some range of text. Here we create the tag with pango markup.
	* GtkTextView tags don't exist by themselves: they exist together with text != "".
	*/

	/* Empty text is no good so, arbitrarily, fall back to the stringified link_dest_id */
	if (text == NULL)
		snprintf (alt, 16, "%d", link_dest_id);

	if (title)
		esc_title = g_markup_printf_escaped (" (%s)", title);
	ret = g_strdup_printf ("<span font=\"@%s%d\" %s>%s</span>%s",
		_tag_info[EGG_MARKDOWN_TAG_DEST_IMAGE_PATH_ID],
		link_dest_id, style, text ? text : alt, esc_title ? esc_title : "");
	g_free (esc_title);
	return ret;
}

/**
 * egg_markdown_imagebuilder_html
 * @text: can be NULL, image alternate text
 * @dest: can be NULL, image path
 * @title: can be NULL
 * @link_dest_id: ignored
 **/
static gchar *
egg_markdown_imagebuilder_html (const gchar *text, const gchar *dest, const gchar *title, const gint link_dest_id)
{
	gchar *ret, *esc_text = NULL, *esc_title = NULL;
	if (text)
		esc_text = g_markup_escape_text (text, -1);
	if (title)
		esc_title = g_markup_escape_text (title, -1);
	if (esc_title)
		ret = g_strdup_printf ("<img src=\"%s\" alt=\"%s\" title=\"%s\" />", dest ? dest : "", esc_text ? esc_text : "", esc_title);
	else
		ret = g_strdup_printf ("<img src=\"%s\" alt=\"%s\" />", dest ? dest : "", esc_text ? esc_text : "");
	g_free (esc_text);
	g_free (esc_title);
	return ret;
}

/**
 * egg_markdown_imagebuilder_text
 * @text: can be NULL, image alternate text
 * @dest: can be NULL
 * @title: can be NULL
 * @link_dest_id: ignored
 **/
static gchar *
egg_markdown_imagebuilder_text (const gchar *text, const gchar *dest, const gchar *title, const gint link_dest_id)
{
	if (title)
		return g_strdup_printf ("%s%s%s%s%s%s(%s)", text ? text : "", text ? " " : "",
				dest ? "{" : "", dest ? dest : "", dest ? "}" : "",
				text || dest ? " " : "", title);
	else
		return g_strdup_printf ("%s%s%s%s%s", text ? text : "", text ? " " : "",
				dest ? "<" : "", dest ? dest : "", dest ? ">" : "");
}

/**
 * egg_markdown_imagebuilder_ansi
 * @text: can be NULL, image alternate text
 * @dest: can be NULL
 * @title: can be NULL
 * @link_dest_id: ignored
 **/
static gchar *
egg_markdown_imagebuilder_ansi (const gchar *text, const gchar *dest, const gchar *title, const gint link_dest_id)
{
#define ESCP "\033"
	if (title)
		return g_strdup_printf (ESCP"[1;93m%s"ESCP"[0m" "%s" ESCP"[4;32m%s%s%s"ESCP"[0m" "%s" ESCP"[37m(%s)"ESCP"[0m",
				text ? text : "", text ? " " : "",
				dest ? "{" : "", dest ? dest : "", dest ? "}" : "",
				text || dest ? " " : "", title);
	else
		return g_strdup_printf (ESCP"[1;93m%s"ESCP"[0m" "%s" ESCP"[4;32m%s%s%s"ESCP"[0m",
			text ? text : "", text ? " " : "",
			dest ? "{" : "", dest ? dest : "", dest ? "}" : "");
#undef ESCP
}

/**
 * egg_markdown_set_output:
 **/
gboolean
egg_markdown_set_output (EggMarkdown *self, EggMarkdownOutput output)
{
	gboolean ret = TRUE;
	g_return_val_if_fail (EGG_IS_MARKDOWN (self), FALSE);

	/* PangoMarkup */
	if (output == EGG_MARKDOWN_OUTPUT_PANGO) {
		self->priv->tags.em_start = "<i>";
		self->priv->tags.em_end = "</i>";
		self->priv->tags.strong_start = "<b>";
		self->priv->tags.strong_end = "</b>";
		self->priv->tags.code_span_start = g_strdup_printf ("<tt><span %s>", egg_markdown_style_pango[EGG_MARKDOWN_STYLE_CODE_SPAN_PANGO]);
		self->priv->tags.code_span_end = "</span></tt>";
		self->priv->tags.codeblock_start = g_strdup_printf ("<tt><span %s>", egg_markdown_style_pango[EGG_MARKDOWN_STYLE_CODEBLOCK_PANGO]);
		self->priv->tags.codeblock_end = "</span></tt>";
		self->priv->tags.codeblock_table_start = "<tt><span>";
		self->priv->tags.codeblock_table_end = "</span></tt>";
		self->priv->tags.codeblock_shebang_start = "<tt><span>";
		self->priv->tags.codeblock_shebang_end = "</span></tt>";
		self->priv->tags.strikethrough_start = "<s>";
		self->priv->tags.strikethrough_end = "</s>";
		self->priv->tags.h1_start = g_strdup_printf ("<b><span %s>", egg_markdown_style_pango[EGG_MARKDOWN_STYLE_H1_PANGO]);
		self->priv->tags.h1_end = "</span></b>";
		self->priv->tags.h2_start = g_strdup_printf ("<b><span %s>", egg_markdown_style_pango[EGG_MARKDOWN_STYLE_H2_PANGO]);
		self->priv->tags.h2_end = "</span></b>";
		self->priv->tags.h3_start = g_strdup_printf ("<b><span %s>", egg_markdown_style_pango[EGG_MARKDOWN_STYLE_H3_PANGO]);
		self->priv->tags.h3_end = "</span></b>";
		self->priv->tags.h4_start = g_strdup_printf ("<b><span %s>", egg_markdown_style_pango[EGG_MARKDOWN_STYLE_H4_PANGO]);
		self->priv->tags.h4_end = "</span></b>";
		self->priv->tags.h5_start = g_strdup_printf ("<b><span %s>", egg_markdown_style_pango[EGG_MARKDOWN_STYLE_H5_PANGO]);
		self->priv->tags.h5_end = "</span></b>";
		self->priv->tags.h6_start = g_strdup_printf ("<b><span %s>", egg_markdown_style_pango[EGG_MARKDOWN_STYLE_H6_PANGO]);
		self->priv->tags.h6_end = "</span></b>";
		self->priv->tags.ulist_start = "";
		self->priv->tags.ulist_end = "";
		self->priv->tags.bullet_start = "  • ";
		self->priv->tags.bullet_end = "";
		self->priv->tags.rule = "⎯⎯⎯⎯⎯⎯⎯⎯⎯⎯⎯⎯⎯⎯⎯⎯⎯⎯⎯⎯⎯⎯\n";
		self->priv->tags.para_start = "";
		self->priv->tags.para_end = "";
		self->priv->tags.br = "\n";
		self->priv->tags.link_builder = egg_markdown_linkbuilder_pango;
		self->priv->tags.image_builder = egg_markdown_imagebuilder_pango;
		self->priv->tags.fence_builder = egg_markdown_fencebuilder_pango;

	/* XHTML */
	} else if (output == EGG_MARKDOWN_OUTPUT_HTML) {
		self->priv->tags.em_start = "<em>";
		self->priv->tags.em_end = "</em>";
		self->priv->tags.strong_start = "<strong>";
		self->priv->tags.strong_end = "</strong>";
		self->priv->tags.code_span_start = "<code>";
		self->priv->tags.code_span_end = "</code>";
		self->priv->tags.codeblock_start = "<pre><code>";
		self->priv->tags.codeblock_end = "</code></pre>";
		self->priv->tags.codeblock_table_start = "<pre><code>";
		self->priv->tags.codeblock_table_end = "</code></pre>";
		self->priv->tags.codeblock_shebang_start = "<pre><code>";
		self->priv->tags.codeblock_shebang_end = "</code></pre>";
		self->priv->tags.strikethrough_start = "<s>";
		self->priv->tags.strikethrough_end = "</s>";
		self->priv->tags.ulist_start = "<ul>\n"; /* needs \n here */
		self->priv->tags.ulist_end = "</ul>\n";  /* needs \n here */
		self->priv->tags.h1_start = "<h1>";
		self->priv->tags.h1_end = "</h1>";
		self->priv->tags.h2_start = "<h2>";
		self->priv->tags.h2_end = "</h2>";
		self->priv->tags.h3_start = "<h3>";
		self->priv->tags.h3_end = "</h3>";
		self->priv->tags.h4_start = "<h4>";
		self->priv->tags.h4_end = "</h4>";
		self->priv->tags.h5_start = "<h5>";
		self->priv->tags.h5_end = "</h5>";
		self->priv->tags.h6_start = "<h6>";
		self->priv->tags.h6_end = "</h6>";
		self->priv->tags.bullet_start = "<li>";
		self->priv->tags.bullet_end = "</li>";
		self->priv->tags.rule = NULL; /* set in egg_markdown_soft_init */
		self->priv->tags.para_start = "<p>";
		self->priv->tags.para_end = "</p>";
		self->priv->tags.br = "<br />\n";  /* needs \n here */
		self->priv->tags.link_builder = egg_markdown_linkbuilder_html;
		self->priv->tags.image_builder = egg_markdown_imagebuilder_html;
		self->priv->tags.fence_builder = egg_markdown_fencebuilder_html;

	/* tty - use Linux/vt100 codes */
	} else if (output == EGG_MARKDOWN_OUTPUT_TTY) {
#define ESCP "\033"
		self->priv->tags.em_start = ESCP "[4m";
		self->priv->tags.em_end = ESCP "[24m";
		self->priv->tags.strong_start = ESCP "[1m";
		self->priv->tags.strong_end = ESCP "[21m";
		self->priv->tags.code_span_start = ESCP "[36m";
		self->priv->tags.code_span_end = ESCP "[39m";
		self->priv->tags.codeblock_start = ESCP "[36m";
		self->priv->tags.codeblock_end = ESCP "[39m";
		self->priv->tags.codeblock_table_start = "";
		self->priv->tags.codeblock_table_end = "";
		self->priv->tags.codeblock_shebang_start = "";
		self->priv->tags.codeblock_shebang_end = "";
		self->priv->tags.strikethrough_start = "--";
		self->priv->tags.strikethrough_end = "--";
		self->priv->tags.h1_start = ESCP "[7;1m";
		self->priv->tags.h1_end = ESCP "[27;21m";
		self->priv->tags.h2_start = ESCP "[7m";
		self->priv->tags.h2_end = ESCP "[27m";
		self->priv->tags.h3_start = ESCP "[1m"; // same as bold
		self->priv->tags.h3_end = ESCP "[21m";
		self->priv->tags.h4_start = ESCP "[1m"; // same as bold
		self->priv->tags.h4_end = ESCP "[21m";
		self->priv->tags.h5_start = ESCP "[1m"; // same as bold
		self->priv->tags.h5_end = ESCP "[21m";
		self->priv->tags.h6_start = ESCP "[1m"; // same as bold
		self->priv->tags.h6_end = ESCP "[21m";
		self->priv->tags.ulist_start = "";
		self->priv->tags.ulist_end = "";
		self->priv->tags.bullet_start = "* ";
		self->priv->tags.bullet_end = "";
		self->priv->tags.rule = " ----- \n";
		self->priv->tags.para_start = "";
		self->priv->tags.para_end = "";
		self->priv->tags.br = "\n";
		self->priv->tags.link_builder = egg_markdown_linkbuilder_text;
		self->priv->tags.image_builder = egg_markdown_imagebuilder_text;
		self->priv->tags.fence_builder = egg_markdown_fencebuilder_text;
#undef ESCP

	/* ansi - use ANSI CSI codes imitating lowdown's */
	} else if (output == EGG_MARKDOWN_OUTPUT_ANSI) {
		/* unlike vt100, ANSI CSI reset is all-or-nothing */
#define ESCP "\033"
		self->priv->tags.em_start = ESCP "[3m";
		self->priv->tags.em_end = ESCP "[0m";
		self->priv->tags.strong_start = ESCP "[1m";
		self->priv->tags.strong_end = ESCP "[0m";
		self->priv->tags.code_span_start = ESCP "[1;94m";
		self->priv->tags.code_span_end = ESCP "[0m";
		self->priv->tags.codeblock_start = ESCP "[1;94m";
		self->priv->tags.codeblock_end = ESCP "[0m";
		self->priv->tags.codeblock_table_start = "";
		self->priv->tags.codeblock_table_end = "";
		self->priv->tags.codeblock_shebang_start = "";
		self->priv->tags.codeblock_shebang_end = "";
		self->priv->tags.strikethrough_start = "--";
		self->priv->tags.strikethrough_end = "--";
		self->priv->tags.h1_start = ESCP "[1;91m";
		self->priv->tags.h1_end = ESCP "[0m";
		self->priv->tags.h2_start = ESCP "[1;36m";
		self->priv->tags.h2_end = ESCP "[0m";
		self->priv->tags.h3_start = ESCP "[1;36m";
		self->priv->tags.h3_end = ESCP "[0m";
		self->priv->tags.h4_start = ESCP "[1;36m";
		self->priv->tags.h4_end = ESCP "[0m";
		self->priv->tags.h5_start = ESCP "[1;36m";
		self->priv->tags.h5_end = ESCP "[0m";
		self->priv->tags.h6_start = ESCP "[1;36m";
		self->priv->tags.h6_end = ESCP "[0m";
		self->priv->tags.ulist_start = "";
		self->priv->tags.ulist_end = "";
		self->priv->tags.bullet_start = ESCP "[93m*" ESCP "[0m ";
		self->priv->tags.bullet_end = "";
		self->priv->tags.rule = " " ESCP "[93m-----" ESCP "[0m \n";
		self->priv->tags.para_start = "";
		self->priv->tags.para_end = "";
		self->priv->tags.br = "\n";
		self->priv->tags.link_builder = egg_markdown_linkbuilder_ansi;
		self->priv->tags.image_builder = egg_markdown_imagebuilder_ansi;
		self->priv->tags.fence_builder = egg_markdown_fencebuilder_text;
#undef ESCP

	/* plain text */
	} else if (output == EGG_MARKDOWN_OUTPUT_TEXT) {
		self->priv->tags.em_start = "";
		self->priv->tags.em_end = "";
		self->priv->tags.strong_start = "";
		self->priv->tags.strong_end = "";
		self->priv->tags.code_span_start = "";
		self->priv->tags.code_span_end = "";
		self->priv->tags.codeblock_start = "";
		self->priv->tags.codeblock_end = "";
		self->priv->tags.codeblock_table_start = "";
		self->priv->tags.codeblock_table_end = "";
		self->priv->tags.codeblock_shebang_start = "";
		self->priv->tags.codeblock_shebang_end = "";
		self->priv->tags.strikethrough_start = "";
		self->priv->tags.strikethrough_end = "";
		self->priv->tags.h1_start = "[";
		self->priv->tags.h1_end = "]";
		self->priv->tags.h2_start = "-";
		self->priv->tags.h2_end = "-";
		self->priv->tags.h3_start = "~";
		self->priv->tags.h3_end = "~";
		self->priv->tags.h4_start = "|";
		self->priv->tags.h4_end = "|";
		self->priv->tags.h5_start = "{";
		self->priv->tags.h5_end = "}";
		self->priv->tags.h6_start = "<";
		self->priv->tags.h6_end = ">";
		self->priv->tags.ulist_start = "";
		self->priv->tags.ulist_end = "";
		self->priv->tags.bullet_start = "* ";
		self->priv->tags.bullet_end = "";
		self->priv->tags.rule = " ----- \n";
		self->priv->tags.para_start = "";
		self->priv->tags.para_end = "";
		self->priv->tags.br = "\n";
		self->priv->tags.link_builder = egg_markdown_linkbuilder_text;
		self->priv->tags.image_builder = egg_markdown_imagebuilder_text;
		self->priv->tags.fence_builder = egg_markdown_fencebuilder_text;

	/* unknown */
	} else {
		g_warning ("unknown output enum");
		ret = FALSE;
	}

	/* save if valid */
	if (ret)
		self->priv->output = output;
	return ret;
}

/**
 * egg_markdown_set_max_lines:
 **/
gboolean
egg_markdown_set_max_lines (EggMarkdown *self, gint max_lines)
{
	g_return_val_if_fail (EGG_IS_MARKDOWN (self), FALSE);
	self->priv->max_lines = max_lines;
	return TRUE;
}

/**
 * egg_markdown_set_smart_text:
 **/
gboolean
egg_markdown_set_smart_text (EggMarkdown *self, gboolean smart_text)
{
	g_return_val_if_fail (EGG_IS_MARKDOWN (self), FALSE);
	self->priv->smart_text = smart_text;
	return TRUE;
}

/**
 * egg_markdown_set_escape:
 **/
gboolean
egg_markdown_set_escape (EggMarkdown *self, gboolean escape)
{
	g_return_val_if_fail (EGG_IS_MARKDOWN (self), FALSE);
	self->priv->escape = escape;
	return TRUE;
}

/**
 * egg_markdown_set_autocode:
 **/
gboolean
egg_markdown_set_autocode (EggMarkdown *self, gboolean autocode)
{
	g_return_val_if_fail (EGG_IS_MARKDOWN (self), FALSE);
	self->priv->autocode = autocode;
	return TRUE;
}

/**
 * egg_markdown_set_pot_output:
 **/
gboolean
egg_markdown_set_pot_output (EggMarkdown *self, FILE *f, gchar *input_filename)
{
	g_return_val_if_fail (EGG_IS_MARKDOWN (self), FALSE);
	self->priv->pot_output = f;
	self->priv->input_filename = g_strdup(input_filename);
	return TRUE;
}

/**
 * egg_markdown_set_use_gettext:
 **/
gboolean
egg_markdown_set_use_gettext (EggMarkdown *self, gboolean use_gettext)
{
	g_return_val_if_fail (EGG_IS_MARKDOWN (self), FALSE);
	self->priv->use_gettext = use_gettext;
	return TRUE;
}

/**
 * egg_markdown_set_cm_block_end:
 **/
gboolean
egg_markdown_set_cm_block_end (EggMarkdown *self, gboolean cm_block_end)
{
	g_return_val_if_fail (EGG_IS_MARKDOWN (self), FALSE);
	self->priv->cm_block_end = cm_block_end;
	return TRUE;
}

/**
 * egg_markdown_set_unsafe_html:
 **/
gboolean
egg_markdown_set_unsafe_html (EggMarkdown *self, gboolean unsafe_html)
{
	g_return_val_if_fail (EGG_IS_MARKDOWN (self), FALSE);
	self->priv->unsafe_html = unsafe_html;
	return TRUE;
}

/**
 * egg_markdown_set_extensions:
 **/
gboolean
egg_markdown_set_extensions (EggMarkdown *self, const EggMarkdownExtensions ext)
{
	g_return_val_if_fail (EGG_IS_MARKDOWN (self), FALSE);
	self->priv->extensions = ext;
	return TRUE;
}

/**
 * egg_markdown_parse:
 **/
gchar *
egg_markdown_parse (EggMarkdown *self, const gchar *markdown)
{
	gchar **lines;
	guint i, max_loop;
	gchar *temp;
	gboolean ret;

	g_return_val_if_fail (EGG_IS_MARKDOWN (self), NULL);
	g_return_val_if_fail (self->priv->output != EGG_MARKDOWN_OUTPUT_UNKNOWN, NULL);

	DEBUG ("input='%s'", markdown);

	/* process */
	self->priv->mode = EGG_MARKDOWN_MODE_UNKNOWN;
	self->priv->codeblock_mode = EGG_MARKDOWN_CODEBLOCK_NONE;
	self->priv->codeblock_syntax = EGG_MARKDOWN_CODEBLOCK_SYNTAX_NONE;
	self->priv->codeblock_start_line_idx = -1;
	self->priv->line_count = 0;
	self->priv->line_curr_idx = -1;
	self->priv->line_delta = 0;
	self->priv->eof = FALSE;
	self->priv->ulist_loose = FALSE;
	self->priv->ulist_length = 0;
	g_string_truncate (self->priv->pending, 0);
	g_string_truncate (self->priv->processed, 0);
	self->priv->lines = lines = g_regex_split_simple ("\\R", markdown, 0, 0);
	self->priv->line_last_idx = g_strv_length (lines) - 1;
	self->priv->autocode_shebang = FALSE;

	/* process each line */
	i = 0;
	max_loop = self->priv->line_last_idx * 10; /* cumulative seek budget */
	do {
		gint n;
		/* handle request to rewind/advance the loop */
		self->priv->line_curr_idx = i;
		ret = egg_markdown_to_text_line_process (self);
		if (!ret)
			break;
		if (self->priv->line_delta) {
			n = i + self->priv->line_delta;
			self->priv->line_delta = 0;
			if (-1 <= n && n < (gint)self->priv->line_last_idx)
				i = n + 1;
			else
				++i;
		}
		else
			++i;
		--max_loop;
	} while (ret && i <= self->priv->line_last_idx && max_loop > 0);

	if (max_loop == 0)
		g_printerr ("internal error (max_loop)\n");

	g_strfreev (lines);
	self->priv->eof = TRUE;
	egg_markdown_flush_pending (self);

	/* remove trailing \n */
#define S self->priv->processed
	if (S->len && S->str[S->len - 1] == '\n')
		g_string_truncate (S, S->len - 1);
#undef S

	g_string_truncate (self->priv->pending, 0);
	temp = self->priv->processed->str;
	g_string_free (self->priv->processed, FALSE);
	self->priv->processed = g_string_new ("");

	DEBUG ("output='%s'", temp);

	return temp;
}

/**
 * egg_markdown_class_init:
 * @klass: The EggMarkdownClass
 **/
static void
egg_markdown_class_init (EggMarkdownClass *klass)
{
	GObjectClass *object_class = G_OBJECT_CLASS (klass);
	object_class->finalize = egg_markdown_finalize;
G_GNUC_BEGIN_IGNORE_DEPRECATIONS;
	g_type_class_add_private (klass, sizeof (EggMarkdownPrivate));
G_GNUC_END_IGNORE_DEPRECATIONS;
}

/**
 * egg_markdown_init:
 **/
static void
egg_markdown_init (EggMarkdown *self)
{
	self->priv = EGG_MARKDOWN_GET_PRIVATE (self);

	self->priv->extensions = 0xffff; // enable all extensions
	self->priv->mode = EGG_MARKDOWN_MODE_UNKNOWN;
	self->priv->blockcomment_prev_mode = EGG_MARKDOWN_MODE_UNKNOWN;
	self->priv->codeblock_mode = EGG_MARKDOWN_CODEBLOCK_NONE;
	self->priv->codeblock_syntax = EGG_MARKDOWN_CODEBLOCK_SYNTAX_NONE;
	self->priv->output = EGG_MARKDOWN_OUTPUT_UNKNOWN;
	self->priv->pending = g_string_new ("");
	self->priv->processed = g_string_new ("");
	self->priv->regex_table = g_array_set_size(g_array_sized_new(FALSE, TRUE, sizeof (GRegex *), EGG_MARKDOWN_REGEX_LEN), EGG_MARKDOWN_REGEX_LEN);
	self->priv->code_table = g_array_new(FALSE, FALSE, sizeof(gchar *));
	self->priv->link_table = g_array_new(FALSE, FALSE, sizeof(gchar *));
	self->priv->html_table = g_array_new(FALSE, FALSE, sizeof(gchar *));
	self->priv->max_lines = -1;
	self->priv->smart_text = FALSE;
	self->priv->escape = FALSE;
	self->priv->autocode = FALSE;
	self->priv->autocode_shebang = FALSE;
	self->priv->pot_output = NULL;
	self->priv->temporary_disable_pot = FALSE;
	self->priv->use_gettext = FALSE;
	self->priv->codeblock_fence_info = NULL;
#ifndef COMMONMARK_FENCED_CODEBLOCK_LINE_ENDING
	self->priv->cm_block_end = FALSE;
#else
	self->priv->cm_block_end = TRUE;
#endif
	self->priv->unsafe_html = FALSE;
	self->priv->pot_output = NULL;
	self->priv->input_filename = NULL;
}

/**
 * egg_markdown_finalize:
 * @object: The object to finalize
 **/
static void
egg_markdown_finalize (GObject *object)
{
	EggMarkdown *self;
	int i;

	g_return_if_fail (EGG_IS_MARKDOWN (object));

	self = EGG_MARKDOWN (object);

	g_return_if_fail (self->priv != NULL);
	g_string_free (self->priv->pending, TRUE);
	g_string_free (self->priv->processed, TRUE);
	g_free (self->priv->codeblock_fence_info);

	for (i = 0; i < self->priv->link_table->len; i++) {
		g_free(g_array_index(self->priv->link_table, gchar *, i));
	}
	g_array_free (self->priv->link_table, TRUE);
	for (i = 0; i < self->priv->code_table->len; i++) {
		g_free(g_array_index(self->priv->code_table, gchar *, i));
	}
	g_array_free (self->priv->code_table, TRUE);
	g_free (self->priv->input_filename);
	for (i = 0; i < self->priv->html_table->len; i++) {
		g_free(g_array_index(self->priv->html_table, gchar *, i));
	}
	g_array_free (self->priv->html_table, TRUE);
	g_free (self->priv->input_filename);

	for (i = 0; i < self->priv->regex_table->len; i++) {
		GRegex *p = g_array_index (self->priv->regex_table, GRegex *, i);
		if (p)
			g_regex_unref (p);
	}
	g_array_free (self->priv->regex_table, TRUE);

	G_OBJECT_CLASS (egg_markdown_parent_class)->finalize (object);
}

/**
 * egg_markdown_new:
 *
 * Return value: a new EggMarkdown object.
 **/
EggMarkdown *
egg_markdown_new (void)
{
	EggMarkdown *self;
	self = g_object_new (EGG_TYPE_MARKDOWN, NULL);
	return EGG_MARKDOWN (self);
}

/***************************************************************************
 ***                          MAKE CHECK TESTS                           ***
 ***************************************************************************/
#ifdef EGG_TEST
#include "egg-test.h"

void
egg_markdown_test (EggTest *test)
{
	EggMarkdown *self;
	gchar *text;
	gboolean ret;
	const gchar *markdown;
	const gchar *markdown_expected;

	if (!egg_test_start (test, "EggMarkdown"))
		return;

	/************************************************************
	 ****************        line_is_rule          **************
	 ************************************************************/
	ret = egg_markdown_to_text_line_is_rule ("* * *");
	egg_test_title_assert (test, "is rule (1)", ret);

	/************************************************************/
	ret = egg_markdown_to_text_line_is_rule ("***");
	egg_test_title_assert (test, "is rule (2)", ret);

	/************************************************************/
	ret = egg_markdown_to_text_line_is_rule ("*****");
	egg_test_title_assert (test, "is rule (3)", ret);

	/************************************************************/
	ret = egg_markdown_to_text_line_is_rule ("- - -");
	egg_test_title_assert (test, "is rule (4)", ret);

	/************************************************************/
	ret = egg_markdown_to_text_line_is_rule ("---------------------------------------");
	egg_test_title_assert (test, "is rule (5)", ret);

	/************************************************************/
	ret = egg_markdown_to_text_line_is_rule ("");
	egg_test_title_assert (test, "is rule (blank)", !ret);

	/************************************************************/
	ret = egg_markdown_to_text_line_is_rule ("richard hughes");
	egg_test_title_assert (test, "is rule (text)", !ret);

	/************************************************************/
	ret = egg_markdown_to_text_line_is_rule ("- richard-hughes");
	egg_test_title_assert (test, "is rule (bullet)", !ret);

	/************************************************************/
	ret = egg_markdown_to_text_line_is_blank ("");
	egg_test_title_assert (test, "is blank (blank)", ret);

	/************************************************************/
	ret = egg_markdown_to_text_line_is_blank (" \t ");
	egg_test_title_assert (test, "is blank (mix)", ret);

	/************************************************************/
	ret = egg_markdown_to_text_line_is_blank ("richard hughes");
	egg_test_title_assert (test, "is blank (name)", !ret);

	/************************************************************/
	ret = egg_markdown_to_text_line_is_blank ("ccccccccc");
	egg_test_title_assert (test, "is blank (full)", !ret);

	/************************************************************
	 ****************          formatter           **************
	 ************************************************************/
	text = egg_markdown_to_text_line_formatter ("**is important** text", "**", "<b>", "</b>");
	egg_test_title (test, "formatter (left)");
	if (g_str_equal (text, "<b>is important</b> text"))
		egg_test_success (test, NULL);
	else
		egg_test_failed (test, "failed, got %s", text);
	g_free (text);

	/************************************************************/
	text = egg_markdown_to_text_line_formatter ("this is **important**", "**", "<b>", "</b>");
	egg_test_title (test, "formatter (right)");
	if (g_str_equal (text, "this is <b>important</b>"))
		egg_test_success (test, NULL);
	else
		egg_test_failed (test, "failed, got %s", text);
	g_free (text);

	/************************************************************/
	text = egg_markdown_to_text_line_formatter ("**important**", "**", "<b>", "</b>");
	egg_test_title (test, "formatter (only)");
	if (g_str_equal (text, "<b>important</b>"))
		egg_test_success (test, NULL);
	else
		egg_test_failed (test, "failed, got %s", text);
	g_free (text);

	/************************************************************/
	text = egg_markdown_to_text_line_formatter ("***important***", "**", "<b>", "</b>");
	egg_test_title (test, "formatter (only)");
	if (g_str_equal (text, "<b>*important</b>*"))
		egg_test_success (test, NULL);
	else
		egg_test_failed (test, "failed, got %s", text);
	g_free (text);

	/************************************************************/
	text = egg_markdown_to_text_line_formatter ("I guess * this is * not bold", "*", "<i>", "</i>");
	egg_test_title (test, "formatter (with spaces)");
	if (g_str_equal (text, "I guess * this is * not bold"))
		egg_test_success (test, NULL);
	else
		egg_test_failed (test, "failed, got %s", text);
	g_free (text);

	/************************************************************/
	text = egg_markdown_to_text_line_formatter ("this **is important** text in **several** places", "**", "<b>", "</b>");
	egg_test_title (test, "formatter (middle, multiple)");
	if (g_str_equal (text, "this <b>is important</b> text in <b>several</b> places"))
		egg_test_success (test, NULL);
	else
		egg_test_failed (test, "failed, got %s", text);
	g_free (text);

	/************************************************************/
	text = egg_markdown_discover_auto_code_spans ("this is http://www.hughsie.com/with_spaces_in_url inline link");
	egg_test_title (test, "auto formatter (url)");
	if (g_str_equal (text, "this is `http://www.hughsie.com/with_spaces_in_url` inline link"))
		egg_test_success (test, NULL);
	else
		egg_test_failed (test, "failed, got %s", text);
	g_free (text);

	/************************************************************/
	text = egg_markdown_to_text_line_formatter ("this was \"triffic\" it was", "\"", "“", "”");
	egg_test_title (test, "formatter (quotes)");
	if (g_str_equal (text, "this was “triffic” it was"))
		egg_test_success (test, NULL);
	else
		egg_test_failed (test, "failed, got %s", text);
	g_free (text);

	/************************************************************/
	text = egg_markdown_to_text_line_formatter ("This isn't a present", "'", "‘", "’");
	egg_test_title (test, "formatter (one quote)");
	if (g_str_equal (text, "This isn't a present"))
		egg_test_success (test, NULL);
	else
		egg_test_failed (test, "failed, got %s", text);
	g_free (text);

	/************************************************************
	 ****************          markdown            **************
	 ************************************************************/
	egg_test_title (test, "get EggMarkdown object");
	self = egg_markdown_new ();
	egg_test_assert (test, self != NULL);

	/************************************************************/
	ret = egg_markdown_set_output (self, EGG_MARKDOWN_OUTPUT_PANGO);
	egg_test_title_assert (test, "set pango output", ret);

	/************************************************************/
	markdown = "OEMs\n"
		   "====\n"
		   " - bullet\n";
	markdown_expected =
		   "<big>OEMs</big>\n"
		   "• bullet";
	egg_test_title (test, "markdown (setext-style header)");
	text = egg_markdown_parse (self, markdown);
	if (g_str_equal (text, markdown_expected))
		egg_test_success (test, NULL);
	else
		egg_test_failed (test, "failed, got '%s', expected '%s'", text, markdown_expected);
	g_free (text);


	/************************************************************/
	markdown = "this is http://www.hughsie.com/with_spaces_in_url inline link\n";
	markdown_expected = "this is <tt>http://www.hughsie.com/with_spaces_in_url</tt> inline link";
	egg_test_title (test, "markdown (autocode)");
	egg_markdown_set_autocode (self, TRUE);
	text = egg_markdown_parse (self, markdown);
	if (g_str_equal (text, markdown_expected))
		egg_test_success (test, NULL);
	else
		egg_test_failed (test, "failed, got '%s', expected '%s'", text, markdown_expected);
	g_free (text);

	/************************************************************/
	markdown = "*** This software is currently in alpha state ***\n";
	markdown_expected = "<b><i> This software is currently in alpha state </b></i>";
	egg_test_title (test, "markdown some invalid header");
	text = egg_markdown_parse (self, markdown);
	if (g_str_equal (text, markdown_expected))
		egg_test_success (test, NULL);
	else
		egg_test_failed (test, "failed, got '%s', expected '%s'", text, markdown_expected);
	g_free (text);

	/************************************************************/
	markdown = " - This is a *very*\n"
		   "   short paragraph\n"
		   "   that is not usual.\n"
		   " - Another";
	markdown_expected =
		   "• This is a <i>very</i> short paragraph that is not usual.\n"
		   "• Another";
	egg_test_title (test, "markdown (complex1)");
	text = egg_markdown_parse (self, markdown);
	if (g_str_equal (text, markdown_expected))
		egg_test_success (test, NULL);
	else
		egg_test_failed (test, "failed, got '%s', expected '%s'", text, markdown_expected);
	g_free (text);

	/************************************************************/
	markdown = "*  This is a *very*\n"
		   "   short paragraph\n"
		   "   that is not usual.\n"
		   "*  This is the second\n"
		   "   bullet point.\n"
		   "*  And the third.\n"
		   " \n"
		   "* * *\n"
		   " \n"
		   "Paragraph one\n"
		   "isn't __very__ long at all.\n"
		   "\n"
		   "Paragraph two\n"
		   "isn't much better.";
	markdown_expected =
		   "• This is a <i>very</i> short paragraph that is not usual.\n"
		   "• This is the second bullet point.\n"
		   "• And the third.\n"
		   "⎯⎯⎯⎯⎯⎯⎯⎯⎯⎯⎯⎯⎯⎯⎯⎯⎯⎯⎯⎯⎯⎯\n"
		   "Paragraph one isn't <b>very</b> long at all.\n"
		   "Paragraph two isn't much better.";
	egg_test_title (test, "markdown (complex1)");
	text = egg_markdown_parse (self, markdown);
	if (g_str_equal (text, markdown_expected))
		egg_test_success (test, NULL);
	else
		egg_test_failed (test, "failed, got '%s', expected '%s'", text, markdown_expected);
	g_free (text);

	/************************************************************/
	markdown = "This is a spec file description or\n"
		   "an **update** description in bohdi.\n"
		   "\n"
		   "* * *\n"
		   "# Big title #\n"
		   "\n"
		   "The *following* things 'were' fixed:\n"
		   "- Fix `dave`\n"
		   "* Fubar update because of \"security\"\n";
	markdown_expected =
		   "This is a spec file description or an <b>update</b> description in bohdi.\n"
		   "⎯⎯⎯⎯⎯⎯⎯⎯⎯⎯⎯⎯⎯⎯⎯⎯⎯⎯⎯⎯⎯⎯\n"
		   "<big>Big title</big>\n"
		   "The <i>following</i> things 'were' fixed:\n"
		   "• Fix <tt>dave</tt>\n"
		   "• Fubar update because of \"security\"";
	egg_test_title (test, "markdown (complex2)");
	text = egg_markdown_parse (self, markdown);
	if (g_str_equal (text, markdown_expected))
		egg_test_success (test, NULL);
	else
		egg_test_failed (test, "failed, got '%s', expected '%s'", text, markdown_expected);
	g_free (text);

	/************************************************************/
	markdown = "* list seporated with spaces -\n"
		   "  first item\n"
		   "\n"
		   "* second item\n"
		   "\n"
		   "* third item\n";
	markdown_expected =
		   "• list seporated with spaces - first item\n"
		   "• second item\n"
		   "• third item";
	egg_test_title (test, "markdown (list with spaces)");
	text = egg_markdown_parse (self, markdown);
	if (g_str_equal (text, markdown_expected))
		egg_test_success (test, NULL);
	else
		egg_test_failed (test, "failed, got '%s', expected '%s'", text, markdown_expected);
	g_free (text);

	/************************************************************/
	ret = egg_markdown_set_max_lines (self, 1);
	egg_test_title_assert (test, "set max_lines", ret);

	/************************************************************/
	markdown = "* list seporated with spaces -\n"
		   "  first item\n"
		   "* second item\n";
	markdown_expected =
		   "• list seporated with spaces - first item";
	egg_test_title (test, "markdown (one line limit)");
	text = egg_markdown_parse (self, markdown);
	if (g_str_equal (text, markdown_expected))
		egg_test_success (test, NULL);
	else
		egg_test_failed (test, "failed, got '%s', expected '%s'", text, markdown_expected);
	g_free (text);

	/************************************************************/
	ret = egg_markdown_set_max_lines (self, 1);
	egg_test_title_assert (test, "set max_lines", ret);

	/************************************************************/
	markdown = "* list & spaces";
	markdown_expected =
		   "• list & spaces";
	egg_test_title (test, "markdown (escaping)");
	text = egg_markdown_parse (self, markdown);
	if (g_str_equal (text, markdown_expected))
		egg_test_success (test, NULL);
	else
		egg_test_failed (test, "failed, got '%s', expected '%s'", text, markdown_expected);
	g_free (text);

	/************************************************************/
	egg_test_title (test, "markdown (free text)");
	text = egg_markdown_parse (self, "This isn't a present");
	if (g_str_equal (text, "This isn't a present"))
		egg_test_success (test, NULL);
	else
		egg_test_failed (test, "failed, got '%s'", text);
	g_free (text);

	/************************************************************/
	egg_test_title (test, "markdown (autotext underscore)");
	text = egg_markdown_parse (self, "This isn't CONFIG_UEVENT_HELPER_PATH present");
	if (g_str_equal (text, "This isn't <tt>CONFIG_UEVENT_HELPER_PATH</tt> present"))
		egg_test_success (test, NULL);
	else
		egg_test_failed (test, "failed, got '%s'", text);
	g_free (text);

	/************************************************************/
	markdown = "*Thu Mar 12 12:00:00 2009* Dan Walsh <dwalsh@redhat.com> - 2.0.79-1\n"
		   "- Update to upstream \n"
		   " * Netlink socket handoff patch from Adam Jackson.\n"
		   " * AVC caching of compute_create results by Eric Paris.\n"
		   "\n"
		   "*Tue Mar 10 12:00:00 2009* Dan Walsh <dwalsh@redhat.com> - 2.0.78-5\n"
		   "- Add patch from ajax to accellerate X SELinux \n"
		   "- Update eparis patch\n";
	markdown_expected =
		   "<i>Thu Mar 12 12:00:00 2009</i> Dan Walsh <tt>&lt;dwalsh@redhat.com&gt;</tt> - 2.0.79-1\n"
		   "• Update to upstream\n"
		   "• Netlink socket handoff patch from Adam Jackson.\n"
		   "• AVC caching of compute_create results by Eric Paris.\n"
		   "<i>Tue Mar 10 12:00:00 2009</i> Dan Walsh <tt>&lt;dwalsh@redhat.com&gt;</tt> - 2.0.78-5\n"
		   "• Add patch from ajax to accellerate X SELinux\n"
		   "• Update eparis patch";
	egg_test_title (test, "markdown (end of bullet)");
	egg_markdown_set_escape (self, TRUE);
	ret = egg_markdown_set_max_lines (self, 1024);
	text = egg_markdown_parse (self, markdown);
	if (g_str_equal (text, markdown_expected))
		egg_test_success (test, NULL);
	else
		egg_test_failed (test, "failed, got '%s', expected '%s'", text, markdown_expected);
	g_free (text);

	g_object_unref (self);

	egg_test_end (test);
}
#endif

